Package: moocore
Type: Package
Title: Core Mathematical Functions for Multi-Objective Optimization
Version: 0.3.1
Authors@R: c(person("Manuel", "López-Ibáñez", role = c("aut", "cre"),
                    email = "manuel.lopez-ibanez@manchester.ac.uk",
                    comment = c(ORCID = "0000-0001-9974-1295")),
             person("Carlos", "Fonseca", role = "ctb"),
             person("Luís", "Paquete", role = "ctb"),
             person(c("Andreia", "P."), "Guerreiro", role = "ctb"),
             person("Mickaël", "Binois", role = "ctb"),
             person(c("Michael", "H."), "Buselli", role = "cph", comment = "AVL-tree library"),
             person("Wessel", "Dankers", role = "cph", comment = "AVL-tree library"),
             person("NumPy Developers", role = "cph", comment = "RNG and ziggurat constants"),
             person("Jean-Sebastien", "Roy", role = "cph", comment = "mt19937 library"),
             person("Makoto", "Matsumoto", role = "cph", comment = "mt19937 library"),
             person("Takuji", "Nishimura", role = "cph", comment = "mt19937 library"))
Description: Fast implementations of mathematical operations and performance metrics for multi-objective optimization, including filtering and ranking of dominated vectors according to Pareto optimality, hypervolume metric, C.M. Fonseca, L. Paquete, M. López-Ibáñez (2006) <doi:10.1109/CEC.2006.1688440>, epsilon indicator, inverted generational distance, computation of the empirical attainment function, V.G. da Fonseca, C.M. Fonseca, A.O. Hall (2001) <doi:10.1007/3-540-44719-9_15>, and  Vorob'ev threshold, expectation and deviation, M. Binois, D. Ginsbourger, O. Roustant (2015) <doi:10.1016/j.ejor.2014.07.032>, among others.
Depends: R (>= 4.1)
Imports: matrixStats, Rdpack (>= 2.4)
Suggests: doctest (>= 0.4.0), spelling, testthat (>= 3.2.0), withr
License: LGPL (>= 2.1)
Copyright: file COPYRIGHTS
URL: https://multi-objective.github.io/moocore/r/,
        https://github.com/multi-objective/moocore
BugReports: https://github.com/multi-objective/moocore/issues
LazyLoad: true
LazyData: true
Encoding: UTF-8
UseLTO: true
RoxygenNote: 7.3.3
SystemRequirements: GNU make
RdMacros: Rdpack
Config/testthat/edition: 3
Language: en-GB
Config/Needs/website: Hmisc, data.table, dplyr, ggplot2, htmltools,
        kableExtra, knitr, plotly, rmarkdown, scales, tidyr
NeedsCompilation: yes
Packaged: 2026-05-04 07:15:31 UTC; manu
Author: Manuel López-Ibáñez [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-9974-1295>),
  Carlos Fonseca [ctb],
  Luís Paquete [ctb],
  Andreia P. Guerreiro [ctb],
  Mickaël Binois [ctb],
  Michael H. Buselli [cph] (AVL-tree library),
  Wessel Dankers [cph] (AVL-tree library),
  NumPy Developers [cph] (RNG and ziggurat constants),
  Jean-Sebastien Roy [cph] (mt19937 library),
  Makoto Matsumoto [cph] (mt19937 library),
  Takuji Nishimura [cph] (mt19937 library)
Maintainer: Manuel López-Ibáñez <manuel.lopez-ibanez@manchester.ac.uk>
Repository: CRAN
Date/Publication: 2026-05-04 13:10:03 UTC
Built: R 4.6.0; x86_64-apple-darwin20; 2026-05-04 23:56:49 UTC; unix

Package: moocore
Type: Package
Title: Core Mathematical Functions for Multi-Objective Optimization
Version: 0.3.2
Authors@R: c(person("Manuel", "López-Ibáñez", role = c("aut", "cre"),
                    email = "manuel.lopez-ibanez@manchester.ac.uk",
                    comment = c(ORCID = "0000-0001-9974-1295")),
             person("Carlos", "Fonseca", role = "ctb"),
             person("Luís", "Paquete", role = "ctb"),
             person(c("Andreia", "P."), "Guerreiro", role = "ctb"),
             person("Mickaël", "Binois", role = "ctb"),
             person(c("Michael", "H."), "Buselli", role = "cph", comment = "AVL-tree library"),
             person("Wessel", "Dankers", role = "cph", comment = "AVL-tree library"),
             person("NumPy Developers", role = "cph", comment = "RNG and ziggurat constants"),
             person("Jean-Sebastien", "Roy", role = "cph", comment = "mt19937 library"),
             person("Makoto", "Matsumoto", role = "cph", comment = "mt19937 library"),
             person("Takuji", "Nishimura", role = "cph", comment = "mt19937 library"))
Description: Fast implementations of mathematical operations and performance metrics for multi-objective optimization, including filtering and ranking of dominated vectors according to Pareto optimality, hypervolume metric, C.M. Fonseca, L. Paquete, M. López-Ibáñez (2006) <doi:10.1109/CEC.2006.1688440>, epsilon indicator, inverted generational distance, computation of the empirical attainment function, V.G. da Fonseca, C.M. Fonseca, A.O. Hall (2001) <doi:10.1007/3-540-44719-9_15>, and  Vorob'ev threshold, expectation and deviation, M. Binois, D. Ginsbourger, O. Roustant (2015) <doi:10.1016/j.ejor.2014.07.032>, among others.
Depends: R (>= 4.1)
Imports: matrixStats, Rdpack (>= 2.4)
Suggests: doctest (>= 0.4.0), spelling, testthat (>= 3.2.0), withr
License: LGPL (>= 2.1)
Copyright: file COPYRIGHTS
URL: https://multi-objective.github.io/moocore/r/,
        https://github.com/multi-objective/moocore
BugReports: https://github.com/multi-objective/moocore/issues
LazyLoad: true
LazyData: true
Encoding: UTF-8
UseLTO: true
RoxygenNote: 7.3.3
SystemRequirements: GNU make
RdMacros: Rdpack
Config/testthat/edition: 3
Language: en-GB
Config/Needs/website: Hmisc, data.table, dplyr, ggplot2, htmltools,
        kableExtra, knitr, plotly, rmarkdown, scales, tidyr
NeedsCompilation: yes
Packaged: 2026-07-12 12:24:38 UTC; manu
Author: Manuel López-Ibáñez [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-9974-1295>),
  Carlos Fonseca [ctb],
  Luís Paquete [ctb],
  Andreia P. Guerreiro [ctb],
  Mickaël Binois [ctb],
  Michael H. Buselli [cph] (AVL-tree library),
  Wessel Dankers [cph] (AVL-tree library),
  NumPy Developers [cph] (RNG and ziggurat constants),
  Jean-Sebastien Roy [cph] (mt19937 library),
  Makoto Matsumoto [cph] (mt19937 library),
  Takuji Nishimura [cph] (mt19937 library)
Maintainer: Manuel López-Ibáñez <manuel.lopez-ibanez@manchester.ac.uk>
Repository: CRAN
Date/Publication: 2026-07-12 14:30:02 UTC

# Generated by roxygen2: do not edit by hand

export(any_dominated)
export(as_double_matrix)
export(attsurf2df)
export(avg_hausdorff_dist)
export(choose_eafdiff)
export(compute_eaf_call)
export(compute_eafdiff_call)
export(eaf)
export(eaf_as_list)
export(eafdiff)
export(epsilon_additive)
export(epsilon_mult)
export(filter_dominated)
export(generate_ndset)
export(hv_approx)
export(hv_contributions)
export(hypervolume)
export(igd)
export(igd_plus)
export(is_nondominated)
export(largest_eafdiff)
export(normalise)
export(pareto_rank)
export(r2_exact)
export(rbind_datasets)
export(read_datasets)
export(total_whv_rect)
export(transform_maximise)
export(vorob_dev)
export(vorob_t)
export(whv_hype)
export(whv_rect)
export(write_datasets)
importFrom(Rdpack,reprompt)
importFrom(matrixStats,colRanges)
importFrom(stats,rexp)
importFrom(stats,rnorm)
importFrom(utils,modifyList)
importFrom(utils,tail)
importFrom(utils,write.table)
useDynLib(moocore, .registration = TRUE)

**moocore**: Core Algorithms for Multi-Objective Optimization
=============================================================

<!-- badges: start -->
[![CRAN version](https://www.r-pkg.org/badges/version-last-release/moocore)][r-moocore-cran]
[![CRAN Status](https://badges.cranchecks.info/worst/moocore.svg)][r-moocore-cran-results]
[![CRAN Downloads](https://cranlogs.r-pkg.org/badges/grand-total/moocore)][r-moocore-cran]
[![R build status][r-build-badge]][r-build-link]
[![coverage][r-coverage-badge]][r-coverage-link]
[![r-universe version][r-universe-version-badge]][r-universe-version]
[![r-universe build status][r-universe-build-badge]][r-universe-build-link]
<!-- badges: end -->

[ [**Homepage**][r-moocore-homepage] ]
[ [**GitHub**][r-moocore-github] ]

**Maintainer:** [Manuel López-Ibáñez](https://lopez-ibanez.eu)

**Contributors:**
    [Manuel López-Ibáñez](https://lopez-ibanez.eu),
    [Carlos M. Fonseca](https://eden.dei.uc.pt/~cmfonsec/),
    [Luís Paquete](https://eden.dei.uc.pt/~paquete/),
    and Mickaël Binois.

---------------------------------------

Introduction
============

This webpage documents the `moocore` R package. There is also a [`moocore` Python package](https://multi-objective.github.io/moocore/python/)


The goal of the [**moocore** project](https://github.com/multi-objective/moocore/) is to collect and document fast implementations of core mathematical functions and algorithms for multi-objective optimization and make them available to different programming languages via similar interfaces. These functions include:

 * Generate and transform nondominated sets.
 * Identify and filter dominated vectors.
 * Quality metrics such as (weighted) hypervolume, epsilon, IGD, etc.
 * Computation of the Empirical Attainment Function. The empirical attainment function (EAF) describes the probabilistic
distribution of the outcomes obtained by a stochastic algorithm in the
objective space.

Most critical functionality is implemented in C, with the R and Python packages providing convenient interfaces to the C code.


**Keywords**: empirical attainment function, summary attainment surfaces, EAF
differences, multi-objective optimization, bi-objective optimization,
performance measures, performance assessment

**Relevant literature:**

 1. <a name="LopPaqStu09emaa"/>Manuel López-Ibáñez, Luís Paquete, and Thomas Stützle. [Exploratory Analysis of Stochastic Local Search Algorithms in Biobjective Optimization](https://doi.org/10.1007/978-3-642-02538-9_9). In T. Bartz-Beielstein, M. Chiarandini, L. Paquete, and M. Preuss, editors, *Experimental Methods for the Analysis of Optimization Algorithms*, pages 209–222. Springer, Berlin, Germany, 2010.<br>
    (This chapter is also available in a slightly extended form as Technical Report TR/IRIDIA/2009-015).<br>
    [ [bibtex](https://lopez-ibanez.eu/LopezIbanez_bib.html#LopPaqStu09emaa) |
    doi: [10.1007/978-3-642-02538-9_9](https://doi.org/10.1007/978-3-642-02538-9_9)
    | [Presentation](https://lopez-ibanez.eu/doc/gecco2010moworkshop.pdf) ]


Download and installation
-------------------------

The **moocore** package is implemented in R. Therefore,
a basic knowledge of R is recommended to make use of all features.

The first step before installing the **moocore** package is to [install R](https://cran.r-project.org/). Once R is installed in the system, there are two methods for installing the **moocore** package:

 1. Install within R (automatic download, internet connection required). Invoke
    R, then

    ```r
        install.packages("moocore")
    ```

 2. [Download the **moocore** package from CRAN][r-moocore-cran] (you may also need to download and install
    first the [package matrixStats](https://cran.r-project.org/package=matrixStats)), and invoke at the command-line:

    ```bash
        R CMD INSTALL <package>
    ```
    where `<package>` is one of the three versions available: `.tar.gz` (Unix/BSD/GNU/Linux), `.tgz` (MacOS X), or `.zip` (Windows).

Search the [R documentation](https://cran.r-project.org/faqs.html) if you need more help to install an R package on your system.

Compiled command-line executables for various purposes can be found under `system.file(package="moocore", "bin")`.

If you wish to be notified of bugfixes and new versions, please subscribe to the [low-volume emo-list](https://lists.dei.uc.pt/mailman/listinfo/emo-list), where announcements will be made.

[ [Download **moocore** package from CRAN][r-moocore-cran] ]  [ [Documentation][r-moocore-homepage] ]  [ [Development version (GitHub)][r-moocore-github] ]


GitHub (Development version)
----------------------------

If you wish to try the development version, you can install it by executing the
following command within the R console:

```r
install.packages('moocore', repos = c('https://multi-objective.r-universe.dev', 'https://cloud.r-project.org'))
```

Usage
-----

Once the **moocore** package is installed, the following R commands will give more information:
```r
    library(moocore)
    ?eaf
    ?read_datasets
```


License
--------

This software is Copyright (C) 2024 Manuel López-Ibáñez, Carlos M. Fonseca, Luís Paquete, Mickaël Binois.

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the [GNU
Lesser General Public License](https://www.gnu.org/licenses/old-licenses/lgpl-2.1.html) for more details.

**IMPORTANT NOTE**: Please be aware that the fact that this program is released
as Free Software does not excuse you from scientific propriety, which obligates
you to give appropriate credit! If you write a scientific paper describing
research that made substantive use of this program, it is your obligation as a
scientist to (a) mention the fashion in which this software was used in the
Methods section; (b) mention the algorithm in the References section. The
appropriate citation is:

 * Manuel López-Ibáñez, Luís Paquete, and Thomas Stützle. **Exploratory Analysis of Stochastic Local Search Algorithms in Biobjective Optimization.** In T. Bartz-Beielstein, M. Chiarandini, L. Paquete, and M. Preuss, editors, *Experimental Methods for the Analysis of Optimization Algorithms*, pages 209–222. Springer, Berlin, Germany, 2010.  doi: 10.1007/978-3-642-02538-9_9

Moreover, as a personal note, I would appreciate it if you would email
`manuel.lopez-ibanez@manchester.ac.uk` with citations of papers referencing
this work so I can mention them to my funding agent and tenure committee.

[r-build-badge]: https://github.com/multi-objective/moocore/actions/workflows/R.yml/badge.svg?event=push
[r-build-link]: https://github.com/multi-objective/moocore/actions/workflows/R.yml
[r-coverage-badge]: https://codecov.io/gh/multi-objective/moocore/branch/main/graph/badge.svg?flag=R
[r-coverage-link]: https://app.codecov.io/gh/multi-objective/moocore/tree/main/r
[r-moocore-cran-results]: https://cran.r-project.org/web/checks/check_results_moocore.html
[r-moocore-cran]: https://cran.r-project.org/package=moocore
[r-moocore-github]: https://github.com/multi-objective/moocore/tree/main/r#readme
[r-moocore-homepage]: https://multi-objective.github.io/moocore/r/
[r-universe-build-badge]: https://github.com/r-universe/multi-objective/actions/workflows/build.yml/badge.svg
[r-universe-build-link]: https://github.com/r-universe/multi-objective/actions/workflows/build.yml
[r-universe-version]: https://multi-objective.r-universe.dev/moocore
[r-universe-version-badge]: https://multi-objective.r-universe.dev/badges/moocore

**moocore**: Core Mathematical Functions for Multi-Objective Optimization
=========================================================================

<!-- badges: start -->
[ [**C library**][c-moocore-homepage] ] [ [GitHub][c-moocore-github] ] [![Build status][c-build-badge]][c-build-link] [![Testsuite status][testsuite-badge]][testsuite-link]
<!-- badges: end -->

**Contributors:**
    [Manuel López-Ibáñez](https://lopez-ibanez.eu),
    [Carlos M. Fonseca](https://eden.dei.uc.pt/~cmfonsec/),
    [Luís Paquete](https://eden.dei.uc.pt/~paquete/),
    Andreia P. Guerreiro.



Introduction
============

This is the C library and associated command-line tools of the [**moocore** project](https://github.com/multi-objective/moocore). This README describes just this component. There is also a [`moocore` Python package][py-moocore-homepage] and a [`moocore` R package][r-moocore-homepage].


The [**moocore** project](https://github.com/multi-objective/moocore/) collects fast implementations of core mathematical functions and algorithms for multi-objective optimization and makes them available to different programming languages via similar interfaces. These functions include:

 * Identifying and filtering dominated vectors.
 * Quality metrics such as (weighted) hypervolume, epsilon, IGD, etc.
 * Computation of the Empirical Attainment Function. The empirical attainment
   function (EAF) describes the probabilistic distribution of the outcomes
   obtained by a stochastic algorithm in the objective space.

The goal of **moocore** is to provide an implementation of these functions that is efficient, thoroughly tested, well-documented, multi-platform (Windows, Linux, MacOS), multi-language (C, R, and Python, so far), with minimal dependencies and reusable from other libraries and packages, thus providing an easy-to-use and high-performance building block.

Most critical functionality is implemented in C, with the R and Python packages providing convenient interfaces to the C code.

**Keywords**: empirical attainment function, summary attainment surfaces, EAF differences, multi-objective optimization, bi-objective optimization, performance measures, performance assessment


Compilation
===========

We recommend that you compile specifically for your architecture using GCC option `-march=`. The default compilation is done with:

    make all

This uses the option `"-march=native"`. If your GCC version does not support `"native"`, you can give an explicit architecture, e.g.,

    make all MARCH=x86-64-v2

See the [GCC manual](https://gcc.gnu.org/onlinedocs/gcc/Submodel-Options.html) for the names of the architectures supported by your version of GCC.

You can compile the code with various runtime checks enabled using the option `DEBUG=1`.  This will slow down the code significantly.  If this is too slow for your purposes, you can disable the [sanitizers](https://gcc.gnu.org/onlinedocs/gcc/Instrumentation-Options.html#index-fsanitize_003daddress) using `make all DEBUG=1 SANITIZERS=""`.

The build system will try to pick good compiler flags for you, but if you need to, you can override them by passing the option `OPT_CFLAGS`. For example, to disable compiler optimization and enabled debug symbols, you could run:

    make all OPT_CFLAGS="-O0 -g"


If you do not want to see the command line of each compiler
invocation, also pass `S=1` to `make`.


Embedding (shared library)
==================================

Most functions are implemented in headers such as `nondominated.h`. For other functions, it is possible to build a shared library for embedding the code into your own C/C++ programs:

    make shlibs

Functions exported by the library are marked with `MOOCORE_API`.


Testsuite
=========

The **moocore** executables are validated using a [comprehensive testsuite](https://github.com/multi-objective/testsuite). Running the testsuite requires [Python](https://www.python.org/downloads/) `>= 3.10` and additional Python packages (see [`testsuite/requirements.txt`](https://github.com/multi-objective/testsuite/blob/main/requirements.txt)).  To run the testsuite yourself, follow these steps:

```bash
# Download moocore
git clone https://github.com/multi-objective/moocore moocore
# Download the testsuite
git clone https://github.com/multi-objective/testsuite moocore/testsuite
# Install required python packages
python3 -m pip install -r testsuite/requirements.txt
# Testing
make -C moocore/c/ test
# Timing
make -C moocore/c/ time
```

Under `moocore/c/`, `make test` will compile the code with `DEBUG=1 SANTIZERS="" OPT_CFLAGS="-Og -g3"`. This is useful for debugging failures.

`make time` will compile the code with `DEBUG=0 SANITIZERS="" OPT_CLAGS="-O3 -ftlo -march=native"`. This is useful for benchmarking code and making sure compiler optimizations do not break anything.

Neither `make test` nor `make time` recompile the code that has not been modified. So you can compile the code first with your preferred compiler flags and then run the testsuite. For example, if you compile with `make all DEBUG=1`, the code will be compiler with several [sanitizers](https://gcc.gnu.org/onlinedocs/gcc/Instrumentation-Options.html#index-fsanitize_003daddress) enabled. You can then call `make test` to run the testsuite with the sanitizers enabled.  This is much slower, but very helpful to track crashes.

Under `c/` there are various targets to run only parts of the testsuite, for example:

    make test-hv
    make test-nondominated
    ...



Command-line executables
========================

Input format
------------

All command-line tools can read the input as a filename in the command line:

    hv datafile

or from standard input:

    cat datafile | hv

In an input file, each point is given in a separate line, and each coordinate
within a line is separated by whitespace. An empty line denotes a separate
set. Most tools have an `--union` option that ignores those empty lines.  All
tools assume that all objectives must be minimized but this can be changed via
options `--maximise` or `--obj=`.  Also, maximization objectives may be
multiplied by -1 to convert them to minimization (see the option `--agree` of `nondominated`).


nondominated
------------

Check dominance, filter and transform the sets given as input.

```
Usage:
       nondominated [OPTIONS] [FILES]
       nondominated [OPTIONS] < [INPUT] > [OUTPUT]

Options:
 -h, --help          print this summary and exit;
     --version       print version number (and compilation flags) and exit;
 -v, --verbose       print some extra information;
 -q, --quiet         print as little as possible;
     --no-check      do not check nondominance of sets (faster but unsafe);
 -o, --obj=[+|-]...  specify whether each objective should be minimised (-)
                     or maximised (+). By default all are minimised;
     --maximise      all objectives must be maximised;
 -u, --upper-bound POINT defines an upper bound to check, e.g. "10 5 30";
 -l, --lower-bound POINT defines a lower bound to check;
 -U, --union         consider each file as a whole approximation set,
                     (by default, approximation sets are separated by an
                     empty line within a file);
 -s, --suffix=STRING suffix to add to output files. Default is "_dat".
                     The empty string means overwrite the input file.
                     This is ignored when reading from stdin because output
                     is sent to stdout.

 The following options OVERWRITE output files:
 -a, --agree=<max|min> transform objectives so all are maximised (or
                       minimised). See also the option --obj.
 -f, --filter        check and filter out dominated points;
 -b, --force-bound   remove points that do not satisfy the bounds;
 -n, --normalise RANGE normalise all objectives to a range, e.g., "1 2".
                       If bounds are given with -l and -u, they are used
                       for the normalisation.
 -L, --log=[1|0]...  specify whether each objective should be transformed
                     to logarithmic scale (1) or not (0).
```

hv
--

Computes the exact hypervolume. A reference point can be given by the option `-r`.

    hv -r "10 10 10" data

If no reference point is given, the default is `maximum + 0.1 * (maximum - minimum)`, where the maximum and minimum values are calculated for each
coordinate from the union of all input points.

For other options, check the output of `hv --help`.


hvapprox
--------

Approximate the hypervolume value of each input set of each input file.  The
approximation uses (quasi-)Monte-Carlo sampling, thus gets more accurate with
larger values of `--nsamples`. With no file, or when file is `-`, read standard
input.

```
Usage: hvapprox [OPTIONS] [FILE...]
       hvapprox [OPTIONS] < [INPUT] > [OUTPUT]

Options:
 -h, --help          print this summary and exit;
     --version       print version number (and compilation flags) and exit;
 -v, --verbose       print some information (time, maximum, etc).
 -q, --quiet         print just the hypervolume (as opposed to --verbose).
 -u, --union         treat all input sets within a FILE as a single set.
 -r, --reference=POINT use POINT as reference point. POINT must be within
                     quotes, e.g., "10 10 10". If no reference point is
                     given, it is taken as max + 0.1 * (max - min) for each
                     coordinate from the union of all input points.
 -s, --suffix=STRING Create an output file for each input file by appending
                     this suffix. This is ignored when reading from stdin.
                     If missing, output is sent to stdout.
 -n, --nsamples=N    Number of Monte-Carlo samples (N is a positive integer).
 -m, --method=M      1: Monte-Carlo sampling using normal distribution;
                     2: Hua-Wang deterministic sampling (default).
 -S, --seed=SEED     Seed of the random number generator (positive integer).
                     Only method=1.
```

epsilon
-------

Calculates the epsilon measure for the sets given as input.

```
Usage:
       epsilon [OPTIONS] [FILES]
       epsilon [OPTIONS] < [INPUT] > [OUTPUT]

Options:
 -h, --help          print this summary and exit;
     --version       print version number (and compilation flags) and exit;
 -v, --verbose        print some information (time, number of points, etc.).
 -q, --quiet         print as little as possible;
 -a, --additive       epsilon additive value (default).
 -m, --multiplicative epsilon multiplicative value .
 -r, --reference FILE file that contains the reference set
 -o, --obj=[+|-]...  specify whether each objective should be minimised (-)
                     or maximised (+). By default all are minimised;
     --maximise      all objectives must be maximised;
 -s, --suffix=STRING  Create an output file for each input file by appending
                      this suffix. This is ignored when reading from stdin.
                      If missing, output is sent to stdout.
```

igd
---

Calculates quality metrics related to the generational distance (GD, IGD, IGD+, avg Hausdorff distance).

```
Usage:
       igd [OPTIONS] [FILES]
       igd [OPTIONS] < [INPUT] > [OUTPUT]


Options:
 -h, --help          print this summary and exit;
     --version       print version number (and compilation flags) and exit;
 -v, --verbose       print some information (time, number of points, etc.)
 -q, --quiet         print as little as possible;
   , --gd            report classical GD
   , --igd           report classical IGD
   , --gd-p          report GD_p (p=1 by default)
   , --igd-p         (default) report IGD_p (p=1 by default)
   , --igd-plus      report IGD+
   , --hausdorff     report avg Hausdorff distance = max (GD_p, IGD_p)
 -a, --all           compute everything
 -p,                 exponent that averages the distances
 -r, --reference FILE file that contains the reference set
 -o, --obj=[+|-]...  specify whether each objective should be minimised (-)
                     or maximised (+). By default all are minimised;
     --maximise      all objectives must be maximised;
 -s, --suffix=STRING Create an output file for each input file by appending
                     this suffix. This is ignored when reading from stdin.
                     If missing, output is sent to stdout.
```

eaf
---

Computes the empirical attainment function (EAF) of all input files.
With no file, or when file is `-`, read standard input.

```
Usage:  eaf [OPTIONS] [FILE...]

Options:
 -h, --help          print this summary and exit;
     --version       print version number (and compilation flags) and exit;
 -v, --verbose       print some information (time, input points, output
                     points, etc) in stderr. Default is --quiet
 -o, --output FILE   write output to FILE instead of standard output.
 -q, --quiet         print just the EAF (as opposed to --verbose)
 -b, --best          compute best attainment surface
 -m, --median        compute median attainment surface
 -w, --worst         compute worst attainment surface
 -p, --percentile REAL compute the given percentile of the EAF
 -l, --level  LEVEL    compute the given level of the EAF
 -i[FILE], --indices[=FILE]  write attainment indices to FILE.
                     If FILE is '-', print to stdout.
                     If FILE is missing use the same file as for output.
 -d[FILE], --diff[=FILE] write difference between half of runs to FILE.
                     If FILE is '-', print to stdout.
                     If FILE is missing use the same file as for output.
        , --polygons Write EAF as R polygons.
```

dominatedsets
-------------

Calculates the number of sets from one file that dominate the sets of the other
files.

```
Usage: dominatedsets [OPTIONS] [FILE...]

Options:
 -h, --help          print this summary and exit;
     --version       print version number (and compilation flags) and exit;
 -v, --verbose       print some information (time, number of points, etc.)
 -q, --quiet         print as little as possible;
 -p, --percentages   print results also as percentages.
     --no-check      do not check nondominance of sets (faster but unsafe).
 -o, --obj=[+|-]...  specify whether each objective should be minimised (-)
                     or maximised (+). By default all are minimised;
```

ndsort
------

Perform nondominated sorting in a list of points.

```
Usage: ndsort [OPTIONS] [FILE...]

Options:
 -h, --help          print this summary and exit;
     --version       print version number (and compilation flags) and exit;
 -v, --verbose       print some information (time, number of points, etc.)
 -q, --quiet         print as little as possible;
 -k, --keep-uevs     keep uniquely extreme values
 -r, --rank          don't break ties using hypervolume contribution
 -o, --obj=[+|-]...  specify whether each objective should be minimised (-)
                     or maximised (+). By default all are minimised;
```

License
========

See the [LICENSE](/LICENSE) and [COPYRIGHTS](/r/inst/COPYRIGHTS) files.


[c-build-badge]: https://github.com/multi-objective/moocore/actions/workflows/C.yml/badge.svg?event=push
[c-build-link]: https://github.com/multi-objective/moocore/actions/workflows/C.yml
[c-moocore-github]: https://github.com/multi-objective/moocore/tree/main/c#readme
[c-moocore-homepage]: https://github.com/multi-objective/moocore/tree/main/c#readme
[py-build-badge]: https://github.com/multi-objective/moocore/actions/workflows/python.yml/badge.svg?event=push
[py-build-link]: https://github.com/multi-objective/moocore/actions/workflows/python.yml
[py-coverage-badge]: https://codecov.io/gh/multi-objective/moocore/branch/main/graph/badge.svg?flag=python
[py-coverage-link]: https://app.codecov.io/gh/multi-objective/moocore/tree/main/python
[py-moocore-github]: https://github.com/multi-objective/moocore/tree/main/python#readme
[py-moocore-homepage]: https://multi-objective.github.io/moocore/python/
[py-moocore-pypi]: https://pypi.org/project/moocore/
[r-build-badge]: https://github.com/multi-objective/moocore/actions/workflows/R.yml/badge.svg?event=push
[r-build-link]: https://github.com/multi-objective/moocore/actions/workflows/R.yml
[r-coverage-badge]: https://codecov.io/gh/multi-objective/moocore/branch/main/graph/badge.svg?flag=R
[r-coverage-link]: https://app.codecov.io/gh/multi-objective/moocore/tree/main/r
[r-moocore-cran-results]: https://cran.r-project.org/web/checks/check_results_moocore.html
[r-moocore-cran]: https://cran.r-project.org/package=moocore
[r-moocore-github]: https://github.com/multi-objective/moocore/tree/main/r#readme
[r-moocore-homepage]: https://multi-objective.github.io/moocore/r/
[testsuite-badge]: https://github.com/multi-objective/testsuite/actions/workflows/moocore.yml/badge.svg
[testsuite-link]: https://github.com/multi-objective/testsuite/actions/workflows/moocore.yml

#' Exact R2 indicator
#'
#' Computes the exact R2 indicator with respect to a given ideal/utopian
#' reference point assuming minimization of all objectives.
#'
#' @inherit epsilon params
#'
#' @param reference `numeric()`\cr Reference (ideal) point as a vector of numerical
#'   values.
#'
#' @return `numeric(1)` A single numerical value.
#'
#' @details
#'
#' The unary R2 indicator is a quality indicator for a set \eqn{A \subset
#' \mathbb{R}^m}{A in R^m} w.r.t. an ideal or utopian reference point
#' \eqn{\vec{r} \in \mathbb{R}^m}{r in R^m}.  It was originally proposed by
#' \citet{HanJas1998} and is defined as the expected Tchebycheff utility under a
#' uniform distribution of weight vectors (w.l.o.g. assuming minimization):
#'
#' \deqn{R2(A) := \int_{w \in W} \min_{a \in A} \left\{ \max_{i=1,\dots,m} w_i
#' (a_i - r_i) \right\} \, dw,}{R2(A) := integral over w in W of min over a in
#' A of max over i of w_i (a_i - r_i) dw,}
#'
#' where \eqn{W} denotes the uniform distribution across weights:
#'
#' \deqn{W = \{w \in \mathbb{R}^m \mid w_i \geq 0, \sum_{i=1}^m w_i = 1\}.}{W
#' = \{w in R^m | w_i >= 0, sum_i w_i = 1\}.}
#'
#' The R2 indicator is to be minimized and has an optimal value of 0 when
#' \eqn{\vec{r} \in A}{r in A}.
#'
#' The exact R2 indicator is strongly Pareto-compliant, i.e., compatible with
#' Pareto-optimality:
#'
#' \deqn{\forall A, B \subset \mathbb{R}^m: A \prec B \Rightarrow R2(A) <
#' R2(B).}{for all A, B in R^m: A dominates B implies R2(A) < R2(B).}
#'
#' Given an ideal or utopian reference point, which is available in most
#' scenarios, all non-dominated solutions always contribute to the value of
#' the exact R2 indicator.  However, it is scale-dependent and care should be
#' taken such that all objectives contribute approximately equally to the
#' indicator, e.g., by normalizing the Pareto front to the unit hypercube.
#'
#' The current implementation exclusively supports bi-objective solution sets
#' and runs in \eqn{O(n \log n)} following \citet{SchKer2025r2v2}.
#'
#' @references
#'
#' \insertAllCited{}
#'
#' @doctest
#' dat <- matrix(c(5, 5, 4, 6, 2, 7, 7, 4), ncol = 2, byrow = TRUE)
#' @expect equal(tolerance = 1e-9, 2.5941919191919194)
#' r2_exact(dat, reference = c(0, 0))
#'
#' # This function assumes minimisation by default. We can easily specify maximisation:
#' @expect equal(tolerance = 1e-9, 2.5196969696969695)
#' r2_exact(dat, reference = c(10, 10), maximise = TRUE)
#'
#' # Merge all the sets of a dataset by removing the set number column:
#' extdata_path <- system.file(package="moocore","extdata")
#' dat <- read_datasets(file.path(extdata_path, "example1_dat"))[, 1:2]
#' @expect equal(65L)
#' nrow(dat)
#'
#' # Dominated points are ignored, so this:
#' @expect equal(tolerance = 1e-9, 3865393.493470812)
#' r2_exact(dat, reference = 0)
#'
#' # gives the same exact R2 value as this:
#' dat <- filter_dominated(dat)
#' @expect equal(9L)
#' nrow(dat)
#' @expect equal(tolerance = 1e-9, 3865393.493470812)
#' r2_exact(dat, reference = 0)
#'
#' @export
#' @concept metrics
r2_exact <- function(x, reference, maximise = FALSE)
{
  x <- as_double_matrix(x)
  nobjs <- ncol(x)
  if (nobjs != 2L) stop("sorry: only 2 objectives supported")
  if (!is.numeric(reference))
    stop("a numerical reference vector must be provided")
  if (length(reference) == 1L) reference <- rep_len(reference, nobjs)

  if (any(maximise)) {
    x <- transform_maximise(x, maximise)
    if (all(maximise)) {
      reference <- -reference
    } else {
      reference[maximise] <- -reference[maximise]
    }
  }
  .Call(r2_exact_C,
    t(x),
    as.double(reference))
}

# Generated by doctest: do not edit by hand
# Please edit file in R/r2.R

test_that("Doctest: r2_exact", {
  # Created from @doctest for `r2_exact`
  # Source file: R/r2.R
  dat <- matrix(c(5, 5, 4, 6, 2, 7, 7, 4), ncol = 2, byrow = TRUE)
  expect_equal(r2_exact(dat, reference = c(0, 0)), tolerance = 1e-09,
  2.59419191919192)
  expect_equal(r2_exact(dat, reference = c(10, 10), maximise = TRUE), tolerance = 1e-09,
  2.51969696969697)
  extdata_path <- system.file(package = "moocore", "extdata")
  dat <- read_datasets(file.path(extdata_path, "example1_dat"))[, 1:2]
  expect_equal(nrow(dat), 65L)
  expect_equal(r2_exact(dat, reference = 0), tolerance = 1e-09, 3865393.49347081)
  dat <- filter_dominated(dat)
  expect_equal(nrow(dat), 9L)
  expect_equal(r2_exact(dat, reference = 0), tolerance = 1e-09, 3865393.49347081)
})

test_that("r2_exact mixed maximise", {
  dat <- matrix(c(5, 5, 4, 6, 2, 7, 7, 4), ncol = 2, byrow = TRUE)
  dat[, 1] <- -dat[, 1]
  expect_equal(tolerance = 1e-09,
    r2_exact(dat, reference=c(0,0), maximise=c(TRUE,FALSE)), 2.59419191919192)
  expect_equal(tolerance = 1e-09,
    r2_exact(dat, reference = c(-10, 10), maximise = c(FALSE, TRUE)), 2.51969696969697)
})

test_that("r2_exact", {
  # perfect approximation of ideal ref yields 0.0
  expect_equal(r2_exact(matrix(c(0, 0), ncol = 2), reference = c(0, 0)), 0.0)

  # [1,1] in normalized objective space should yield r2 = 0.75
  expect_equal(r2_exact(matrix(c(2, 2), ncol = 2), reference = c(1, 1)), 0.75)

  # [[1,0],[0,1]] should yield r2 = 0.25
  expect_equal(r2_exact(matrix(c(0, 1, 1, 0), ncol = 2, byrow = TRUE), reference = c(0, 0)), 0.25)

  x <- matrix(c(0, 1, 0.2, 0.8, 0.4, 0.6, 0.6, 0.4, 0.8, 0.2, 1, 0), ncol = 2, byrow = TRUE)
  expect_equal(r2_exact(x, reference = c(0, 0)), 0.1833333333333333, tolerance = 1e-10)

  # a closely sampled linear front should yield a value close to 1/6
  xvals <- seq(0, 1, length.out = 1000001)
  pf <- cbind(xvals, 1 - xvals)
  expect_equal(r2_exact(pf, reference = c(0, 0)), 1 / 6, tolerance = 1e-6)
})

test_that("r2_exact from doc", {
  dat <- matrix(c(5, 5, 4, 6, 2, 7, 7, 4), ncol = 2, byrow = TRUE)
  expect_equal(r2_exact(dat, reference = c(0, 0)), 2.5941919191919194, tolerance = 1e-9)

  expect_equal(r2_exact(dat, reference = c(10, 10), maximise = TRUE), 2.5196969696969695, tolerance = 1e-9)

  extdata_path <- system.file(package = "moocore", "extdata")
  dat <- read_datasets(file.path(extdata_path, "example1_dat"))[, 1:2]
  r2_val <- r2_exact(dat, reference = c(0, 0))
  dat_nondom <- filter_dominated(dat)
  r2_val_nondom <- r2_exact(dat_nondom, reference = c(0, 0))
  # Dominated points are ignored, so both should give the same value
  expect_equal(r2_val, r2_val_nondom)
})

files <- c(
  "dominatedsets",
  "eaf",
  "epsilon",
  "hv",
  "hvapprox",
  "igd",
  "ndsort",
  "nondominated"
)

file_move <- function(from, to) {
  file.copy(from = from, to = to, overwrite = TRUE)
  file.remove(from)
}

if (WINDOWS) files <- paste0(files, ".exe")
if (any(file.exists(files))) {
  dest <- file.path(R_PACKAGE_DIR,  paste0('bin', R_ARCH))
  dir.create(dest, recursive = TRUE, showWarnings = FALSE)
  file_move(files, dest)
}

files <- Sys.glob(paste0("*", SHLIB_EXT))
if (any(file.exists(files))) {
  dest <- file.path(R_PACKAGE_DIR, paste0('libs', R_ARCH))
  dir.create(dest, recursive = TRUE, showWarnings = FALSE)
  file.copy(files, dest, overwrite = TRUE)
}
if (file.exists("symbols.rds"))
  file.copy("symbols.rds", dest, overwrite = TRUE)

#' @keywords internal
"_PACKAGE"

## usethis namespace: start
#' @importFrom matrixStats colRanges
#' @importFrom Rdpack reprompt
#' @importFrom utils modifyList write.table tail
#' @importFrom stats rexp rnorm
#' @useDynLib moocore, .registration = TRUE
## usethis namespace: end
NULL

#' Results of Hybrid GA on Vanzyl and Richmond water networks
#'
#'@format
#'  A list with two data frames, each of them with three columns, as
#'  produced by [read_datasets()].
#'  \describe{
#'    \item{`$vanzyl`}{data frame of results on Vanzyl network}
#'    \item{`$richmond`}{data frame of results on Richmond
#'      network. The second column is filled with `NA`}
#'  }
#'
#'@source \insertRef{LopezIbanezPhD}{moocore}.
#'
#' @examples
#'data(HybridGA)
#'print(HybridGA$vanzyl)
#'print(HybridGA$richmond)
#' @keywords datasets
"HybridGA"

#' Results of SPEA2 when minimising electrical cost and maximising the
#' minimum idle time of pumps on Richmond water network.
#'
#' @format
#'  A data frame as produced by [read_datasets()]. The second
#'  column measures time in seconds and corresponds to a maximisation problem.
#'
#' @source \insertRef{LopezIbanezPhD}{moocore}
#'
#' @examples
#' data(SPEA2minstoptimeRichmond)
#' str(SPEA2minstoptimeRichmond)
#' @keywords datasets
"SPEA2minstoptimeRichmond"

#' Results of SPEA2 with relative time-controlled triggers on Richmond water
#' network.
#'
#' @format
#'  A data frame as produced by [read_datasets()].
#'
#' @source \insertRef{LopezIbanezPhD}{moocore}
#'
#'@examples
#' data(SPEA2relativeRichmond)
#' str(SPEA2relativeRichmond)
#' @keywords datasets
"SPEA2relativeRichmond"

#' Results of SPEA2 with relative time-controlled triggers on Vanzyl's
#' water network.
#'
#' @inherit SPEA2relativeRichmond format source
#'
#'@examples
#' data(SPEA2relativeVanzyl)
#' str(SPEA2relativeVanzyl)
#' @keywords datasets
"SPEA2relativeVanzyl"


#' Conditional Pareto fronts obtained from Gaussian processes simulations.
#'
#' The data has the only goal of providing an example of use of [vorob_t()] and
#' [vorob_dev()]. It has been obtained by fitting two Gaussian processes on 20
#' observations of a bi-objective problem, before generating conditional
#' simulation of both GPs at different locations and extracting non-dominated
#' values of coupled simulations.
#'
#' @format
#'  A data frame with 2967 observations on the following 3 variables.
#'  \describe{
#'    \item{`f1`}{first objective values.}
#'    \item{`f2`}{second objective values.}
#'    \item{`set`}{indices of corresponding conditional Pareto fronts.}
#'  }
#'
#'@source
#' \insertRef{BinGinRou2015gaupar}{moocore}
#'
#'@examples
#' data(CPFs)
#' vorob_t(CPFs, reference = c(2, 200))
#'@keywords datasets
"CPFs"

#' Various strategies of Two-Phase Local Search applied to the Permutation
#' Flowshop Problem with Makespan and Weighted Tardiness objectives.
#'
#' @format
#'  A data frame with 1511 observations of 4 variables:
#'  \describe{
#'    \item{`algorithm`}{TPLS search strategy}
#'    \item{`Makespan`}{first objective values.}
#'    \item{`WeightedTardiness`}{second objective values.}
#'    \item{`run`}{index of the run.}
#'  }
#'
#'@source
#'
#' \insertRef{DubLopStu2011amai}{moocore}
#'
#' @examples
#' data(tpls50x20_1_MWT)
#' str(tpls50x20_1_MWT)
#' @keywords datasets
"tpls50x20_1_MWT"

.libmoocore_constants <- NULL

.onLoad <- function(lib, pkg) {
  .libmoocore_constants <<-  .Call("libmoocore_constants")
  Rdpack::Rdpack_bibstyles(package = pkg, authors = "LongNames")
  invisible(NULL)
}

#' Exact computation of the Empirical Attainment Function (EAF)
#'
#' This function computes the EAF given a set of 2D or 3D points and a vector
#' `set` that indicates to which set each point belongs.
#'
#' @param x `matrix()`|`data.frame()`\cr Matrix or data frame of numerical
#'   values that represents multiple sets of points, where each row represents
#'   a point. If `sets` is missing, the last column of `x` gives the sets.
#'
#' @param sets `integer()`\cr Vector that indicates the set of each point in
#'   `x`. If missing, the last column of `x` is used instead.
#'
#' @param percentiles `numeric()`\cr Vector indicating which percentiles are
#'   computed.  `NULL` computes all.
#'
#' @inheritParams is_nondominated
#'
#' @param groups `factor()`\cr Indicates that the EAF must be computed
#'   separately for data belonging to different groups.
#'
#' @return `data.frame()`\cr A data frame containing the exact representation
#'   of EAF. The last column gives the percentile that corresponds to each
#'   point. If groups is not `NULL`, then an additional column indicates to
#'   which group the point belongs.
#'
#' @author  Manuel \enc{López-Ibáñez}{Lopez-Ibanez}
#'
#' @details
#'
#' The empirical first-order attainment function (EAF) is used to assess the
#' performance of stochastic multiobjective optimisers such as multiobjective
#' evolutionary algorithms \citep{Grunert01}.  It is an estimator for the
#' first-order attainment function, which provides information about the
#' location and, to some extent, the variability of the random sets of
#' nondominated objective vectors produced by such optimisers when applied to
#' given problem instances.
#'
#' Given a set \eqn{A \subset \mathbb{R}^d}, the attainment function of
#' \eqn{A}, denoted by \eqn{\alpha_{A}\colon \mathbb{R}^d\to \{0,1\}},
#' specifies which points in the objective space are weakly dominated by
#' \eqn{A}, where \eqn{\alpha_A(\vec{z}) = 1} if \eqn{\exists \vec{a} \in A,
#' \vec{a} \leq \vec{z}}, and \eqn{\alpha_A(\vec{z}) = 0}, otherwise.
#'
#' Let \eqn{\mathcal{A} = \{A_1, \dots, A_n\}} be a multi-set of \eqn{n} sets
#' \eqn{A_i \subset \mathbb{R}^d}, the EAF \citep{Grunert01,GruFon2009:emaa} is
#' the function \eqn{\hat{\alpha}_{\mathcal{A}}\colon \mathbb{R}^d\to [0,1]},
#' such that:
#'
#'  \deqn{\hat{\alpha}_{\mathcal{A}}(\vec{z}) = \frac{1}{n}\sum_{i=1}^n \alpha_{A_i}(\vec{z})}
#'
#' The EAF is a coordinate-wise non-decreasing step function, similar to the
#' empirical cumulative distribution function (ECDF)
#' \citep{LopVerDreDoe2025}.  Thus, a finite representation of the EAF
#' can be computed as the set of minima, in terms of Pareto optimality, with a
#' value of the EAF not smaller than a given \eqn{t/n}, where
#' \eqn{t=1,\dots,n} \citep{FonGueLopPaq2011emo}. Formally, the EAF can
#' be represented by the sequence \eqn{(L_1, L_2, \dots, L_n)}, where:
#'
#' \deqn{L_t = \min \{\vec{z} \in \mathbb{R}^d : \hat{\alpha}_{\mathcal{A}}(\vec{z}) \geq t/n\}}
#'
#' It is also common to refer to the \eqn{k\% \in [0,100]} percentile. For
#' example, the *median* (or 50%) attainment surface corresponds to
#' \eqn{L_{\lceil n/2 \rceil}} and it is the lower boundary of the vector space
#' attained by at least 50% of the input sets \eqn{A_i}. Similarly, \eqn{L_1}
#' is called the *best* attainment surface (\eqn{\frac{1}{n}}%) and
#' represents the lower boundary of the space attained by at least one input
#' set, whereas \eqn{L_{100}} is called the *worst* attainment surface (100%)
#' and represents the lower boundary of the space attained by all input sets.
#'
#' In the current implementation, the EAF is computed using the algorithms
#' proposed by \citet{FonGueLopPaq2011emo}, which have complexity \eqn{\Theta(m\log
#' m + nm)} in 2D and \eqn{O(n^2 m \log m)} in 3D, where \eqn{n} is the number
#' of input sets and \eqn{m} is the total number of input points.
#'
#' @note
#'
#' There are several examples of data sets in
#' `system.file(package="moocore","extdata")`.  The current implementation only
#' supports two and three dimensional points.
#'
#' @references
#'
#' \insertAllCited{}
#'
#'@seealso [read_datasets()]
#'
#'@examples
#' extdata_path <- system.file(package="moocore", "extdata")
#'
#' x <- read_datasets(file.path(extdata_path, "example1_dat"))
#' # Compute full EAF (sets is the last column)
#' str(eaf(x))
#'
#' # Compute only best, median and worst
#' str(eaf(x[,1:2], sets = x[,3], percentiles = c(0, 50, 100)))
#'
#' x <- read_datasets(file.path(extdata_path, "spherical-250-10-3d.txt"))
#' y <- read_datasets(file.path(extdata_path, "uniform-250-10-3d.txt"))
#' x <- rbind(data.frame(x, groups = "spherical"),
#'            data.frame(y, groups = "uniform"))
#' # Compute only median separately for each group
#' z <- eaf(x[,1:3], sets = x[,4], groups = x[,5], percentiles = 50)
#' str(z)
#' @concept eaf
#' @export
eaf <- function (x, sets, percentiles = NULL, maximise = FALSE, groups = NULL)
{
  if (missing(sets)) {
    sets <- x[, ncol(x)]
    x <- x[, -ncol(x), drop=FALSE]
  }

  x <- as_double_matrix(x)
  maximise <- as.logical(maximise)
  x <- transform_maximise(x, maximise)

  if (!is.null(percentiles)) {
    # FIXME: We should handle only integral levels inside the C code.
    percentiles <- unique.default(sort.int(as.numeric(percentiles)))
  }

  if (is.null(groups))
    return(compute_1_eaf(x, sets = sets, percentiles = percentiles, maximise = maximise))

  groups <- factor(groups)
  # FIXME: Is there a multi-variate tapply? Maybe data.table?
  do.call("rbind", lapply(levels(groups), function(g){
    where <- groups == g
    data.frame(compute_1_eaf(x[where, , drop=FALSE], sets = sets[where], percentiles = percentiles, maximise = maximise),
      groups = g)
  }))
}

#' Convert a list of attainment surfaces to a single EAF `data.frame`.
#'
#' @param x `list()`\cr List of `data.frames` or matrices. The names of the list
#'   give the percentiles of the attainment surfaces.  This is the format
#'   returned by [eaf_as_list()].
# FIXME: and [mooplot::eafplot()].
#'
#' @return `data.frame()`\cr Data frame with as many columns as objectives and an additional column `percentiles`.
#'
#' @seealso [eaf_as_list()]
#' @examples
#'
#' data(SPEA2relativeRichmond)
#' attsurfs <- eaf_as_list(eaf(SPEA2relativeRichmond, percentiles = c(0,50,100)))
#' str(attsurfs)
#' eaf_df <- attsurf2df(attsurfs)
#' str(eaf_df)
#' @concept eaf
#' @export
attsurf2df <- function(x)
{
  if (!is.list(x) || is.data.frame(x))
    stop("'x' must be a list of data.frames or matrices")

  percentiles <- as.numeric(names(x))
  percentiles <- rep.int(percentiles, sapply(x, nrow))
  x <- do.call("rbind", x)
  # Remove duplicated points (keep only the higher values)
  uniq <- !duplicated(x, fromLast = TRUE)
  cbind(x[uniq, , drop = FALSE], percentiles = percentiles[uniq])
}

compute_1_eaf <- function(x, sets, percentiles, maximise)
{
  # This function assumes that x has been already transformed according to maximise.
  if (anyNA(sets)) stop("'sets' must have only non-NA numerical values")
  order_sets <- order(sets)
  sets <- sets[order_sets]
  # The C code expects points within a set to be contiguous.
  x <- x[order_sets, , drop=FALSE]
  nsets <- nunique(sets)
  ## if (is.null(percentiles)) {
  ##   # FIXME: We should compute this in the C code.
  ##   percentiles <- 1L:nsets * (100 / nsets)
  ## }
  x <- compute_eaf_call(x, cumsizes = cumsum(unique_counts(sets)), percentiles = percentiles)
  # Undo previous transformation
  if (any(maximise))
      x[,-ncol(x)] <- transform_maximise(x[, -ncol(x), drop=FALSE], maximise)
  x
}

#' Same as [eaf()] but performs no checks and does not transform the input or
#' the output. This function should be used by other packages that want to
#' avoid redundant checks and transformations.
#'
#' @seealso [as_double_matrix()] [transform_maximise()]
#' @inherit eaf title params return
#' @param cumsizes `integer()`\cr Cumulative size of the different sets of points in `x`.
#' @concept eaf
#' @export
compute_eaf_call <- function(x, cumsizes, percentiles)
  .Call(compute_eaf_C,
    t(x),
    cumsizes,
    percentiles)

#' Convert an EAF data frame to a list of data frames, where each element
#' of the list is one attainment surface. The function [attsurf2df()] can be
#' used to convert the list into a single data frame.
#'
#' @param eaf `data.frame()`|`matrix()`\cr Data frame or matrix that represents the EAF.
#'
#' @return `list()`\cr A list of data frames. Each `data.frame` represents one attainment surface.
#'
#' @seealso [eaf()] [attsurf2df()]
#'
#' @examples
#' extdata_path <- system.file(package="moocore", "extdata")
#' x <- read_datasets(file.path(extdata_path, "example1_dat"))
#' attsurfs <- eaf_as_list(eaf(x, percentiles = c(0, 50, 100)))
#' str(attsurfs)
#' @concept eaf
#' @export
eaf_as_list <- function(eaf)
{
  # FIXME: Add groups argument.
  setcol <- ncol(eaf)
  if (!is.null(colnames(eaf)) && colnames(eaf)[setcol] == "groups") {
    return(split.data.frame(eaf[, -c(setcol-1L, setcol), drop=FALSE],
      list(eaf[, setcol-1L], eaf[, setcol]), sep="-"))
  }
  split.data.frame(eaf[,-setcol, drop=FALSE], factor(eaf[,setcol]))
}

### Local Variables:
### ess-indent-level: 2
### ess-continued-statement-offset: 2
### ess-brace-offset: 0
### ess-expression-offset: 4
### ess-else-offset: 0
### ess-brace-imaginary-offset: 0
### ess-continued-brace-offset: 0
### ess-arg-function-offset: 2
### ess-close-brace-offset: 0
### indent-tabs-mode: nil
### ess-fancy-comments: nil
### End:

source("helper-common.R")

test_that("eaf", {

  test_eaf_dataset <- function(name, percentiles = NULL) {
    dataset <- get(name)
    x <- eaf(dataset, percentiles = percentiles)
    # FIXME: work-around for change in the computation
    x[,3] <- floor(x[,3])
    #saveRDS(x, paste0(name, "-eaf.rds"))
    return(x)
  }
  test_eaf_file <- function(file, percentiles = NULL) {
    dataset <- read_datasets(file)
    x <- eaf(dataset, percentiles = percentiles)
    #saveRDS(x, paste0(basename(file), "-eaf.rds"))
    return(x)
  }
  expect_equal(test_eaf_file(extdata_path("ALG_1_dat.xz")),
               readRDS("ALG_1_dat-eaf.rds"))
  expect_equal(test_eaf_dataset("SPEA2relativeRichmond"),
               readRDS("SPEA2relativeRichmond-eaf.rds"))

  for (i in seq_len(399))
    expect_equal(anyDuplicated(eaf(cbind(0:i, 0:i), sets=0:i)[,1]), 0L)
})

test_that("eafs_sets_non_numeric", {
  x <- matrix(1:10, ncol=2)
  expect_equal(eaf(x, sets=1:5), eaf(x, sets=letters[1:5]))
})

#' Combine datasets `x` and `y` by row taking care of making all sets unique.
#'
#' @param x,y `matrix`|`data.frame()`\cr Each dataset has at least three
#'   columns, the last one is the set of each point. See also
#'   [read_datasets()].
#'
#' @return `matrix()|`data.frame()`\cr A dataset.
#' @examples
#' x <- data.frame(f1 = 5:10, f2 = 10:5, set = 1:6)
#' y <- data.frame(f1 = 15:20, f2 = 20:15, set = 1:6)
#' rbind_datasets(x,y)
#' @export
rbind_datasets <- function(x, y)
{
  setcol <- ncol(x)
  stopifnot(setcol > 2L)
  stopifnot(ncol(x) == ncol(y))
  # FIXME: We could relax this condition by re-encoding  the column.
  stopifnot(min(x[,setcol]) == 1L)
  stopifnot(min(y[,setcol]) == 1L)
  # We have to make all sets unique.
  y[,setcol] <- y[,setcol] + max(x[,setcol])
  rbind(x, y)
}


combine_cumsizes_sets <- function(sets_x, sets_y)
{
  cumsizes <- cumsum(unique_counts(sets_x))
  c(cumsizes, cumsizes[length(cumsizes)] + cumsum(unique_counts(sets_y)))
}

#' Approximate the hypervolume indicator.
#'
#' Approximate the value of the hypervolume metric with respect to a given
#' reference point assuming minimization of all objectives. Methods
#' `"Rphi-FWE+"` and `"DZ2019-HW"` are deterministic and ignore the parameter
#' `seed`, while `method="DZ2019-MC"` relies on Monte-Carlo sampling
#' \citep{DenZha2019approxhv}.  All methods tend to get more accurate with
#' higher values of `nsamples`, but the increase in accuracy is not monotonic,
#' as shown in the vignette \HTMLVignette{hv_approx}{Approximating the
#' hypervolume}.
#'
#' @author Manuel \enc{López-Ibáñez}{Lopez-Ibanez}
#'
#' @details
#'
#' All available methods approximate the hypervolume as a
#' \eqn{(m-1)}-dimensional integral over the surface of hypersphere
#' \citep{DenZha2019approxhv}:
#'
#' \deqn{\widehat{HV}_r(A) = \frac{2\pi^\frac{m}{2}}{\Gamma(\frac{m}{2})}\frac{1}{m 2^m}\frac{1}{n}\sum_{i=1}^n \max_{y \in A} s(w^{(i)}, y)^m}
#'
#' where \eqn{m} is the number of objectives, \eqn{w^{(i)}} are weights
#' uniformly distributed on \eqn{S_{+}}, i.e., the positive orthant of the
#' \eqn{(m-1)}-D unit hypersphere, \eqn{n} is the number of weights sampled,
#' \eqn{\Gamma()} is the gamma function [gamma()], i.e., the analytical
#' continuation of the factorial function, and \eqn{s(w, y) = \min_{k=1}^m (r_k
#' - y_k)/w_k}.
#'
#' In the default `method="Rphi-FWE+"` \citep{Lop2026hvapprox}, the weights
#' \eqn{w^{(i)}, i=1\ldots n} are defined using the deterministic
#' low-discrepancy sequence \eqn{R_\phi} \citep{Rob2018unreasonable} mapped to
#' the positive orthant of the hypersphere using a modified version of Fang and
#' Wang efficient mapping \citep{FanWan1994numtheory}.
#'
#' In `method="DZ2019-HW"` \citep{DenZha2019approxhv}, the weights
#' \eqn{w^{(i)}, i=1\ldots n} are defined using a deterministic low-discrepancy
#' sequence. The weight values depend on their number (`nsamples`), thus
#' increasing the number of weights may not necessarily increase accuracy
#' because the set of weights would be different.
#'
#' In `method="DZ2019-MC"` \citep{DenZha2019approxhv}, the weights
#' \eqn{w^{(i)}, i=1\ldots n} are sampled from the unit normal vector such that
#' each weight \eqn{w = \frac{|x|}{\|x\|_2}} where each component of \eqn{x} is
#' independently sampled from the standard normal distribution
#' \citep{Muller1959sphere}.
#'
#' The original source code in C++/MATLAB for both `"DZ2019-HW"` and
#' `"DZ2019-MC"` methods can be found at
#' <https://github.com/Ksrma/Hypervolume-Approximation-using-polar-coordinate>.
#'
#' \citet{Lop2026hvapprox} empirically shows that `"Rphi-FWE+"` typically
#' produces an approximation error as low as the other methods, with a
#' computational cost similar to `"DZ2019-MC"` and significantly faster than
#' `"DZ2019-HW"`.
#'
#' @inherit whv_hype params return
#'
#' @param method `character(1)`\cr Method to generate the sampling weights. See `Details'.
#'
#' @author Manuel \enc{López-Ibáñez}{Lopez-Ibanez}
#'
#' @references
#'
#' \insertAllCited{}
#'
#' @doctest
#'
#' x <- matrix(c(5, 5, 4, 6, 2, 7, 7, 4), ncol=2, byrow=TRUE)
#' @expect equal(38.0)
#' hypervolume(x, ref=10)
#' @expect equal(37.999979)
#' hv_approx(x, ref=10, method="Rphi-FWE+")
#' @expect equal(37.999958)
#' hv_approx(x, ref=10, method="DZ2019-HW")
#' @expect equal(38.000806)
#' hv_approx(x, ref=10, seed=42, method="DZ2019-MC")
#'
#' @export
#' @concept metrics
hv_approx <- function(x, reference, maximise = FALSE, nsamples = 262144L, seed = NULL,
                      method = c("Rphi-FWE+", "DZ2019-HW", "DZ2019-MC"))
{
  method <- match.arg(method)
  x <- as_double_matrix(x)
  nobjs <- ncol(x)

  if (!is.numeric(reference))
    stop("a numerical reference vector must be provided")
  if (length(reference) == 1L) reference <- rep_len(reference, nobjs)
  stopifnot(length(reference) == nobjs)

  if (length(maximise) == 1L) maximise <- rep_len(maximise, nobjs)
  stopifnot(length(maximise) == nobjs)
  check_dimension_max(nobjs, .libmoocore_constants[["MOOCORE_HVAPPROX_DIMENSION_MAX"]])

  if (!is_wholenumber(nsamples) || nsamples <= 0 || nsamples > 2147483648)
    stop("nsamples (", nsamples, ") must be a positive integer value smaller than 2147483648")

  if (method == "DZ2019-MC") {
    seed <- if (is.null(seed)) get_seed() else as_integer(seed)
    return(.Call(hv_approx_dz2019_mc_C,
      t(x),
      as.double(reference),
      as.logical(maximise),
      as.integer(nsamples),
      seed))
  } else if (method == "DZ2019-HW") {
    return(.Call(hv_approx_dz2019_hw_C,
      t(x),
      as.double(reference),
      as.logical(maximise),
      as.integer(nsamples)))
  } else  {
    # match.arg() should already give an error if method is wrong.
    stopifnot(method == "Rphi-FWE+")
    return(.Call(hv_approx_rphi_fang_wang_plus_C,
      t(x),
      as.double(reference),
      as.logical(maximise),
      as.integer(nsamples)))
  }
}

# Generated by doctest: do not edit by hand
# Please edit file in R/hv_approx.R

test_that("Doctest: hv_approx", {
  # Created from @doctest for `hv_approx`
  # Source file: R/hv_approx.R
  x <- matrix(c(5, 5, 4, 6, 2, 7, 7, 4), ncol = 2, byrow = TRUE)
  expect_equal(hypervolume(x, ref = 10), 38)
  expect_equal(hv_approx(x, ref = 10, method = "Rphi-FWE+"), 37.999979)
  expect_equal(hv_approx(x, ref = 10, method = "DZ2019-HW"), 37.999958)
  expect_equal(hv_approx(x, ref = 10, seed = 42, method = "DZ2019-MC"), 38.000806)
})

test_that("hv_approx errors", {
  x <- matrix(c(5, 5, 4, 6), ncol = 2, byrow = TRUE)
  expect_error(hv_approx(x, ref = NULL), "a numerical reference vector must be provided")
  expect_error(hv_approx(x, ref = 10, method = "None"), "'arg' should be one of ")
  expect_equal(hv_approx(x, ref = 1), 0)
})

for (dim in seq(3L, 10L)) {
  test_that(paste0("hv_approx dim=", dim), {
    x <- matrix(replicate(dim, 0.5), ncol=dim)
    ref <- 1.0
    true_hv <- hypervolume(x, ref=ref)

    ## Precision goes down significantly with higher dimensions.
    signif <- if (dim < 4) 3 else 2
    expect_equal(hv_approx(x, ref=ref, method="DZ2019-MC", seed=42), true_hv,
      tolerance = 10**-signif, info = paste0("dim=", dim, ", signif=", signif,
        " error=", log10(abs(true_hv - appr_hv) / true_hv)))

    signif <- if (dim < 8) 4 else if (dim < 10) 3 else 2
    expect_equal(hv_approx(x, ref=ref, method="DZ2019-HW"), true_hv,
      tolerance = 10**-signif, info = paste0("dim=", dim, ", signif=", signif,
        " error=", log10(abs(true_hv - appr_hv) / true_hv)))

    # method="Rphi-FWE+" is the default
    signif <- if (dim < 6) 4 else if (dim < 8) 3 else 2
    expect_equal(hv_approx(x, ref=ref), true_hv,
      tolerance = 10**-signif, info = paste0("dim=", dim, ", signif=", signif,
        " error=", log10(abs(true_hv - appr_hv) / true_hv)))
  })
}

range_df_list <- function(x, col)
{
  # FIXME: How to make this faster?
  do.call(range, lapply(x, `[`, , col))
}

get_ideal <- function(x, maximise)
{
  # FIXME: Is there a better way to do this?
  minmax <- colRanges(as.matrix(x))
  lower <- minmax[,1L]
  upper <- minmax[,2L]
  ifelse(maximise, upper, lower)
}

nunique <- function(x) length(unique.default(x))

# FIXME: There must be something faster than table
unique_counts <- function(x) as.vector(table(x))

#' Transform matrix according to maximise parameter
#'
#' @inheritParams is_nondominated
#'
#' @return `x` transformed such that every column where `maximise` is `TRUE` is multiplied by `-1`.
#'
#' @examples
#' x <- data.frame(f1=1:10, f2=101:110)
#' rownames(x) <- letters[1:10]
#' transform_maximise(x, maximise=c(FALSE,TRUE))
#' transform_maximise(x, maximise=TRUE)
#' x <- as.matrix(x)
#' transform_maximise(x, maximise=c(FALSE,TRUE))
#' transform_maximise(x, maximise=TRUE)
#'
#' @export
transform_maximise <- function(x, maximise)
{
  if (any(maximise)) {
    if (all(maximise))
      return(-x)
    if (length(maximise) != ncol(x))
      stop("length of maximise must be either 1 or ncol(x)")
    x[,maximise] <- -x[,maximise, drop=FALSE]
  }
  x
}

#' Convert input to a matrix with `"double"` storage mode ([base::storage.mode()]).
#'
#' @param x `data.frame()`|`matrix()`\cr A numerical data frame or matrix with at least 1 row and 2 columns.
#' @return `x` is coerced to a numerical `matrix()`.
#' @export
as_double_matrix <- function(x)
{
  name <- deparse(substitute(x))   # FIXME: Only do this if there is an error.
  if (length(dim(x)) != 2L)
    stop("'", name, "' must be a data.frame or a matrix")
  if (nrow(x) < 1L)
    stop("not enough points (rows) in '", name, "'")
  if (ncol(x) < 2L)
    stop("'", name, "' must have at least 2 columns")
  x <- as.matrix(x)
  if (!is.numeric(x))
    stop("'", name, "' must be numeric")
  if (storage.mode(x) != "double")
    storage.mode(x) <- "double"
  x
}

# Same a as_double_matrix() but allow 1-column matrix.
as_double_matrix_1 <- function(x)
{
  name <- deparse(substitute(x))   # FIXME: Only do this if there is an error.
  if (length(dim(x)) != 2L)
    stop("'", name, "' must be a data.frame or a matrix")
  if (nrow(x) < 1L)
    stop("not enough points (rows) in '", name, "'")
  if (ncol(x) < 1L)
    stop("'", name, "' must have at least 1 column")
  x <- as.matrix(x)
  if (!is.numeric(x))
    stop("'", name, "' must be numeric")
  if (storage.mode(x) != "double")
    storage.mode(x) <- "double"
  x
}

is_wholenumber <- function(x, tol = .Machine$double.eps^0.5)
  is.finite(x) & (abs(x - round(x)) < tol)

as_integer <- function(x)
{
  if (!is_wholenumber(x)) {
    stop("'", deparse(substitute(x)), "' is not an integer: ", x)
  }
  as.integer(x)
}

check_dimension_max <- function(ncol, ncol_max)
{
  if (ncol > ncol_max)
    stop("number of columns (", ncol, ") cannot be larger than ", ncol_max)
}

#' Compute empirical attainment function differences
#'
#' Calculate the differences between the empirical attainment functions of two
#' data sets.
#'
#' @param x,y `matrix`|`data.frame()`\cr Data frames corresponding to the input data of
#'   left and right sides, respectively. Each data frame has at least three
#'   columns, the last one is the set of each point. See also
#'   [read_datasets()].
#'
#' @param intervals `integer(1)`\cr The absolute range of the differences
#'   \eqn{[0, 1]} is partitioned into the number of intervals provided.
#'
#' @inheritParams is_nondominated
#'
#' @param rectangles `logical(1)`\cr If TRUE, the output is in the form of rectangles of the same color.
#'
#' @details
#'   This function calculates the differences between the EAFs of two
#'   data sets.
#'
#' @return With `rectangle=FALSE`, a `data.frame` containing points where there
#'   is a transition in the value of the EAF differences.  With
#'   `rectangle=TRUE`, a `matrix` where the first 4 columns give the
#'   coordinates of two corners of each rectangle. In both cases, the last
#'   column gives the difference in terms of sets in `x` minus sets in `y` that
#'   attain each point (i.e., negative values are differences in favour `y`).
#'
#' @seealso    [read_datasets()]
# , [mooplot::eafdiffplot()]
#'
#' @doctest
#' A1 <- read_datasets(text='
#'  3 2
#'  2 3
#'
#'  2.5 1
#'  1 2
#'
#'  1 2
#' ')
#'
#' A2 <- read_datasets(text='
#'  4 2.5
#'  3 3
#'  2.5 3.5
#'
#'  3 3
#'  2.5 3.5
#'
#'  2 1
#' ')
#' d <- eafdiff(A1, A2)
#' @omit
#' str(d)
#' @expect true(is.matrix(.))
#' d
#' @testRaw  expect_equal(d,matrix(byrow=TRUE, ncol=3, scan(quiet = TRUE, text='1.0  2.0    2
#' @testRaw   2.0  1.0   -1
#' @testRaw   2.5  1.0    0
#' @testRaw   2.0  2.0    1
#' @testRaw   2.0  3.0    2
#' @testRaw   3.0  2.0    2
#' @testRaw   2.5  3.5    0
#' @testRaw   3.0  3.0    0
#' @testRaw   4.0  2.5    1')))
#'
#' d <- eafdiff(A1, A2, rectangles = TRUE)
#' @omit
#' str(d)
#' d
#' @testRaw expect_equal(d, as.matrix(read.table(header=TRUE, text='
#' @testRaw      xmin ymin xmax ymax diff
#' @testRaw      2.0  1.0  2.5  2.0   -1
#' @testRaw      1.0  2.0  2.0  Inf    2
#' @testRaw      2.5  1.0  Inf  2.0    0
#' @testRaw      2.0  2.0  3.0  3.0    1
#' @testRaw      2.0  3.5  2.5  Inf    2
#' @testRaw      2.0  3.0  3.0  3.5    2
#' @testRaw      3.0  2.5  4.0  3.0    2
#' @testRaw      3.0  2.0  Inf  2.5    2
#' @testRaw      4.0  2.5  Inf  3.0    1')))
#'@concept eaf
#'@export
eafdiff <- function(x, y, intervals = NULL, maximise = FALSE, rectangles = FALSE)
{
  stopifnot(ncol(x) == ncol(y))

  sets_x <- x[, ncol(x)]
  if (anyNA(sets_x)) stop("'sets' must have only non-NA numerical values")
  cumsizes_x <- cumsum(unique_counts(sets_x))
  # The C code expects points within a set to be contiguous.
  x <- as_double_matrix(x[order(sets_x) , -ncol(x), drop=FALSE])

  sets_y <- y[, ncol(y)]
  if (anyNA(sets_y)) stop("'sets' must have only non-NA numerical values")
  cumsizes_y <- cumsum(unique_counts(sets_y))
  # The C code expects points within a set to be contiguous.
  y <- as_double_matrix(y[order(sets_y), -ncol(y), drop=FALSE])

  nsets <- length(cumsizes_x) + length(cumsizes_y)
  if (is.null(intervals)) {
    # Default is nsets / 2
    intervals <- nsets / 2.0
  } else {
    stopifnot(length(intervals) == 1L)
    intervals <- min(intervals, nsets / 2.0)
  }

  maximise <- as.logical(maximise)
  if (any(maximise)) {
    x <- transform_maximise(x, maximise)
    y <- transform_maximise(y, maximise)
  }

  DIFF <- compute_eafdiff_call(x, y, cumsizes_x, cumsizes_y,
    intervals = intervals, ret = if (rectangles) "rectangles" else "points")
  # FIXME: We should remove duplicated rows in C code.
  # FIXME: Check that we do not generate duplicated nor overlapping rectangles
  # with different colors. That would be a bug.
  DIFF <- DIFF[!duplicated(DIFF),]
  if (any(maximise)) {
    if (rectangles && length(maximise) != 1L)
      maximise <- c(maximise, maximise)
    DIFF[, -ncol(DIFF)] <- transform_maximise(DIFF[, -ncol(DIFF), drop=FALSE], maximise)
  }
  # Undo previous transformation
  ## if (any(maximise)) {
  ##   DIFF[,-ncol(x)] <- transform_maximise(x[, -ncol(x), drop=FALSE], maximise)
  ## }
  DIFF
}

#' Same as [eafdiff()] but performs no checks and does not transform the input
#' or the output. This function should be used by other packages that want to
#' avoid redundant checks and transformations.
#'
#' @seealso [as_double_matrix()] [transform_maximise()]
#' @inherit eafdiff title params return
#' @param cumsizes_x,cumsizes_y Cumulative size of the different sets of points in `x` and `y`.
#' @param ret (`"points"|"rectangles"|"polygons"`)\cr The format of the returned EAF differences.
#' @concept eaf
#' @export
compute_eafdiff_call <- function(x, y, cumsizes_x, cumsizes_y, intervals, ret)
{
  x <- cbind(t(x), t(y))
  cumsizes <- c(cumsizes_x, cumsizes_x[length(cumsizes_x)] + cumsizes_y)
  intervals <- as.integer(intervals)
  switch(ret,
    rectangles = .Call(compute_eafdiff_rectangles_C, x, cumsizes, intervals),
    polygons = .Call(compute_eafdiff_polygon_C, x, cumsizes, intervals),
    points = .Call(compute_eafdiff_C, x, cumsizes, intervals))
}

compute_eafdiff <- function(x, y, sets_x, sets_y, intervals)
  compute_eafdiff_call(x, y, cumsum(unique_counts(sets_x)), cumsum(unique_counts(sets_y)), intervals, ret = "points")

# FIXME: The default intervals should be nsets / 2
compute_eafdiff_rectangles <- function(x, y, sets_x, sets_y, intervals)
  compute_eafdiff_call(x, y, cumsum(unique_counts(sets_x)), cumsum(unique_counts(sets_y)), intervals, ret = "rectangles")

# FIXME: The default intervals should be nsets / 2
compute_eafdiff_polygon <- function(x, y, sets_x, sets_y, intervals)
  compute_eafdiff_call(x, y, cumsum(unique_counts(sets_x)), cumsum(unique_counts(sets_y)), intervals, ret = "polygons")



## compute_eafdiff <- function(data, intervals)
## {
##   DIFF <- compute_eafdiff_call(data[,c(1L,2L), drop=FALSE], sets = data[,3L], intervals = intervals)
##   #print(DIFF)
##   # FIXME: Do this computation in C code. See compute_eafdiff_area_C
##   setcol <- ncol(data)
##   eafval <- DIFF[, setcol]
##   eafdiff <- list(left  = unique(DIFF[ eafval >= 1L, , drop=FALSE]),
##                   right = unique(DIFF[ eafval <= -1L, , drop=FALSE]))
##   eafdiff$right[, setcol] <- -eafdiff$right[, setcol]
##   eafdiff
## }

# This file is loaded automatically by testthat (supposedly)
extdata_path <- function(file)
  file.path(system.file(package = "moocore"), "extdata", file)

read_extdata <- function(file) read_datasets(extdata_path(file))

save_csv_xz <- function(code, pattern)
{
  path <- tempfile(fileext = ".csv.xz")
  res <- code
  write.table(file = withr::local_connection(xzfile(path, "wb")),
    res, row.names = FALSE, col.names=FALSE, sep=",")
  path
}

compare_file_text_compressed <- function(old, new)
{
  if (compare_file_binary(old, new))
    return(TRUE)
  old <- base::readLines(withr::local_connection(gzfile(old, open = "rb")), warn = FALSE)
  new <- base::readLines(withr::local_connection(gzfile(new, open = "rb")), warn = FALSE)
  identical(old, new)
}

expect_snapshot_csv_xz <- function(name, code)
{
  # skip_on_ci() # Skip for now until we implement this: https://github.com/tidyverse/ggplot2/blob/main/tests/testthat/helper-vdiffr.R
  name <- paste0(name, ".csv.xz")
  # Announce the file before touching `code`. This way, if `code`
  # unexpectedly fails or skips, testthat will not auto-delete the
  # corresponding snapshot file.
  testthat::announce_snapshot_file(name = name)
  path <- save_csv_xz(code)
  testthat::expect_snapshot_file(path, name = name, compare = compare_file_text_compressed)
}

is_wholenumber <- function(x, tol = .Machine$double.eps^0.5)
  abs(x - round(x)) < tol

source("helper-common.R")

withr::with_output_sink("test-whv_hype.Rout", {

find_minmax <- function(...)
{
  args <- list(...)
  lower <- list()
  upper <- list()
  for (x in args) {
    stopifnot(is.numeric(x))
    minmax <- apply(x[,1:2], 2L, range)
    lower <- c(lower, list(minmax[1L, , drop = FALSE]))
    upper <- c(upper, list(minmax[2L, , drop = FALSE]))
  }
  lower <- apply(do.call("rbind", lower), 2L, min)
  upper <- apply(do.call("rbind", upper), 2L, min)
  list(lower = lower, upper = upper)
}

test_that("whv_hype", {

  x <- read_extdata("wrots_l10w100_dat")
  y <- read_extdata("wrots_l100w10_dat")

  r <- find_minmax(x, y)
  ideal <- r$lower
  ref <- 1.1 * r$upper
  goal <- colMeans(x[,1:2])

  x_list <- split.data.frame(x[,1:2], x[,3])
  y_list <- split.data.frame(y[,1:2], y[,3])
  set.seed(12345)
  whv_x <- whv_hype(x_list[[1]], reference = ref, ideal = ideal)
  expect_equal(whv_x, 2480979524388, tolerance=10)

  whv_x <- whv_hype(x_list[[1]], reference = ref, ideal = ideal,
                    dist = "point", mu = goal)
  expect_equal(whv_x, 1496335657875, tolerance=10)

  whv_x <- whv_hype(x_list[[1]], reference = ref, ideal = ideal,
                    dist = "exponential", mu=0.2)
  expect_equal(whv_x, 1903385037871, tolerance=10)
})


test_that("whv_hype mixed min/max", {
  expect_equal(whv_hype(matrix(-2, ncol = 2), reference = -4, ideal = -1, seed = 42, maximise=TRUE),
    3.99807)
  expect_equal(whv_hype(matrix(c(-2, 2), ncol = 2), reference = c(-4, 4), ideal = c(-1, 1), seed = 42, maximise=c(TRUE,FALSE)),
    3.99807)

})

test_that("whv_hype requires mu for non-uniform dist", {
  expect_error(
    whv_hype(matrix(2, ncol = 2), reference = 4, ideal = 1, dist = "point"),
    "mu.*required"
  )
})



}) # withr::with_output_sink

## FIXME: We need smaller data!
test_that("eafdiff2d", {
  test_eafdiff <- function(a, b, maximise = FALSE, rectangles = FALSE, ...) {
    A1 <- read_datasets(file.path(system.file(package="moocore"), "extdata", a))
    A2 <- read_datasets(file.path(system.file(package="moocore"), "extdata", b))
    minmax <- ""
    if (any(maximise)) {
      minmax <- paste0("-",
        paste0(ifelse(maximise, "max", "min"), collapse="-"))
      A1[, which(maximise)] <- -A1[, which(maximise)]
      A2[, which(maximise)] <- -A2[, which(maximise)]
    }
    # FIXME: It would be much faster to just test the case maximise=FALSE
    # against the snapshot then test that all possible maximise variants are
    # equivalent to maximise=FALSE after negating the appropriate columns.
    a <- tools::file_path_sans_ext(basename(a))
    b <- tools::file_path_sans_ext(basename(b))
    res <- as.data.frame(eafdiff(A1, A2, maximise = maximise, ...))
    expect_true(all(is_wholenumber(res[, ncol(res)])))
    # If we do not save the last column as integer we get spurious differences.
    res[, ncol(res)] <- as.integer(round(res[,ncol(res)]))

    expect_snapshot_csv_xz(paste0("eafdiff-", a, "-", b, minmax, ifelse(rectangles,"-rect","")),
      res)
  }
  test_eafdiff("wrots_l10w100_dat", "wrots_l100w10_dat")
  test_eafdiff("wrots_l10w100_dat", "wrots_l100w10_dat", maximise = c(TRUE, FALSE))
  test_eafdiff("wrots_l10w100_dat", "wrots_l100w10_dat", maximise = c(FALSE, TRUE))
  test_eafdiff("wrots_l10w100_dat", "wrots_l100w10_dat", maximise = c(TRUE, TRUE))
  test_eafdiff("tpls.xz", "rest.xz")
  test_eafdiff("ALG_1_dat.xz", "ALG_2_dat.xz")
})

#' Vorob'ev threshold, expectation and deviation
#'
#' Compute Vorob'ev threshold, expectation and deviation. Also, displaying the
#' symmetric deviation function is possible.  The symmetric deviation
#' function is the probability for a given target in the objective space to
#' belong to the symmetric difference between the Vorob'ev expectation and a
#' realization of the (random) attained set.
#'
#' @inheritParams eaf
#' @inheritParams hypervolume
#'
#' @return `vorob_t` returns a list with elements `threshold`,
#'   `ve`, and `avg_hyp` (average hypervolume)
#'
#' @details
#'
#' Let \eqn{\mathcal{A} = \{A_1, \dots, A_n\}} be a multi-set of \eqn{n} sets
#' \eqn{A_i \subset \mathbb{R}^d} of mutually nondominated vectors, with finite
#' (but not necessarily equal) cardinality.  If bounded by a reference point
#' \eqn{\vec{r}} that is strictly dominated by any point in any set, then these
#' sets can be seen a samples from a random closed set
#' \citep{Molchanov2005theory}.
#'
#' Let the \eqn{\beta}-quantile be the subset of the empirical attainment
#' function \eqn{\mathcal{Q}_\beta = \{\vec{z}\in \mathbb{R}^d :
#' \hat{\alpha}_{\mathcal{A}}(\vec{z}) \geq \beta\}}.
#'
#' The Vorob'ev *expectation* is the \eqn{\beta^{*}}-quantile set
#' \eqn{\mathcal{Q}_{\beta^{*}}} such that the mean value hypervolume of the
#' sets is equal (or as close as possible) to the hypervolume of
#' \eqn{\mathcal{Q}_{\beta^{*}}}, that is, \eqn{\text{hyp}(\mathcal{Q}_\beta)
#' \leq \mathbb{E}[\text{hyp}(\mathcal{A})] \leq
#' \text{hyp}(\mathcal{Q}_{\beta^{*}})}, \eqn{\forall \beta > \beta^{*}}. Thus,
#' the Vorob'ev expectation provides a definition of the notion of *mean*
#' nondominated set.
#'
#' The value \eqn{\beta^{*} \in [0,1]} is called the Vorob'ev
#' *threshold*. Large differences from the median quantile (0.5) indicate a
#' skewed distribution of \eqn{\mathcal{A}}.
#'
#' The Vorob'ev *deviation* is the mean hypervolume of the symmetric difference
#' between the Vorob'ev expectation and any set in \eqn{\mathcal{A}}, that is,
#' \eqn{\mathbb{E}[\text{hyp}(\mathcal{Q}_{\beta^{*}} \ominus \mathcal{A})]},
#' where the symmetric difference is defined as \eqn{A \ominus B = (A \setminus
#' B) \cup (B \setminus A)}.  Low deviation values indicate that the sets are
#' very similar, in terms of the location of the weakly dominated space, to the
#' Vorob'ev expectation.
#'
#' For more background, see
#' \citet{BinGinRou2015gaupar,Molchanov2005theory,CheGinBecMol2013moda}.
#'
#' @rdname Vorob
#' @author Mickael Binois
#' @doctest
#' data(CPFs)
#' res <- vorob_t(CPFs, reference = c(2, 200))
#' @expect equal(44.140625)
#' res$threshold
#' @expect equal(8943.3332)
#' res$avg_hyp
#' # Now print Vorob'ev deviation
#' vd <- vorob_dev(CPFs, ve = res$ve, reference = c(2, 200))
#' @expect equal(3017.1299)
#' vd
#' @references
#' \insertAllCited{}
#'
#' @concept eaf
#' @export
vorob_t <- function(x, sets, reference, maximise = FALSE)
{
  if (missing(sets)) {
    sets <- x[, ncol(x)]
    x <- x[, -ncol(x), drop=FALSE]
  }
  x <- as_double_matrix(x)

  if (any(maximise)) {
    x <- transform_maximise(x, maximise)
    if (all(maximise)) {
      reference <- -reference
    } else {
      reference[maximise] <- -reference[maximise]
    }
  }

  # First step: compute average hypervolume over conditional Pareto fronts
  avg_hyp <- mean(sapply(split.data.frame(x, sets), hypervolume, reference = reference))

  prev_hyp <- diff <- Inf # hypervolume of quantile at previous step
  a <- 0
  b <- 100
  setcol <- ncol(x) + 1L
  while (diff != 0) {
    c <- (a + b) / 2
    ve <- eaf(x, sets = sets, percentiles = c)[,-setcol]
    tmp <- hypervolume(ve, reference = reference)
    if (tmp > avg_hyp) a <- c else b <- c
    diff <- prev_hyp - tmp
    prev_hyp <- tmp
  }
  ve <- transform_maximise(ve, maximise)
  list(threshold = c, ve = ve, avg_hyp = avg_hyp)
}

#' @concept eaf
#' @rdname Vorob
#' @param ve `matrix()`\cr Vorob'ev expectation, e.g., as returned by [vorob_t()].
#' @return `vorob_dev` returns the Vorob'ev deviation.
#' @export
vorob_dev <- function(x, sets, reference, ve = NULL, maximise = FALSE)
{
  # FIXME: Does it make sense to call this function with 'x' different than the
  # one used to calculate ve? If not, then we should merge them and avoid a lot
  # of redundant work.
  if (missing(sets)) {
    sets <- x[, ncol(x)]
    x <- x[, -ncol(x), drop=FALSE]
  }
  if (is.null(ve)) {
    ve <- vorob_t(x, sets, reference = reference, maximise = maximise)$ve
  } else {
    x <- as_double_matrix(x)
  }

  if (any(maximise)) {
    x <- transform_maximise(x, maximise)
    ve <- transform_maximise(ve, maximise)
    if (all(maximise)) {
      reference <- -reference
    } else {
      reference[maximise] <- -reference[maximise]
    }
  }
  setcol <- ncol(x)
  # Hypervolume of the symmetric difference between A and B:
  # 2 * H(AUB) - H(A) - H(B)
  h2 <- hypervolume(ve, reference = reference)
  x_split <- split.data.frame(x, sets)
  h1 <- mean(sapply(x_split, hypervolume, reference = reference))

  hv_union_ve <- function(y)
    hypervolume(rbind(y, ve), reference = reference)

  vd <- 2 * sum(sapply(x_split, hv_union_ve))
  nruns <- length(x_split)
  ((vd / nruns) - h1 - h2)
}

test_that("vorob", {
  # FIXME: What is this test exactly testing?
  test_data <- expand.grid(seq(0, 1, length.out = 51),
                           seq(0, 1, length.out = 51))
  test_data <- as.matrix(cbind(test_data, nrun = 1:nrow(test_data)))

  # Average hypervolume is known to be 0.25
  # avg_hyp <- mean(sapply(split.data.frame(test_data[,1:2], test_data[,3]),
  #                        hypervolume, reference = c(1, 1)))
  expect_equal(hypervolume(vorob_t(x = test_data, reference = c(1, 1))$ve, reference = c(1, 1)), 0.25, tolerance = 1e-1)
  expect_equal(vorob_dev(x = test_data, ve = vorob_t(test_data, reference = c(1,1))$ve, reference = c(1, 1)), 0.218, tolerance = 1e-1)

})

# Generated by doctest: do not edit by hand
# Please edit file in R/vorob.R

test_that("Doctest: vorob_t", {
  # Created from @doctest for `vorob_t`
  # Source file: R/vorob.R
  data(CPFs)
  res <- vorob_t(CPFs, reference = c(2, 200))
  expect_equal(res$threshold, 44.140625)
  expect_equal(res$avg_hyp, 8943.3332)
  vd <- vorob_dev(CPFs, ve = res$ve, reference = c(2, 200))
  expect_equal(vd, 3017.1299)
})

#' Normalise points
#'
#' Normalise points per coordinate to a range, e.g., `c(1,2)`, where the
#' minimum value will correspond to 1 and the maximum to 2. If bounds are
#' given, they are used for the normalisation.
#'
#' @inheritParams is_nondominated
#'
#' @param to_range `numerical(2)`\cr Normalise values to this range. If the objective is
#'   maximised, it is normalised to `c(to_range[1], to_range[0])`
#'   instead.
#'
#' @param lower,upper `numerical()`\cr Bounds on the values. If `NA`, the maximum and minimum
#'   values of each coordinate are used.
#'
#' @return `matrix()`\cr A numerical matrix
#'
#' @author Manuel \enc{López-Ibáñez}{Lopez-Ibanez}
#'
#' @examples
#'
#' data(SPEA2minstoptimeRichmond)
#' # The second objective must be maximized
#' head(SPEA2minstoptimeRichmond[, 1:2])
#'
#' head(normalise(SPEA2minstoptimeRichmond[, 1:2], maximise = c(FALSE, TRUE)))
#'
#' head(normalise(SPEA2minstoptimeRichmond[, 1:2], to_range = c(0,1), maximise = c(FALSE, TRUE)))
#'
#' @export
normalise <- function(x, to_range = c(1, 2), lower = NA, upper = NA, maximise = FALSE)
{
  x <- as_double_matrix(x)
  nobjs <- ncol(x)
  check_dimension_max(nobjs, .libmoocore_constants[["MOOCORE_DIMENSION_MAX"]])

  lower <- rep_len(as.double(lower), nobjs)
  upper <- rep_len(as.double(upper), nobjs)
  # Handle NA
  no.lower <- is.na(lower)
  no.upper <- is.na(upper)
  minmax <- colRanges(x)
  lower[no.lower] <- minmax[no.lower, 1L]
  upper[no.upper] <- minmax[no.upper, 2L]
  maximise <- rep_len(as.logical(maximise), nobjs)

  if (length(to_range) != 2L)
    stop("'to_range' must be a vector of length 2")

  x <- t.default(x)
  .Call(normalise_C,
    x, # This is modified by normalise_C
    as.double(to_range),
    lower, upper, maximise)
  # FIXME: Transposing in C may avoid a copy.
  t.default(x)
}

source("helper-common.R")

test_that("normalise", {
  my.2d.matrix <- function(...) matrix(c(...), ncol = 2, byrow=FALSE)
  x = my.2d.matrix(0, 0.5, 1, 0, 0.5, 1)

  expect_equal(normalise(x), my.2d.matrix(1, 1.5, 2, 1, 1.5, 2))

  expect_equal(normalise(x, maximise = c(FALSE,TRUE)),
               my.2d.matrix(1, 1.5, 2, 2, 1.5, 1))

  expect_equal(normalise(x, to_range = c(0, 1), maximise = c(FALSE,TRUE)),
               my.2d.matrix(0, 0.5, 1, 1, 0.5, 0))

  expect_equal(normalise(my.2d.matrix(1,1,2,2)), my.2d.matrix(1,1,1,1))

  df <- as.data.frame(x)
  expect_equal(normalise(df),
               as.matrix(data.frame(V1=c(1,1.5,2), V2=c(1,1.5,2))))
})

#' Compute (total) weighted hypervolume given a set of rectangles
#'
#' Calculates the hypervolume weighted by a set of rectangles (with zero weight
#' outside the rectangles). The function [total_whv_rect()] calculates the
#' total weighted hypervolume as [hypervolume()]` + scalefactor *
#' abs(prod(reference - ideal)) * whv_rect()`. The details of the computation
#' are given by \citet{DiaLop2020ejor}.
#'
#' @inherit hypervolume params return
#'
#' @param rectangles `matrix()`\cr Weighted rectangles that will bias the
#'   computation of the hypervolume. Maybe generated by [eafdiff()] with
#'   `rectangles=TRUE` or by [choose_eafdiff()].
#'
#' @details
#'   TODO
#'
#' @seealso [read_datasets()], [eafdiff()], [choose_eafdiff()], [whv_hype()]
#'
#' @doctest
#' rectangles <- as.matrix(read.table(header=FALSE, text='
#'  1.0  3.0  2.0  Inf    1
#'  2.0  3.5  2.5  Inf    2
#'  2.0  3.0  3.0  3.5    3
#' '))
#' @expect equal(4)
#' whv_rect(matrix(2, ncol=2), rectangles, reference = 6)
#' @expect equal(4)
#' whv_rect(matrix(c(2, 1), ncol=2), rectangles, reference = 6)
#' @expect equal(7)
#' whv_rect(matrix(c(1, 2), ncol=2), rectangles, reference = 6)
#'
#' @expect equal(26)
#' total_whv_rect(matrix(2, ncol=2), rectangles, reference = 6, ideal = c(1,1))
#' @expect equal(30)
#' total_whv_rect(matrix(c(2, 1), ncol=2), rectangles, reference = 6, ideal = c(1,1))
#' @expect equal(37.5)
#' total_whv_rect(matrix(c(1, 2), ncol=2), rectangles, reference = 6, ideal = c(1,1))
#'
#' @references
#' \insertAllCited{}
#'
#'@concept metrics
#'@export
whv_rect <- function(x, rectangles, reference, maximise = FALSE)
{
  x <- as_double_matrix(x)
  nobjs <- ncol(x)
  if (nobjs != 2L) stop("sorry: only 2 objectives supported")
  if (ncol(rectangles) != 5L) stop("rectangles: invalid number of columns")
  if (is.null(reference)) stop("reference cannot be NULL")
  if (length(reference) == 1L) reference <- rep_len(reference, nobjs)
  # FIXME: This code does not handle maximisation yet.
  stopifnot(maximise == FALSE)
  ## if (any(maximise)) {
  ##   x <- transform_maximise(x, maximise)
  ##   if (length(maximise) == 1L) {
  ##     reference <- -reference
  ##     rectangles[,1L:4L] <- -rectangles[,1L:4L, drop = FALSE]
  ##   } else {
  ##     reference[maximise] <- -reference[maximise]
  ##     pos <- as.vector(matrix(1L:4L, nrow=2L)[, maximise, drop=FALSE])
  ##     rectangles[,pos] <- -rectangles[,pos, drop = FALSE]
  ##   }
  ## }
  .Call(rect_weighted_hv2d_C,
    t(x),
    t(rectangles),
    reference)
}


#' @inheritParams largest_eafdiff
#'
#' @param scalefactor `numeric(1)`\cr Real value within \eqn{(0,1]} that scales
#'   the overall weight of the differences. This is parameter psi (\eqn{\psi})
#'   in \citet{DiaLop2020ejor}.
#'
#' @rdname whv_rect
#' @concept metrics
#' @export
total_whv_rect <- function(x, rectangles, reference, maximise = FALSE, ideal = NULL, scalefactor = 0.1)
{
  x <- as.matrix(x)
  nobjs <- ncol(x)
  maximise <- as.logical(rep_len(maximise, nobjs))
  if (nobjs != 2L) stop("sorry: only 2 objectives supported")
  if (ncol(rectangles) != 5L) stop("invalid number of columns in rectangles (should be 5)")
  if (scalefactor <= 0 || scalefactor > 1) stop("scalefactor must be within (0,1]")

  hv <- hypervolume(x, reference, maximise = maximise)
  whv <- whv_rect(x, rectangles, reference, maximise = maximise)
  if (is.null(ideal)) {
    # FIXME: Should we include the range of the rectangles here?
    ideal <- get_ideal(x, maximise = maximise)
  }
  if (length(ideal) != nobjs) {
    stop("'ideal' should have same length as ncol(x)")
  }
  beta <- scalefactor * abs(prod(reference - ideal))
  #cat("beta: ", beta, "\n")
  hv + beta * whv
}

#' Approximation of the (weighted) hypervolume by Monte-Carlo sampling (2D only)
#'
#' Return an estimation of the hypervolume of the space dominated by the input
#' data following the procedure described by \citet{AugBadBroZit2009gecco}. A
#' weight distribution describing user preferences may be specified.
#'
#' @inheritParams hypervolume
#'
#' @param ideal `numeric()`\cr Ideal point as a vector of numerical values.
#'
#' @param nsamples `integer(1)`\cr Number of samples for Monte-Carlo
#'   sampling. Higher values typically produce more accurate approximations of
#'   the true hypervolume, but require more time.
#'
#' @param seed `integer(1)`\cr Random seed.
#'
#' @param dist `character(1)`\cr Weight distribution type. See Details.
#'
#' @param mu `numeric()`\cr Parameter of the weight distribution. See Details.
#'
#' @details
#' The current implementation only supports 2 objectives.
#'
#' A weight distribution  \citep{AugBadBroZit2009gecco} can be provided via the `dist` argument. The ones currently supported are:
#'  * `"uniform"` corresponds to the default hypervolume (unweighted).
#'  * `"point"` describes a goal in the objective space, where the parameter `mu` gives the coordinates of the goal. The resulting weight distribution is a multivariate normal distribution centred at the goal.
#' * `"exponential"` describes an exponential distribution with rate parameter `1/mu`, i.e., \eqn{\lambda = \frac{1}{\mu}}.
#'
#' @return A single numerical value.
#'
#' @references
#' \insertAllCited{}
#'
#' @seealso    [read_datasets()], [eafdiff()], [whv_rect()]
#'
#' @doctest
#' @expect equal(3.99807)
#' whv_hype(matrix(2, ncol=2), reference = 4, ideal = 1, seed = 42)
#' @expect equal(3.00555)
#' whv_hype(matrix(c(3,1), ncol=2), reference = 4, ideal = 1, seed = 42)
#' @expect equal(1.14624)
#' whv_hype(matrix(2, ncol=2), reference = 4, ideal = 1, seed = 42,
#'          dist = "exponential", mu=0.2)
#' @expect equal(1.66815)
#' whv_hype(matrix(c(3,1), ncol=2), reference = 4, ideal = 1, seed = 42,
#'          dist = "exponential", mu=0.2)
#' @expect equal(0.64485)
#' whv_hype(matrix(2, ncol=2), reference = 4, ideal = 1, seed = 42,
#'          dist = "point", mu=c(2.9,0.9))
#' @expect equal(4.03632)
#' whv_hype(matrix(c(3,1), ncol=2), reference = 4, ideal = 1, seed = 42,
#'          dist = "point", mu=c(2.9,0.9))
#' @concept metrics
#' @export
whv_hype <- function(x, reference, ideal, maximise = FALSE,
                     nsamples = 100000L, seed = NULL, dist = "uniform", mu = NULL)
{
  x <- as_double_matrix(x)
  nobjs <- ncol(x)

  if (is.null(reference)) stop("reference cannot be NULL")
  if (length(reference) == 1L) reference <- rep_len(reference, nobjs)
  if (is.null(ideal)) stop("ideal cannot be NULL")
  if (length(ideal) == 1L) ideal <- rep_len(ideal, nobjs)
  if (nobjs != 2L) stop("sorry: only 2 objectives supported")

  if (any(maximise)) {
    if (all(maximise)) {
      x <- -x
      reference <- -reference
      ideal <- -ideal
    } else if (length(maximise) != nobjs) {
      stop("length of maximise must be either 1 or ncol(x)")
    } else {
      x[,maximise] <- -x[,maximise]
      reference[maximise] <- -reference[maximise]
      ideal[maximise] <- -ideal[maximise]
    }
  }

  seed <- if (is.null(seed)) get_seed() else as_integer(seed)

  if (dist != "uniform" && is.null(mu)) {
    stop("parameter 'mu' is required when dist ='", dist, "'")
  }

  .Call(whv_hype_C,
    t(x),
    as.double(ideal),
    as.double(reference),
    as.integer(nsamples),
    dist,
    seed,
    mu)
}

get_seed <- function() sample.int(.Machine$integer.max, 1)

# Generated by doctest: do not edit by hand
# Please edit file in R/whv.R

test_that("Doctest: whv_hype", {
  # Created from @doctest for `whv_hype`
  # Source file: R/whv.R
  expect_equal(whv_hype(matrix(2, ncol = 2), reference = 4, ideal = 1, seed = 42),
  3.99807)
  expect_equal(whv_hype(matrix(c(3, 1), ncol = 2), reference = 4, ideal = 1,
  seed = 42), 3.00555)
  expect_equal(whv_hype(matrix(2, ncol = 2), reference = 4, ideal = 1, seed = 42,
  dist = "exponential", mu = 0.2), 1.14624)
  expect_equal(whv_hype(matrix(c(3, 1), ncol = 2), reference = 4, ideal = 1,
  seed = 42, dist = "exponential", mu = 0.2), 1.66815)
  expect_equal(whv_hype(matrix(2, ncol = 2), reference = 4, ideal = 1, seed = 42,
  dist = "point", mu = c(2.9, 0.9)), 0.64485)
  expect_equal(whv_hype(matrix(c(3, 1), ncol = 2), reference = 4, ideal = 1,
  seed = 42, dist = "point", mu = c(2.9, 0.9)), 4.03632)
})

# Generated by doctest: do not edit by hand
# Please edit file in R/whv.R

test_that("Doctest: whv_rect", {
  # Created from @doctest for `whv_rect`
  # Source file: R/whv.R
  rectangles <- as.matrix(read.table(header = FALSE, text = "\n 1.0  3.0  2.0  Inf    1\n 2.0  3.5  2.5  Inf    2\n 2.0  3.0  3.0  3.5    3\n"))
  expect_equal(whv_rect(matrix(2, ncol = 2), rectangles, reference = 6), 4)
  expect_equal(whv_rect(matrix(c(2, 1), ncol = 2), rectangles, reference = 6), 4)
  expect_equal(whv_rect(matrix(c(1, 2), ncol = 2), rectangles, reference = 6), 7)
  expect_equal(total_whv_rect(matrix(2, ncol = 2), rectangles, reference = 6,
  ideal = c(1, 1)), 26)
  expect_equal(total_whv_rect(matrix(c(2, 1), ncol = 2), rectangles, reference = 6,
  ideal = c(1, 1)), 30)
  expect_equal(total_whv_rect(matrix(c(1, 2), ncol = 2), rectangles, reference = 6,
  ideal = c(1, 1)), 37.5)
})

#' Epsilon metric
#'
#' Computes the epsilon metric, either additive or multiplicative.
#'
#' @inheritParams is_nondominated
#'
#' @param reference `matrix`|`data.frame`\cr Reference set as a matrix or
#'   data.frame of numerical values.
#'
#' @return  `numeric(1)` A single numerical value.
#'
#' @name epsilon
#'
#' @author Manuel \enc{López-Ibáñez}{Lopez-Ibanez}
#'
#' @details
#'
#' The epsilon metric of a set \eqn{A \subset \mathbb{R}^m} with respect to a
#' reference set \eqn{R \subset \mathbb{R}^m} is defined as
#'
#' \deqn{epsilon(A,R) = \max_{r \in R} \min_{a \in A} \max_{1 \leq i \leq m} epsilon(a_i, r_i)}
#'
#' where \eqn{a} and \eqn{r} are objective vectors of length \eqn{m}.
#'
#' In the case of minimization of objective \eqn{i}, \eqn{epsilon(a_i,r_i)} is
#' computed as \eqn{a_i/r_i} for the multiplicative variant (respectively,
#' \eqn{a_i - r_i} for the additive variant), whereas in the case of
#' maximization of objective \eqn{i}, \eqn{epsilon(a_i,r_i) = r_i/a_i} for the
#' multiplicative variant (respectively, \eqn{r_i - a_i} for the additive
#' variant). This allows computing a single value for problems where some
#' objectives are to be maximized while others are to be minimized. Moreover, a
#' lower value corresponds to a better approximation set, independently of the
#' type of problem (minimization, maximization or mixed). However, the meaning
#' of the value is different for each objective type. For example, imagine that
#' objective 1 is to be minimized and objective 2 is to be maximized, and the
#' multiplicative epsilon computed here for \eqn{epsilon(A,R) = 3}. This means
#' that \eqn{A} needs to be multiplied by 1/3 for all \eqn{a_1} values and by 3
#' for all \eqn{a_2} values in order to weakly dominate \eqn{R}.
#'
#' The multiplicative variant can be computed as \eqn{\exp(epsilon_{+}(\log(A),
#' \log(R)))}, which makes clear that the computation of the multiplicative
#' version for zero or negative values doesn't make sense. See the examples
#' below.
#'
#' The current implementation uses the naive algorithm that requires
#' \eqn{O(m \cdot |A| \cdot |R|)}, where \eqn{m} is the number of objectives
#' (dimension of vectors).
#'
#' @references
#'
#' \insertRef{ZitThiLauFon2003:tec}{moocore}
#'
NULL
#> NULL

#' @rdname epsilon
#' @concept metrics
#' @doctest
#' # Fig 6 from Zitzler et al. (2003).
#' A1 <- matrix(c(9,2,8,4,7,5,5,6,4,7), ncol=2, byrow=TRUE)
#' A2 <- matrix(c(8,4,7,5,5,6,4,7), ncol=2, byrow=TRUE)
#' A3 <- matrix(c(10,4,9,5,8,6,7,7,6,8), ncol=2, byrow=TRUE)
#' @omit
#' if (requireNamespace("graphics", quietly = TRUE)) {
#'    plot(A1, xlab=expression(f[1]), ylab=expression(f[2]),
#'         panel.first=grid(nx=NULL), pch=4, cex=1.5, xlim = c(0,10), ylim=c(0,8))
#'    points(A2, pch=0, cex=1.5)
#'    points(A3, pch=1, cex=1.5)
#'    legend("bottomleft", legend=c("A1", "A2", "A3"), pch=c(4,0,1),
#'           pt.bg="gray", bg="white", bty = "n", pt.cex=1.5, cex=1.2)
#' }
#' @expect equal(0.9)
#' epsilon_mult(A1, A3) # A1 epsilon-dominates A3 => e = 9/10 < 1
#' @expect equal(1)
#' epsilon_mult(A1, A2) # A1 weakly dominates A2 => e = 1
#' @expect equal(2)
#' epsilon_mult(A2, A1) # A2 is epsilon-dominated by A1 => e = 2 > 1
#' # Equivalence between additive and multiplicative
#' @expect equal(2)
#' exp(epsilon_additive(log(A2), log(A1)))
#'
#' # A more realistic example
#' extdata_path <- system.file(package="moocore","extdata")
#' path.A1 <- file.path(extdata_path, "ALG_1_dat.xz")
#' path.A2 <- file.path(extdata_path, "ALG_2_dat.xz")
#' A1 <- read_datasets(path.A1)[,1:2]
#' A2 <- read_datasets(path.A2)[,1:2]
#' ref <- filter_dominated(rbind(A1, A2))
#' @expect equal(199090640)
#' epsilon_additive(A1, ref)
#' @expect equal(132492066)
#' epsilon_additive(A2, ref)
#' # Multiplicative version of epsilon metric
#' ref <- filter_dominated(rbind(A1, A2))
#' @expect equal(1.05401476)
#' epsilon_mult(A1, ref)
#' @expect equal(1.023755)
#' epsilon_mult(A2, ref)
#' @export
epsilon_additive <- function(x, reference, maximise = FALSE)
  unary_common(data = x, reference = reference, maximise = maximise,
    fn = function(DATA, REFERENCE, MAXIMISE) .Call(epsilon_add_C, DATA, REFERENCE, MAXIMISE))

#' @rdname epsilon
#' @concept metrics
#' @export
epsilon_mult <- function(x, reference, maximise = FALSE)
  unary_common(data = x, reference = reference, maximise = maximise,
    fn = function(DATA, REFERENCE, MAXIMISE) .Call(epsilon_mul_C, DATA, REFERENCE, MAXIMISE))

unary_common <- function(data, reference, maximise, fn)
{
  data <- as_double_matrix(data)
  nobjs <- ncol(data)
  if (is.null(reference)) stop("reference cannot be NULL")
  reference <- as_double_matrix(reference)
  if (ncol(reference) != nobjs)
    stop("data and reference must have the same number of columns")
  if (nobjs < 2L)
    stop("the input must have a least 2 columns")
  check_dimension_max(nobjs, .libmoocore_constants[["MOOCORE_DIMENSION_MAX"]])
  fn(t(data), t(reference), as.logical(rep_len(maximise, nobjs)))
}

test_that("epsilon", {
  ref = matrix(c(10, 1, 6, 1, 2, 2, 1, 6, 1, 10), ncol = 2L, byrow = TRUE)
  A = matrix(c(4, 2, 3, 3, 2, 4), ncol = 2L, byrow = TRUE)
  expect_equal(epsilon_additive(A, ref), 1.0)
  expect_equal(epsilon_mult(A, ref), 2.0)
  expect_equal(epsilon_mult(A, ref, maximise=TRUE), 2.5)
  expect_equal(epsilon_additive(A, ref, maximise=TRUE), 6.0)
})

# Generated by doctest: do not edit by hand
# Please edit file in R/epsilon.R

test_that("Doctest: epsilon_additive", {
  # Created from @doctest for `epsilon_additive`
  # Source file: R/epsilon.R
  A1 <- matrix(c(9, 2, 8, 4, 7, 5, 5, 6, 4, 7), ncol = 2, byrow = TRUE)
  A2 <- matrix(c(8, 4, 7, 5, 5, 6, 4, 7), ncol = 2, byrow = TRUE)
  A3 <- matrix(c(10, 4, 9, 5, 8, 6, 7, 7, 6, 8), ncol = 2, byrow = TRUE)
  expect_equal(epsilon_mult(A1, A3), 0.9)
  expect_equal(epsilon_mult(A1, A2), 1)
  expect_equal(epsilon_mult(A2, A1), 2)
  expect_equal(exp(epsilon_additive(log(A2), log(A1))), 2)
  extdata_path <- system.file(package = "moocore", "extdata")
  path.A1 <- file.path(extdata_path, "ALG_1_dat.xz")
  path.A2 <- file.path(extdata_path, "ALG_2_dat.xz")
  A1 <- read_datasets(path.A1)[, 1:2]
  A2 <- read_datasets(path.A2)[, 1:2]
  ref <- filter_dominated(rbind(A1, A2))
  expect_equal(epsilon_additive(A1, ref), 199090640)
  expect_equal(epsilon_additive(A2, ref), 132492066)
  ref <- filter_dominated(rbind(A1, A2))
  expect_equal(epsilon_mult(A1, ref), 1.05401476)
  expect_equal(epsilon_mult(A2, ref), 1.023755)
})

#' Identify largest EAF differences
#'
#' Given a list of datasets, return the indexes of the pair with the largest
#' EAF differences according to the method proposed by \citet{DiaLop2020ejor}.
#'
#' @inheritParams eafdiff
#' @param x `list()`\cr A list of matrices or data frames with at least 3 columns (last column indicates the set).
#'
#' @inheritParams hypervolume
#'
#' @param ideal `numeric()`\cr Ideal point as a vector of numerical values.  If
#'   `NULL`, it is calculated as minimum (or maximum if maximising that
#'   objective) of each objective in the input data.
#'
#' @return  `list()`\cr A list with two components `pair` and `value`.
#'
#'@examples
#' # FIXME: This example is too large, we need a smaller one.
#' data(tpls50x20_1_MWT)
#' nadir <- apply(tpls50x20_1_MWT[,2:3], 2L, max)
#' x <- largest_eafdiff(split.data.frame(tpls50x20_1_MWT[,2:4], tpls50x20_1_MWT[, 1L]),
#'                      reference = nadir)
#' str(x)
#'
#'@references
#' \insertAllCited{}
#'
#'@concept eaf
#'@export
largest_eafdiff <- function(x, maximise = FALSE, intervals = 5L, reference,
                            ideal = NULL)
{
  # FIXME: Check the input data makes sense
  nobjs <- 2L
  maximise <- as.logical(rep_len(maximise, nobjs))
  if (nobjs != 2L) stop("Only 2 objectives currently supported")

  n <- length(x)
  stopifnot(n > 1L)
  if (is.null(ideal)) {
    r1 <- range_df_list(x, 1L)
    r2 <- range_df_list(x, 2L)
    upper <- c(r1[2L], r2[2L])
    lower <- c(r1[1L], r2[1L])
    ideal <- ifelse(maximise, upper, lower)
  }
  # Convert to a 1-row matrix
  if (is.null(dim(ideal))) dim(ideal) <- c(1L,nobjs)

  best_pair <- NULL
  best_value <- 0
  for (a in seq_len(n-1)) {
    for (b in (a+1):n) {
      DIFF <- eafdiff(x[[a]], x[[b]], intervals = intervals,
                      maximise = maximise, rectangles = TRUE)
      # Set color to 1
      a_rectangles <- DIFF[ DIFF[, 5] >= 1L, , drop = FALSE]
      a_rectangles[, ncol(a_rectangles)] <- 1

      a_value <- whv_rect(ideal, a_rectangles, reference, maximise)

      b_rectangles <- DIFF[ DIFF[, 5] <= -1L, , drop = FALSE]
      b_rectangles[, ncol(b_rectangles)] <- 1
      b_value <- whv_rect(ideal, b_rectangles, reference, maximise)
      value <- min(a_value, b_value)
      if (value > best_value) {
        best_value <- value
        best_pair <- c(a, b)
      }
    }
  }
  list(pair=best_pair, value = best_value)
}

#' Interactively choose according to empirical attainment function differences
#'
#' @param x `matrix()`\cr Matrix of rectangles representing EAF differences
#'   returned by [eafdiff()] with `rectangles=TRUE`.
#'
#' @param left `logical(1)`\cr With `left=TRUE` return the rectangles with
#'   positive differences, otherwise return those with negative differences but
#'   differences are converted to positive.
#'
#' @return `matrix()` where the first 4 columns give the coordinates of two
#'   corners of each rectangle and the last column. In both cases, the last
#'   column gives the positive differences in favor of the chosen side.
#'
#' @examples
#' \donttest{
#' extdata_dir <- system.file(package="moocore", "extdata")
#' A1 <- read_datasets(file.path(extdata_dir, "wrots_l100w10_dat"))
#' A2 <- read_datasets(file.path(extdata_dir, "wrots_l10w100_dat"))
#' # Choose A1
#' rectangles <- eafdiff(A1, A2, intervals = 5, rectangles = TRUE)
#' rectangles <- choose_eafdiff(rectangles, left = TRUE)
#' reference <- c(max(A1[, 1], A2[, 1]), max(A1[, 2], A2[, 2]))
#' x <- split.data.frame(A1[,1:2], A1[,3])
#' hv_A1 <- sapply(split.data.frame(A1[, 1:2], A1[, 3]),
#'                  hypervolume, reference=reference)
#' hv_A2 <- sapply(split.data.frame(A2[, 1:2], A2[, 3]),
#'                  hypervolume, reference=reference)
#' print(fivenum(hv_A1))
#' print(fivenum(hv_A2))
#' whv_A1 <- sapply(split.data.frame(A1[, 1:2], A1[, 3]),
#'                  whv_rect, rectangles=rectangles, reference=reference)
#' whv_A2 <- sapply(split.data.frame(A2[, 1:2], A2[, 3]),
#'                  whv_rect, rectangles=rectangles, reference=reference)
#' print(fivenum(whv_A1))
#' print(fivenum(whv_A2))
#' }
#'
#'@concept eaf
#'@export
choose_eafdiff <- function(x, left = stop("'left' must be either TRUE or FALSE"))
{
  if (left) return (x[ x[, ncol(x)] > 0L, , drop = FALSE])
  x <- x[ x[, ncol(x)] < 0L, , drop = FALSE]
  # We always return positive colors.
  x[, ncol(x)] <- abs(x[, ncol(x)])
  x
}

#' Generate a random set of mutually nondominated points
#'
#' @description
#' Generate a random set of `n` mutually nondominated points of dimension `d`
#' with the shape defined by `method`.
#'
#' When `integer = FALSE` (the default), the points are generated within the
#' hypercube \eqn{(0,1)^d} and can be scaled to another range using
#' [normalise()].  Otherwise, points are scaled to a non-negative integer range
#' that keeps the points mutually nondominated.
#'
#' See the visualisations in the vignette \HTMLVignette{generate}{Sampling Random Nondominated Sets}.
#'
#' @param n `integer(1)`\cr Number of rows in the output.
#' @param d `integer(1)`\cr Number of columns in the output.
#' @param method `character(1)`\cr Method used to generate the random nondominated set.
#'   See **Details** below for more information.
#' @param seed `integer(1)`\cr Integer seed for random number generation. If `NULL`,
#'   a random seed is generated.
#' @param integer `logical(1)`\cr If `TRUE`, return integer-valued points.
#'
#' @return A numeric matrix of size `n × d` containing nondominated points.
#'
#' @details
#' The available methods are:
#'
#' \describe{
#'   \item{`'simplex'`, `'linear'`, or `'L'`}{
#'
#'    Uniformly samples points on the standard simplex.  This shape of
#'    nondominated set is also called `'linear'` in the literature
#'    \citep{LacKlaFon2017box}.
#'
#'    The standard \eqn{(d-1)}-simplex is defined by \eqn{\{x \in \mathbb{R}_+^d :
#'    \sum_i x_i = 1\}}.  Each point \eqn{\vec{z} \in (0,1)^d \subset
#'    \mathbb{R}^d} is generated by sampling \eqn{d} independent and
#'    identically distributed values \eqn{(x_1,x_2, \dots, x_d)} from the
#'    exponential distribution, then dividing each value by the L1-norm of the
#'    vector, \eqn{z_i = x_i / \sum_{i=1}^d x_i} \citep{RubMel1998simulation}.
#'    Values sampled from the exponential distribution are guaranteed to be
#'    positive.  Sampling from either the standard normal distribution
#'    \citep{GueFonPaq2021hv} or the uniform distribution
#'    \citep{LacKlaFon2017box} does not produce a uniform distribution when
#'    projected onto the simplex.
#'
#'   }
#'   \item{`'concave-sphere'`, `'sphere'`, or `'C'`}{
#'
#'    Uniformly samples points on the positive orthant of the hyper-sphere,
#'    which is concave when all objectives are minimised.
#'
#'    Each point \eqn{\vec{z} \in (0,1)^d \subset \mathbb{R}^d} is generated by
#'    sampling \eqn{d} independent and identically distributed values
#'    \eqn{\vec{x}=(x_1,x_2, \dots, x_d)} from the standard normal
#'    distribution, then dividing each value by the l2-norm of the vector,
#'    \eqn{z_i = \frac{|x_i|}{\|\vec{x}\|_2}} \citep{Muller1959sphere}. The
#'    absolute value in the numerator ensures that points are sampled on the
#'    positive orthant of the hyper-sphere.  Sampling from the uniform
#'    distribution \citep{LacKlaFon2017box} would not result in a uniform
#'    sampling when projected onto the surface of the hyper-sphere.
#'
#'   }
#'   \item{`'convex-sphere'` or `'X'`}{
#'
#'    Equivalent to `1 - generate_ndset(..., method='concave-sphere')`, which
#'    is convex for minimisation.  This shape has also been called *inverted
#'    convex* \citep{IshHeSha2019regular}. This sampling is uniform.
#'
#'    It corresponds to translating points from the negative orthant of the
#'    hyper-sphere to the positive orthant. Thus, the sampling remains uniform.
#'
#'   }
#'   \item{`'convex-simplex'`}{
#'
#'    Equivalent to `generate_ndset(..., method='simplex')^2`, which is convex
#'    for minimisation. Such a set cannot be obtained by any affine
#'    transformation of a subset of the hyper-sphere.  This sampling is *not*
#'    uniform.
#'
#'    The corresponding surface is equivalent to a simplex curved towards the
#'    origin. The generated points \eqn{\vec{z} \in (0,1)^d \subset
#'    \mathbb{R}^d} satisfy \eqn{\sum_{i=1}^d \sqrt{z_i} = 1}.  Although the
#'    sampling on the simplex is uniform, the transformed points are not.
#'
#'   }
#'   \item{`'concave-simplex'`}{
#'
#'    Equivalent to `1 - generate_ndset(..., method='convex-simplex')`, which
#'    is concave for minimisation.  This shape has also been called *inverted
#'    concave* \citep{IshHeSha2019regular}.  This sampling is *not* uniform
#'    because `method='convex-simplex'` is also not uniform.
#'
#'   }
#'   \item{`'inverted-simplex'` or `'inverted-linear'`}{
#'
#'    Equivalent to `1 - generate_ndset(..., method='simplex')`. This sampling
#'    is uniform.
#'
#'   }
#' }
#'
#' Methods `'inverted-simplex'`, `'concave-simplex'` and `'convex-sphere'` are
#' translations of `'simplex'`, `'convex-simplex'` and `'concave'`, respectively.
#' These translations have been called *inverted* shapes in the literature and
#' have different properties than their *regular* counterparts
#' \citep{IshHeSha2019regular}.
#'
#' @references
#'
#' \insertAllCited{}
#'
#' @examples
#' generate_ndset(5, 3, "simplex", seed = 42)
#' generate_ndset(5, 3, "simplex", seed = 42, integer = TRUE)
#' generate_ndset(4, 2, "sphere", seed = 123)
#' generate_ndset(3, 5, "convex-sphere", seed = 123)
#' generate_ndset(4, 4, "convex-simplex", seed = 123)
#'
#' @export
generate_ndset <- function(n, d, method, seed = NULL, integer = FALSE)
{
  seed <- if (is.null(seed)) get_seed() else as_integer(seed)
  withr::local_seed(seed)

  sample_simplex <- function() {
    x <- rexp(n * d)
    dim(x) <- c(n, d)
    x <- x / rowSums(x)
    x
  }

  sample_sphere <- function() {
    x <- abs(rnorm(n * d))
    dim(x) <- c(n, d)
    x <- x / sqrt(rowSums(x * x))
    x
  }

  sample_convex_sphere <- function() 1. - sample_sphere()
  sample_convex_simplex <- function() sample_simplex()^2
  sample_inverted_simplex <- function() 1. - sample_simplex()
  sample_concave_simplex <- function() 1. - sample_convex_simplex()

  sample_fun <- (
    if (method %in% c("simplex", "linear", "L")) sample_simplex
    else if (method %in% c("concave-sphere", "sphere", "C")) sample_sphere
    else if (method %in% c("convex-sphere", "X")) sample_convex_sphere
    else if (method == "convex-simplex") sample_convex_simplex
    else if (method %in% c("inverted-simplex", "inverted-linear")) sample_inverted_simplex
    else if (method == "concave-simplex") sample_concave_simplex
    else stop("Unknown method =", method)
  )

  repeat {
    x <- sample_fun()
    # Due to rounding or bad luck, we may actually have dominated points.
    if (!any_dominated(x)) {
      if (!integer) return(x)
      break
    }
  }

  x <- x * 64
  repeat {
    y <- trunc(x)
    if (!any_dominated(y)) return(y)
    x <- x * 2
  }
}

test_that("generate_ndset", {
  n <- 10L
  one <- replicate(n,1)
  for (dim in seq(5,10)) {
    points <- generate_ndset(n, dim, "simplex")
    expect_equal(rowSums(points), one)
    points <- generate_ndset(n, dim, "concave-sphere")
    expect_equal(rowSums(points**2), one)
    points <- generate_ndset(n, dim, "convex-simplex")
    expect_equal(rowSums(sqrt(points)), one)
  }
})

#' Inverted Generational Distance (IGD and IGD+) and Averaged Hausdorff Distance
#'
#' Functions to compute the inverted generational distance (IGD and IGD+) and
#' the averaged Hausdorff distance between nondominated sets of points.
#'
#' @inherit epsilon params return
#'
#' @author Manuel \enc{López-Ibáñez}{Lopez-Ibanez}
#'
#' @details
#'
#' The generational distance (GD) of a set \eqn{A} is defined as the distance
#' between each point \eqn{a \in A} and the closest point \eqn{r} in a
#' reference set \eqn{R}, averaged over the size of \eqn{A}. Formally,
#'
#' \deqn{GD_p(A,R) = \left(\frac{1}{|A|}\sum_{a\in A}\min_{r\in R} d(a,r)^p\right)^{\frac{1}{p}} }{GD(A,R) = (1/|A|) * ( sum_{a in A} min_{r in R} d(a,r)^p )^(1/p)}
#' where the distance in our implementation is the Euclidean distance:
#' \deqn{d(a,r) = \sqrt{\sum_{k=1}^m (a_k - r_k)^2} }{d(a,r) = sqrt( sum_{k=1}^m (a_k - r_k)^2)}
#'
#' The inverted generational distance (IGD) is calculated as \eqn{IGD_p(A,R) = GD_p(R,A)}.
#'
#' The modified inverted generational distanced (IGD+) was proposed by
#' \citet{IshMasTanNoj2015igd} to ensure that IGD+ is weakly Pareto compliant,
#' similarly to [epsilon_additive()] or [epsilon_mult()]. Formally,
#'
#' \deqn{IGD^+(A,R) = \left(\frac{1}{|R|}\sum_{r\in R}\min_{a\in A} d^+(r,a)\right)}{IGD+(A,R) = (1/|R|) * ( sum_{r in R} min_{a in A} d^+(r,a))}
#'
#' Assuming minimization of all objectives, IGD+ uses the following modified
#' distance:
#'
#' \deqn{d^+(r,a) = \sqrt{\sum_{k=1}^m (\max\{a_k - r_k, 0\})^2}}{d^+(r,a) = sqrt(sum_{k=1}^m (max {a_k - r_k, 0 })^2)}
#'
#' The average Hausdorff distance (\eqn{\Delta_p}) was proposed by
#' \citet{SchEsqLarCoe2012tec} and it is calculated as:
#'
#' \deqn{\Delta_p(A,R) = \max\{ IGD_p(A,R), IGD_p(R,A) \}}
#'
#' IGDX \citep{ZhoZhaJin2009igdx} is the application of IGD to decision vectors
#' instead of objective vectors to measure closeness and diversity in decision
#' space. One can use the functions `igd()` or `igd_plus()` (recommended)
#' directly, just passing the decision vectors as `data`.
#'
#' There are different formulations of the GD and IGD metrics in the literature
#' that differ on the value of \eqn{p}, on the distance metric used and on
#' whether the term \eqn{|A|^{-1}} is inside (as above) or outside the exponent
#' \eqn{1/p}.  GD was first proposed by \citet{VelLam1998gp} with \eqn{p=2} and
#' the term \eqn{|A|^{-1}} outside the exponent. IGD seems to have been
#' mentioned first by \citet{CoeSie2004igd}, however, some people also used the
#' name D-metric for the same concept with \eqn{p=1} and later papers have
#' often used IGD/GD with \eqn{p=1}. \citet{SchEsqLarCoe2012tec} proposed to
#' place the term \eqn{|A|^{-1}} inside the exponent, as in the formulation
#' shown above.  This has a significant effect for GD and less so for IGD given
#' a constant reference set. IGD+ also follows this formulation.  We refer to
#' \citet{IshMasTanNoj2015igd} and \citet{BezLopStu2017emo} for a more detailed
#' historical perspective and a comparison of the various variants.
#'
#' Following \citet{IshMasTanNoj2015igd}, we always use \eqn{p=1} in our
#' implementation of IGD and IGD+ because (1) it is the setting most used in
#' recent works; (2) it makes irrelevant whether the term \eqn{|A|^{-1}} is
#' inside or outside the exponent \eqn{1/p}; and (3) the meaning of IGD becomes
#' the average Euclidean distance from each reference point to its nearest
#' objective vector. It is also slightly faster to compute.
#'
#' GD should never be used directly to compare the quality of approximations to
#' a Pareto front, because it is not weakly Pareto-compliant and often
#' contradicts Pareto optimality.
#'
#' IGD is still popular due to historical reasons, but we strongly recommend
#' IGD+ instead of IGD, because IGD contradicts Pareto optimality in some cases
#' (see examples below) whereas IGD+ is weakly Pareto-compliant.
#'
#' The average Hausdorff distance \eqn{\Delta_p(A,R)} is also not weakly
#' Pareto-compliant, as shown in the examples below.
#'
#' @references
#'
#' \insertAllCited{}
#'
#' @doctest
#  This is already tested in test-igd.R
#' @omit
#' # Example 4 from Ishibuchi et al. (2015)
#' ref <- matrix(c(10,0,6,1,2,2,1,6,0,10), ncol=2, byrow=TRUE)
#' A <- matrix(c(4,2,3,3,2,4), ncol=2, byrow=TRUE)
#' B <- matrix(c(8,2,4,4,2,8), ncol=2, byrow=TRUE)
#' if (requireNamespace("graphics", quietly = TRUE)) {
#'    plot(ref, xlab=expression(f[1]), ylab=expression(f[2]),
#'         panel.first=grid(nx=NULL), pch=23, bg="gray", cex=1.5)
#'    points(A, pch=1, cex=1.5)
#'    points(B, pch=19, cex=1.5)
#'    legend("topright", legend=c("Reference", "A", "B"), pch=c(23,1,19),
#'           pt.bg="gray", bg="white", bty = "n", pt.cex=1.5, cex=1.2)
#' }
#' cat("A is better than B in terms of Pareto optimality,\n however, IGD(A)=",
#'     igd(A, ref), "> IGD(B)=", igd(B, ref),
#'     "and AvgHausdorff(A)=", avg_hausdorff_dist(A, ref),
#'     "> AvgHausdorff(B)=", avg_hausdorff_dist(B, ref),
#'     ", which both contradict Pareto optimality.\nBy contrast, IGD+(A)=",
#'     igd_plus(A, ref), "< IGD+(B)=", igd_plus(B, ref), ", which is correct.\n")
#' @resume
#' # A less trivial example.
#' extdata_path <- system.file(package="moocore","extdata")
#' path.A1 <- file.path(extdata_path, "ALG_1_dat.xz")
#' path.A2 <- file.path(extdata_path, "ALG_2_dat.xz")
#' A1 <- read_datasets(path.A1)[,1:2]
#' A2 <- read_datasets(path.A2)[,1:2]
#' ref <- filter_dominated(rbind(A1, A2))
#' @expect equal(91888189)
#' igd(A1, ref)
#' @expect equal(11351992)
#' igd(A2, ref)
#'
#' # IGD+ (Pareto compliant)
#' @expect equal(82695357)
#' igd_plus(A1, ref)
#' @expect equal(10698269.3)
#' igd_plus(A2, ref)
#'
#' # Average Haussdorff distance
#' @expect equal(268547627)
#' avg_hausdorff_dist(A1, ref)
#' @expect equal(352613092)
#' avg_hausdorff_dist(A2, ref)
#' @rdname igd
#' @concept metrics
#' @aliases IGDX
#' @export
igd <- function(x, reference, maximise = FALSE) {
  unary_common(data = x, reference = reference, maximise = maximise,
    fn = function(DATA, REFERENCE, MAXIMISE) .Call(igd_C, DATA, REFERENCE, MAXIMISE))
}
#' @rdname igd
#' @concept metrics
#' @export
igd_plus <- function(x, reference, maximise = FALSE)
  unary_common(data = x, reference = reference, maximise = maximise,
    fn = function(DATA, REFERENCE, MAXIMISE) .Call(igd_plus_C, DATA, REFERENCE, MAXIMISE))

#' @rdname igd
#' @param p `integer(1)`\cr Hausdorff distance parameter (default: `1L`).
#' @concept metrics
#' @export
avg_hausdorff_dist <- function(x, reference, maximise = FALSE, p = 1L)
  unary_common(data = x, reference = reference, maximise = maximise,
    fn = function(DATA, REFERENCE, MAXIMISE) .Call(avg_hausdorff_dist_C, DATA, REFERENCE, MAXIMISE, as.integer(p)))

# Example 4 from Ishibuchi et al. (2015)
ref = matrix(c(10, 0, 6, 1, 2, 2, 1, 6, 0, 10), ncol = 2L, byrow = TRUE)
A = matrix(c(4, 2, 3, 3, 2, 4), ncol = 2L, byrow = TRUE)
B = matrix(c(8, 2, 4, 4, 2, 8), ncol = 2L, byrow = TRUE)

test_that("igd", {
  expect_equal(igd(A, ref), 3.707092031609239)
  expect_equal(igd(B, ref), 2.59148346584763)
})

test_that("igd plus", {
  expect_equal(igd_plus(A, ref), 1.482842712474619)
  expect_equal(igd_plus(B, ref), 2.260112615949154)

})
test_that("avg_hausdorff_dist", {
  expect_equal(avg_hausdorff_dist(A, ref), 3.707092031609239)
  expect_equal(avg_hausdorff_dist(B, ref), 2.59148346584763)
})

# Generated by doctest: do not edit by hand
# Please edit file in R/igd.R

test_that("Doctest: igd", {
  # Created from @doctest for `igd`
  # Source file: R/igd.R
  extdata_path <- system.file(package = "moocore", "extdata")
  path.A1 <- file.path(extdata_path, "ALG_1_dat.xz")
  path.A2 <- file.path(extdata_path, "ALG_2_dat.xz")
  A1 <- read_datasets(path.A1)[, 1:2]
  A2 <- read_datasets(path.A2)[, 1:2]
  ref <- filter_dominated(rbind(A1, A2))
  expect_equal(igd(A1, ref), 91888189)
  expect_equal(igd(A2, ref), 11351992)
  expect_equal(igd_plus(A1, ref), 82695357)
  expect_equal(igd_plus(A2, ref), 10698269.3)
  expect_equal(avg_hausdorff_dist(A1, ref), 268547627)
  expect_equal(avg_hausdorff_dist(A2, ref), 352613092)
})

#' Identify and remove dominated points according to Pareto optimality
#'
#' Identify nondominated points with [is_nondominated()] and remove dominated
#' ones with [filter_dominated()].
#'
#' @rdname nondominated
#'
#' @param x `matrix()`|`data.frame()`\cr Matrix or data frame of numerical
#'   values, where each row gives the coordinates of a point.
#'
#' @param maximise `logical()`\cr Whether the objectives must be maximised
#'   instead of minimised. Either a single logical value that applies to all
#'   objectives or a vector of logical values, with one value per objective.
#'
#' @param keep_weakly `logical(1)`\cr If `FALSE`, return `FALSE` for any
#'   duplicates of nondominated points, except the last one.
#'
#' @return [is_nondominated()] returns a logical vector of the same length
#'   as the number of rows of `data`, where `TRUE` means that the
#'   point is not dominated by any other point.
#'
#' @author Manuel \enc{López-Ibáñez}{Lopez-Ibanez}
#'
#' @details
#'
#' Given \eqn{n} points of dimension \eqn{m}, the current implementation always
#' uses the best-known \eqn{O(n \log n)} dimension-sweep algorithm
#' \citep{KunLucPre1975jacm} for \eqn{m \leq 3}. For \eqn{m \geq 4}, functions
#' `is_nondominated()` and `filter_dominated()` use the best-known \eqn{O(n
#' \log^{m-2} n)} algorithm \citep{KunLucPre1975jacm} when \eqn{n > `r moocore:::.libmoocore_constants[["KUNG_SMALL_THRESHOLD"]]`}, and
#' the naive \eqn{O(m n^2)} algorithm otherwise.  Function `any_dominated()`
#' always uses the naive algorithm for \eqn{m \geq 4}.
#'
#' @doctest
#' S = matrix(c(1,1,0,1,1,0,1,0), ncol = 2, byrow = TRUE)
#' @expect equal(c(FALSE,TRUE,TRUE,FALSE))
#' is_nondominated(S)
#'
#' @expect equal(c(TRUE,FALSE,FALSE,FALSE))
#' is_nondominated(S, maximise = TRUE)
#'
#' @expect equal(matrix(c(0,1,1,0), ncol = 2, byrow = TRUE))
#' filter_dominated(S)
#'
#' @expect equal(matrix(c(0,1,1,0,1,0), ncol = 2, byrow = TRUE))
#' filter_dominated(S, keep_weakly = TRUE)
#'
#' @expect equal(TRUE)
#' any_dominated(S)
#'
#' @expect equal(TRUE)
#' any_dominated(S, keep_weakly = TRUE)
#'
#' @expect equal(FALSE)
#' any_dominated(filter_dominated(S))
#'
#' @omit
#' path_A1 <- file.path(system.file(package="moocore"),"extdata","ALG_1_dat.xz")
#' set <- read_datasets(path_A1)[,1:2]
#' is_nondom <- is_nondominated(set)
#' cat("There are ", sum(is_nondom), " nondominated points\n")
#'
#' if (requireNamespace("graphics", quietly = TRUE)) {
#'    plot(set, col = "blue", type = "p", pch = 20)
#'    ndset <- filter_dominated(set)
#'    points(ndset[order(ndset[,1]),], col = "red", pch = 21)
#' }
#'
#' @references
#'
#' \insertAllCited{}
#'
#' @seealso [pareto_rank()]
#' @concept dominance
#' @export
is_nondominated <- function(x, maximise = FALSE, keep_weakly = FALSE)
{
  x <- as_double_matrix_1(x)
  nobjs <- ncol(x)
  keep_weakly <- as.logical(keep_weakly)
  maximise <- as.logical(maximise)
  if (nobjs == 1L) { # Handle single-objective
    if (keep_weakly) {
      best <- if (maximise) max(x) else min(x)
      return(as.vector(x == best))
    } else {
      nondom <- logical(nrow(x))
      nondom[if (maximise) which.max(x) else which.min(x)] <- TRUE
      return(nondom)
    }
  }
  check_dimension_max(nobjs, .libmoocore_constants[["MOOCORE_DIMENSION_MAX"]])
  .Call(is_nondominated_C,
    x,
    keep_weakly,
    rep_len(maximise, nobjs))
}

#' @rdname nondominated
#' @concept dominance
#' @return [filter_dominated()] returns a matrix or data.frame with only mutually nondominated points.
#' @export
filter_dominated <- function(x, maximise = FALSE, keep_weakly = FALSE)
  x[is_nondominated(x, maximise = maximise, keep_weakly = keep_weakly), , drop = FALSE]

#' @rdname nondominated
#' @concept dominance
#' @description  [any_dominated()] quickly detects if a set contains any dominated point.
#' @return [any_dominated()] returns `TRUE` if `x` contains any (weakly-)dominated points, `FALSE` otherwise.
#' @export
any_dominated <- function(x, maximise = FALSE, keep_weakly = FALSE)
{
  x <- as_double_matrix_1(x)
  if (keep_weakly) # FIXME: Do this in C.
    x <- x[!duplicated(x), , drop = FALSE]
  nobjs <- ncol(x)
  check_dimension_max(nobjs, .libmoocore_constants[["MOOCORE_DIMENSION_MAX"]])
  .Call(any_dominated_C,
    x,
    rep_len(as.logical(maximise), nobjs))
}

#' Rank points according to Pareto-optimality (nondominated sorting).
#'
#' [pareto_rank()] is meant to be used like `rank()`, but it assigns ranks
#' according to Pareto dominance, where rank 1 indicates those solutions not
#' dominated by any other solution in the input set.  Duplicated points are
#' assigned the same rank.  The resulting ranking can be used to partition
#' points into a list of matrices, each matrix representing a nondominated
#' front \citep{Deb02nsga2} (see examples below).
#'
#' @inherit is_nondominated params
#'
#' @return An integer vector of the same length as the number of rows of the
#'   input `x`, where each value gives the rank of each point (lower is
#'   better).
#'
#' @details
#'
#'   Given a finite set of points \eqn{X \subset \mathbb{R}^m},
#'   the rank of a point \eqn{x \in X} is defined as:
#'
#'   \deqn{\operatorname{rank}(x) = r \iff x \in F^c_{r} \land \nexists y \in F^c_{r}, y \prec x}
#'
#'   where \eqn{y \prec x} means that \eqn{y} dominates \eqn{x} according to
#'   Pareto optimality, \eqn{F^c_r = X \setminus \bigcup_{i=1}^{r-1} F_i} and
#'   \eqn{F_r = \{x \in X \land \operatorname{rank}(x) = r\}}.  The sets
#'   \eqn{F_c}, with \eqn{c=1,\dots,k}, partition \eqn{X} into \eqn{k}
#'   *fronts*, that is, mutually nondominated subsets of \eqn{X}.
#'
#'   With \eqn{m=2}, i.e., `ncol(data)=2`, the code uses the best-known
#'   \eqn{O(n \log n)} algorithm by \citet{Jen03}.  When \eqn{m \geq 3}, it
#'   uses the naive algorithm that identifies one *front* at a time, which
#'   requires \eqn{O(n^2\log n)} for \eqn{m=3}, and \eqn{O(n^2 \log^{m-2} n)}
#'   for \eqn{m \geq 4}.
#'
#' @doctest
#'
#' three_fronts = matrix(c(1, 2, 3,
#'                         3, 1, 2,
#'                         2, 3, 1,
#'                         10, 20, 30,
#'                         30, 10, 20,
#'                         20, 30, 10,
#'                         100, 200, 300,
#'                         300, 100, 200,
#'                         200, 300, 100), ncol=3, byrow=TRUE)
#' @expect equal(c(1,1,1,2,2,2,3,3,3))
#' pareto_rank(three_fronts)
#'
#' split.data.frame(three_fronts, pareto_rank(three_fronts))
#'
#' @omit
#' path_A1 <- file.path(system.file(package="moocore"),"extdata","ALG_1_dat.xz")
#' set <- read_datasets(path_A1)[,1:2]
#' ranks <- pareto_rank(set)
#' str(ranks)
#' if (requireNamespace("graphics", quietly = TRUE)) {
#'    colors <- colorRampPalette(c("red","yellow","springgreen","royalblue"))(max(ranks))
#'    plot(set, col = colors[ranks], type = "p", pch = 20)
#' }
#'
#' @references
#'
#' \insertAllCited{}
#'
#' @seealso [is_nondominated()]
#' @concept dominance
#' @export
pareto_rank <- function(x, maximise = FALSE)
{
  x <- as_double_matrix_1(x)
  nobjs <- ncol(x)
  if (nobjs == 1L) { # Handle single-objective
    x <- as.vector(x)
    if (maximise)
      x <- -x
    return(match(x, sort(unique(x)))) # FIXME: Can we do this faster?
  }
  check_dimension_max(nobjs, .libmoocore_constants[["MOOCORE_DIMENSION_MAX"]])
  x <- transform_maximise(x, maximise)
  .Call(pareto_ranking_C, t(x))
}

source("helper-common.R")

test_that("pareto", {
  test_pareto_rank <- function(extdatafile, maximise = FALSE) {
    data <- read_extdata(extdatafile)
    # Drop set column
    data <- data[,-ncol(data)]
    ranks <- pareto_rank(data, maximise = maximise)
    data2 <- data
    for (r in min(ranks):max(ranks)) {
      # We have to keep_weakly because pareto_rank does the same.
      nondom <- is_nondominated(data2, maximise = maximise, keep_weakly = TRUE)
      expect_equal(data[ranks == r, , drop = FALSE], data2[nondom, , drop = FALSE])
      data2 <- data2[!nondom, , drop = FALSE]
    }
  }
  test_pareto_rank("ALG_2_dat.xz")
  test_pareto_rank("spherical-250-10-3d.txt")
})

test_that("pareto_rank() 1D", {
  x <- matrix(c(0.2, 0.1, 0.2, 0.5, 0.3), ncol = 1L)
  expect_equal(pareto_rank(x), c(2, 1, 2, 4, 3))
  expect_equal(pareto_rank(x, maximise=TRUE), c(3, 4, 3, 1, 2))
})

# Generated by doctest: do not edit by hand
# Please edit file in R/nondominated.R

test_that("Doctest: pareto_rank", {
  # Created from @doctest for `pareto_rank`
  # Source file: R/nondominated.R
  three_fronts = matrix(c(1, 2, 3, 3, 1, 2, 2, 3, 1, 10, 20, 30, 30, 10, 20, 20,
    30, 10, 100, 200, 300, 300, 100, 200, 200, 300, 100), ncol = 3, byrow = TRUE)
  expect_equal(pareto_rank(three_fronts), c(1, 1, 1, 2, 2, 2, 3, 3, 3))
  split.data.frame(three_fronts, pareto_rank(three_fronts))
})

# Generated by doctest: do not edit by hand
# Please edit file in R/nondominated.R

test_that("Doctest: is_nondominated", {
  # Created from @doctest for `is_nondominated`
  # Source file: R/nondominated.R
  S = matrix(c(1, 1, 0, 1, 1, 0, 1, 0), ncol = 2, byrow = TRUE)
  expect_equal(is_nondominated(S), c(FALSE, TRUE, TRUE, FALSE))
  expect_equal(is_nondominated(S, maximise = TRUE), c(TRUE, FALSE, FALSE, FALSE))
  expect_equal(filter_dominated(S), matrix(c(0, 1, 1, 0), ncol = 2, byrow = TRUE))
  expect_equal(filter_dominated(S, keep_weakly = TRUE), matrix(c(0, 1, 1, 0, 1, 0),
  ncol = 2, byrow = TRUE))
  expect_equal(any_dominated(S), TRUE)
  expect_equal(any_dominated(S, keep_weakly = TRUE), TRUE)
  expect_equal(any_dominated(filter_dominated(S)), FALSE)
})

test_that("bug 27", {
  x <- matrix(c(0.5,0.6,0.3,0.1,0.0,0.9,0.0), ncol=1L)
  expect_equal(is_nondominated(x),
      c(FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, FALSE))
  expect_equal(is_nondominated(x, keep_weakly=TRUE),
    c(FALSE, FALSE, FALSE, FALSE, TRUE, FALSE, TRUE))
  expect_equal(is_nondominated(x, maximise=TRUE),
    c(FALSE, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE))
  expect_equal(is_nondominated(x, keep_weakly=TRUE, maximise=TRUE),
    c(FALSE, FALSE, FALSE, FALSE, FALSE, TRUE, FALSE))
  expect_true(any_dominated(x))
  expect_false(any_dominated(matrix(rep(5,5), ncol=1L), keep_weakly=TRUE))

})

test_that("is_nondominated 4D", {
  set.seed(42)
  pts <- matrix(runif(100), ncol = 4L)  # 25 points in 4D

  fd <- filter_dominated(pts)
  expect_true(nrow(fd) >= 1L)
  nd <- is_nondominated(fd)
  expect_true(is.logical(nd))
  expect_equal(length(nd), nrow(fd))
  # The non-dominated points should not dominate each other
  expect_true(all(nd))
  expect_false(any_dominated(fd))
})

test_that("is_nondominated 5D", {
  set.seed(123)
  pts <- matrix(runif(50), ncol = 5)  # 10 points in 5D

  fd <- filter_dominated(pts)
  expect_true(nrow(fd) >= 1L)
  nd <- is_nondominated(fd)
  expect_true(is.logical(nd))
  expect_equal(length(nd), nrow(fd))
  # The non-dominated points should not dominate each other
  expect_true(all(nd))
  expect_false(any_dominated(fd))
})

#' Hypervolume metric
#'
#' Compute the hypervolume metric with respect to a given reference point
#' assuming minimization of all objectives.
#'
#'
#' @inherit epsilon params return
#'
#' @param reference `numeric()`\cr Reference point as a vector of numerical
#'   values.
#'
#' @author Manuel \enc{López-Ibáñez}{Lopez-Ibanez}
#'
#' @details
#'
#' The hypervolume of a set of multidimensional points \eqn{A \subset
#' \mathbb{R}^m}{A in R^m}, where \eqn{m} is the dimension of the points, with
#' respect to a reference point \eqn{\vec{r} \in \mathbb{R}^m}{r in R^m} is the
#' volume of the region dominated by the set and bounded by the reference point
#' \citep{ZitThi1998ppsn}.  Points in \eqn{A} that do not strictly dominate
#' \eqn{\vec{r}} do not contribute to the hypervolume value, thus, ideally, the
#' reference point must be strictly dominated by all points in the true Pareto
#' front.
#'
#' More precisely, the hypervolume is the [Lebesgue measure](https://en.wikipedia.org/wiki/Lebesgue_measure) of the union of
#' axis-aligned hyperrectangles
#' ([orthotopes](https://en.wikipedia.org/wiki/Hyperrectangle)), where each
#' hyperrectangle is defined by one point from \eqn{\vec{a} \in A} and the
#' reference point.  The union of axis-aligned hyperrectangles is also called
#' an _orthogonal polytope_.
#'
#' The hypervolume is compatible with Pareto-optimality
#' \citep{KnoCor2002cec,ZitThiLauFon2003:tec}, that is, \eqn{\nexists A,B
#' \subset \mathbb{R}^m}{it does not exist A,B subsets of R^m}, such that
#' \eqn{A} is better than \eqn{B} according to Pareto-optimality and
#' \eqn{\text{hyp}(A) \leq \text{hyp}(B)}{hyp(A) <= hyp(B)}. In other words, if
#' the hypervolume of a set is larger than the hypervolume of another, then we
#' know for sure than the latter set cannot be better than the former in terms
#' of Pareto-optimality. The hypervolume is also the only unary metric known
#' that is complete according to Pareto-optimality
#' \citep{ZitThiLauFon2003:tec}, that is, if \eqn{A} is better than \eqn{B}
#' according to Pareto-optimality, then \eqn{\text{hyp}(A) > \text{hyp}(B)}.
#'
#' Like most measures of unions of high-dimensional geometric objects,
#' computing the hypervolume is #P-hard \citep{BriFri2010approx}, which means
#' that the best possible algorithm takes an exponential time on the number of
#' objectives or points, in the worst-case.
#'
#' For 2D and 3D, the algorithms used
#' \citep{FonPaqLop06:hypervolume,BeuFonLopPaqVah09:tec} have \eqn{O(n \log n)}
#' complexity, where \eqn{n} is the number of input points. The 3D case uses
#' the \eqn{\text{HV3D}^{+}} algorithm \citep{GueFon2017hv4d}, which has the
#' sample complexity as the HV3D algorithm
#' \citep{FonPaqLop06:hypervolume,BeuFonLopPaqVah09:tec}, but it is faster in
#' practice.
#'
#' For 4D, the algorithm used is \eqn{\text{HV4D}^{+}} \citep{GueFon2017hv4d},
#' which has \eqn{O(n^2)} complexity.  Compared to the [original
#' implementation](https://github.com/apguerreiro/HVC/), this implementation
#' correctly handles weakly dominated points and has been further optimized for
#' speed.
#'
#' For 5D or higher and up to `r moocore:::.libmoocore_constants[["HV_INEX_MAX_ROWS"]]` points, the implementation uses the
#' inclusion-exclusion algorithm \citep{WuAza2001metrics}, which has \eqn{O(m
#' 2^{n})} time and \eqn{O(n\cdot m)} space complexity, but it is very fast for
#' such small sets.  For larger number of points, it uses a recursive algorithm
#' \citep{FonPaqLop06:hypervolume} that computes 4D contributions
#' \citep{GueFon2017hv4d} as the base case, resulting in a \eqn{O(n^{m-2})}
#' time complexity and \eqn{O(n)} space complexity in the worst-case.
#' Experimental results show that the pruning techniques used may reduce the
#' time complexity even further.  The original proposal
#' \citep{FonPaqLop06:hypervolume} had the HV3D algorithm as the base case,
#' giving a time complexity of \eqn{O(n^{m-2} \log n)}.  Andreia P. Guerreiro
#' enhanced the numerical stability of the algorithm by avoiding floating-point
#' comparisons of partial hypervolumes.
#'
#' The hypervolume of 1D inputs is defined as `max(0, ref - min(x))`.
#'
#' @references
#'
#' \insertAllCited{}
#'
#' @doctest
#'
#' dat = matrix(c(5, 5, 4, 6, 2, 7, 7, 4), ncol=2, byrow=TRUE)
#' @expect equal(38)
#' hypervolume(dat, ref=c(10, 10))
#' @expect equal(39)
#' hypervolume(dat, ref=0, maximise=TRUE)
#'
#' data(SPEA2minstoptimeRichmond)
#' # The second objective must be maximized
#' # We calculate the hypervolume of the union of all sets.
#' @expect equal(7911375.69)
#' hypervolume(SPEA2minstoptimeRichmond[, 1:2], reference = c(250, 0),
#'             maximise = c(FALSE, TRUE))
#' @export
#' @concept metrics
hypervolume <- function(x, reference, maximise = FALSE)
{
  x <- as_double_matrix_1(x)
  nobjs <- ncol(x)
  if (!is.numeric(reference))
    stop("a numerical reference vector must be provided")
  if (length(reference) == 1L) reference <- rep_len(reference, nobjs)

  if (any(maximise)) {
    x <- transform_maximise(x, maximise)
    if (all(maximise)) {
      reference <- -reference
    } else {
      reference[maximise] <- -reference[maximise]
    }
  }
  # hypervolume_C already handles nobjs==1.
  check_dimension_max(nobjs, .libmoocore_constants[["MOOCORE_HV_DIMENSION_MAX"]])
  .Call(hypervolume_C,
    t(x),
    as.double(reference))
}

#' Hypervolume contribution of a set of points
#'
#' Computes the hypervolume contribution of each point of a set of points with
#' respect to a given reference point. Duplicated and dominated points have
#' zero contribution.  By default, dominated points are ignored, that is, they
#' do not affect the contribution of other points.  See the Notes below for
#' more details.  For details about the hypervolume, see [hypervolume()].
#'
#' @inheritParams hypervolume
#'
#' @param ignore_dominated `logical(1)`\cr Whether dominated points are ignored
#'   when computing the contribution of nondominated points.  The value of this
#'   parameter has an effect on the return values only if the input contains
#'   dominated points. Setting this to `FALSE` slows down the computation
#'   significantly.  See the Notes below for a detailed explanation.
#'
#' @return `numeric()`\cr A numerical vector
#'
#' @author Manuel \enc{López-Ibáñez}{Lopez-Ibanez}
#'
#' @details
#'
#' The hypervolume contribution of point \eqn{\vec{p} \in X} is defined as
#' \eqn{\text{hvc}(\vec{p}) = \text{hyp}(X) - \text{hyp}(X \setminus
#' \{\vec{p}\})}.  This definition implies that duplicated points have zero
#' contribution even if not dominated, because removing one of the duplicates
#' does not change the hypervolume of the remaining set.  Moreover, dominated
#' points also have zero contribution. However, a point that is dominated by a
#' single (dominating) nondominated point reduces the contribution of the
#' latter, because removing the dominating point makes the dominated one become
#' nondominated.
#'
#' Handling this special case is non-trivial and makes the computation more
#' expensive, thus the default (`ignore_dominated=TRUE`) ignores all dominated
#' points in the input, that is, their contribution is set to zero and their
#' presence does not affect the contribution of any other point.  Setting
#' `ignore_dominated=FALSE` will consider dominated points according to the
#' mathematical definition given above, but the computation will be slower.
#'
#' When the input only consists of mutually nondominated points, the value of
#' `ignore_dominated` does not change the result, but the default value is
#' significantly faster.
#'
#' Like the hypervolume, computing the hypervolume contribution is #P-hard \citep{BriFri2012tcs}.
#'
#' The current implementation uses a \eqn{O(n\log n)} dimension-sweep
#' algorithm for 2D.  With `ignore_dominated=TRUE`, the 3D case uses the HVC3D
#' algorithm \citep{GueFon2017hv4d}, which has \eqn{O(n\log n)} complexity.
#' Otherwise, the implementation uses the naive algorithm that requires
#' calculating the hypervolume \eqn{|X|+1} times.
#'
#' @seealso [hypervolume()]
#'
#' @references
#'
#' \insertAllCited{}
#'
#' @doctest
#'
#' x <- matrix(c(5,1, 1,5, 4,2, 4,4, 5,1), ncol=2, byrow=TRUE)
#' @expect equal(c(0,3,3,0,0))
#' hv_contributions(x, reference=c(6,6))
#' # hvc[(5,1)] = 0 = duplicated
#' # hvc[(1,5)] = 3 = (4 - 1) * (6 - 5)
#' # hvc[(4,2)] = 3 = (5 - 4) * (5 - 2)
#' # hvc[(4,4)] = 0 = dominated
#' # hvc[(5,1)] = 0 = duplicated
#'
#' @expect equal(c(0,3,2,0,0))
#' hv_contributions(x, reference=c(6,6), ignore_dominated = FALSE)
#' # hvc[(5,1)] = 0 = duplicated
#' # hvc[(1,5)] = 3 = (4 - 1) * (6 - 5)
#' # hvc[(4,2)] = 2 = (5 - 4) * (4 - 2)
#' # hvc[(4,4)] = 0 = dominated
#' # hvc[(5,1)] = 0 = duplicated
#'
#' @omit
#' data(SPEA2minstoptimeRichmond)
#' # The second objective must be maximized
#' # We calculate the hypervolume contribution of each point of the union of all sets.
#' hv_contributions(SPEA2minstoptimeRichmond[, 1:2], reference = c(250, 0),
#'             maximise = c(FALSE, TRUE))
#'
#' # Duplicated points show zero contribution above, even if not
#' # dominated. However, filter_dominated removes all duplicates except
#' # one. Hence, there are more points below with nonzero contribution.
#' hv_contributions(filter_dominated(SPEA2minstoptimeRichmond[, 1:2], maximise = c(FALSE, TRUE)),
#'                  reference = c(250, 0), maximise = c(FALSE, TRUE))
#'
#' @export
#' @concept metrics
hv_contributions <- function(x, reference, maximise = FALSE, ignore_dominated = TRUE)
{
  x <- as_double_matrix(x)
  nobjs <- ncol(x)
  if (is.null(reference)) stop("reference cannot be NULL")
  if (length(reference) == 1L) reference <- rep_len(reference, nobjs)

  if (any(maximise)) {
    x <- transform_maximise(x, maximise)
    if (all(maximise)) {
      reference <- -reference
    } else {
      reference[maximise] <- -reference[maximise]
    }
  }
  check_dimension_max(nobjs, .libmoocore_constants[["MOOCORE_HV_DIMENSION_MAX"]])
  .Call(hv_contributions_C,
    t(x),
    as.double(reference),
    as.logical(ignore_dominated))
}

# Generated by doctest: do not edit by hand
# Please edit file in R/hv.R

test_that("Doctest: hypervolume", {
  # Created from @doctest for `hypervolume`
  # Source file: R/hv.R
  dat = matrix(c(5, 5, 4, 6, 2, 7, 7, 4), ncol = 2, byrow = TRUE)
  expect_equal(hypervolume(dat, ref = c(10, 10)), 38)
  expect_equal(hypervolume(dat, ref = 0, maximise = TRUE), 39)
  data(SPEA2minstoptimeRichmond)
  expect_equal(hypervolume(SPEA2minstoptimeRichmond[, 1:2], reference = c(250, 0),
  maximise = c(FALSE, TRUE)), 7911375.69)
})

# Generated by doctest: do not edit by hand
# Please edit file in R/hv.R

test_that("Doctest: hv_contributions", {
  # Created from @doctest for `hv_contributions`
  # Source file: R/hv.R
  x <- matrix(c(5, 1, 1, 5, 4, 2, 4, 4, 5, 1), ncol = 2, byrow = TRUE)
  expect_equal(hv_contributions(x, reference = c(6, 6)), c(0, 3, 3, 0, 0))
  expect_equal(hv_contributions(x, reference = c(6, 6), ignore_dominated = FALSE),
  c(0, 3, 2, 0, 0))
})

source("helper-common.R")

test_that("hypervolume", {

test_hv_file <- function(file, reference, maximise = FALSE) {
  nobj <- length(reference)
  dataset <- if (tools::file_ext(file) == "rds") readRDS(file) else read_datasets(file)
  hypervolume(dataset[,1:nobj], reference = reference, maximise)
}

expect_equal(test_hv_file("DTLZDiscontinuousShape.3d.front.1000pts.10.rds",
  reference = c(10,10,10)),
  719.223555475191)

expect_equal(test_hv_file("duplicated3.inp",
  reference = c(-14324, -14906, -14500, -14654, -14232, -14093)),
  1.52890128312393e+20)

})

test_that("hv_dim0_dim1", {
  x <- matrix(numeric(), nrow=5,ncol=0)
  expect_error(hypervolume(x, ref=c()), "must have at least 1 column")
  x <- matrix(0, ncol=1)
  expect_equal(0.0, hypervolume(x, ref=0.0))
  x <- matrix(runif(5), ncol=1)
  ref <- 1.1
  expect_equal(hypervolume(x, ref=ref), ref - min(x))
})

hv_contributions_slow <- function(dataset, reference, maximise = FALSE)
  hypervolume(dataset, reference, maximise) -
    sapply(seq_len(nrow(dataset)), function(x) hypervolume(dataset[-x, , drop = FALSE], reference, maximise))

hv_contributions_nondom_slow <- function(dataset, reference, maximise = FALSE) {
  nondom <- is_nondominated(dataset, maximise = maximise, keep_weakly=TRUE)
  hvc <- numeric(nrow(dataset))
  dataset <- dataset[nondom, , drop=FALSE]
  hvc[nondom] <- hypervolume(dataset, reference, maximise) -
    sapply(seq_len(nrow(dataset)), function(x) hypervolume(dataset[-x, , drop=FALSE], reference=reference, maximise=maximise))
  hvc
}

test_that("hv_contributions", {
  reference = c(250,0)
  maximise = c(FALSE,TRUE)
  expect_equal(hv_contributions(SPEA2minstoptimeRichmond[, 1:2], reference = reference, maximise = maximise, ignore_dominated=FALSE),
               hv_contributions_slow(SPEA2minstoptimeRichmond[, 1:2], reference = reference, maximise = maximise))
  expect_equal(hv_contributions(SPEA2minstoptimeRichmond[, 1:2], reference = reference, maximise = maximise),
    hv_contributions_nondom_slow(SPEA2minstoptimeRichmond[, 1:2], reference = reference, maximise = maximise))

})

test_that("hv_contributions 3D", {

  set.seed(42)
  pts <- matrix(runif(30L), ncol = 3L)
  ref <- c(2, 2, 2)
  expect_equal(tolerance = 1e-10,
    hv_contributions(pts, reference = ref, ignore_dominated=FALSE),
    hv_contributions_slow(pts, ref))
  expect_equal(tolerance = 1e-10,
    hv_contributions(pts, reference = ref, ignore_dominated=TRUE),
    hv_contributions_nondom_slow(pts, ref))
})

# Generated by doctest: do not edit by hand
# Please edit file in R/eafdiff.R

test_that("Doctest: eafdiff", {
  # Created from @doctest for `eafdiff`
  # Source file: R/eafdiff.R
  A1 <- read_datasets(text = "\n 3 2\n 2 3\n\n 2.5 1\n 1 2\n\n 1 2\n")
  A2 <- read_datasets(text = "\n 4 2.5\n 3 3\n 2.5 3.5\n\n 3 3\n 2.5 3.5\n\n 2 1\n")
  d <- eafdiff(A1, A2)
  expect_true(is.matrix(d))
  expect_equal(d, matrix(byrow = TRUE, ncol = 3, scan(quiet = TRUE, text = "1.0  2.0    2\n\n  2.0  1.0   -1\n\n  2.5  1.0    0\n\n  2.0  2.0    1\n\n  2.0  3.0    2\n\n  3.0  2.0    2\n\n  2.5  3.5    0\n\n  3.0  3.0    0\n\n  4.0  2.5    1")))
  d <- eafdiff(A1, A2, rectangles = TRUE)
  expect_equal(d, as.matrix(read.table(header = TRUE, text = "\n\n     xmin ymin xmax ymax diff\n\n     2.0  1.0  2.5  2.0   -1\n\n     1.0  2.0  2.0  Inf    2\n\n     2.5  1.0  Inf  2.0    0\n\n     2.0  2.0  3.0  3.0    1\n\n     2.0  3.5  2.5  Inf    2\n\n     2.0  3.0  3.0  3.5    2\n\n     3.0  2.5  4.0  3.0    2\n\n     3.0  2.0  Inf  2.5    2\n\n     4.0  2.5  Inf  3.0    1")))
})

test_that("eafdiff3d", {
  lin <- read_datasets("lin.S.txt")
  sph <- read_datasets("sph.S.txt")
  setcol <- ncol(lin)
  # This may stop working once we filter uninteresting values in the C code directly.
  DIFF <- eafdiff(lin, sph)
  x <- as.matrix(read.table("lin.S-sph.S-diff.txt.xz", header = FALSE))
  x[, setcol] <- x[, setcol] - x[, setcol+1]
  dimnames(x) <- NULL
  expect_equal(DIFF[, 1:setcol], x[, 1:setcol])
})

\newcommand{\citep}{\insertCite{#1}{moocore}}%
\newcommand{\citet}{\insertCite{#1;textual}{moocore}}%

% Macro to create link to an HTML vignette from a help page.
% arg 1 is the name of the vignette (without .html)
% arg 2 is the text to show on the link.
%% See also:
%% https://github.com/dmurdoch/rgl/blob/ec1e6cf237f91064a3e00b4627c6e39005779f83/man/macros/macros.Rd#L10
\newcommand{\HTMLVignette}{\href{https://multi-objective.github.io/moocore/r/articles/#1.html}{#2}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/moocore-package.R
\docType{data}
\name{SPEA2relativeRichmond}
\alias{SPEA2relativeRichmond}
\title{Results of SPEA2 with relative time-controlled triggers on Richmond water
network.}
\format{
A data frame as produced by \code{\link[=read_datasets]{read_datasets()}}.
}
\source{
\insertRef{LopezIbanezPhD}{moocore}
}
\usage{
SPEA2relativeRichmond
}
\description{
Results of SPEA2 with relative time-controlled triggers on Richmond water
network.
}
\examples{
data(SPEA2relativeRichmond)
str(SPEA2relativeRichmond)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eaf.R
\name{eaf_as_list}
\alias{eaf_as_list}
\title{Convert an EAF data frame to a list of data frames, where each element
of the list is one attainment surface. The function \code{\link[=attsurf2df]{attsurf2df()}} can be
used to convert the list into a single data frame.}
\usage{
eaf_as_list(eaf)
}
\arguments{
\item{eaf}{\code{data.frame()}|\code{matrix()}\cr Data frame or matrix that represents the EAF.}
}
\value{
\code{list()}\cr A list of data frames. Each \code{data.frame} represents one attainment surface.
}
\description{
Convert an EAF data frame to a list of data frames, where each element
of the list is one attainment surface. The function \code{\link[=attsurf2df]{attsurf2df()}} can be
used to convert the list into a single data frame.
}
\examples{
extdata_path <- system.file(package="moocore", "extdata")
x <- read_datasets(file.path(extdata_path, "example1_dat"))
attsurfs <- eaf_as_list(eaf(x, percentiles = c(0, 50, 100)))
str(attsurfs)
}
\seealso{
\code{\link[=eaf]{eaf()}} \code{\link[=attsurf2df]{attsurf2df()}}
}
\concept{eaf}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/rbind_datasets.R
\name{rbind_datasets}
\alias{rbind_datasets}
\title{Combine datasets \code{x} and \code{y} by row taking care of making all sets unique.}
\usage{
rbind_datasets(x, y)
}
\arguments{
\item{x, y}{\code{matrix}|\code{data.frame()}\cr Each dataset has at least three
columns, the last one is the set of each point. See also
\code{\link[=read_datasets]{read_datasets()}}.}
}
\value{
\verb{matrix()|}data.frame()`\cr A dataset.
}
\description{
Combine datasets \code{x} and \code{y} by row taking care of making all sets unique.
}
\examples{
x <- data.frame(f1 = 5:10, f2 = 10:5, set = 1:6)
y <- data.frame(f1 = 15:20, f2 = 20:15, set = 1:6)
rbind_datasets(x,y)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/largest_eafdiff.R
\name{choose_eafdiff}
\alias{choose_eafdiff}
\title{Interactively choose according to empirical attainment function differences}
\usage{
choose_eafdiff(x, left = stop("'left' must be either TRUE or FALSE"))
}
\arguments{
\item{x}{\code{matrix()}\cr Matrix of rectangles representing EAF differences
returned by \code{\link[=eafdiff]{eafdiff()}} with \code{rectangles=TRUE}.}

\item{left}{\code{logical(1)}\cr With \code{left=TRUE} return the rectangles with
positive differences, otherwise return those with negative differences but
differences are converted to positive.}
}
\value{
\code{matrix()} where the first 4 columns give the coordinates of two
corners of each rectangle and the last column. In both cases, the last
column gives the positive differences in favor of the chosen side.
}
\description{
Interactively choose according to empirical attainment function differences
}
\examples{
\donttest{
extdata_dir <- system.file(package="moocore", "extdata")
A1 <- read_datasets(file.path(extdata_dir, "wrots_l100w10_dat"))
A2 <- read_datasets(file.path(extdata_dir, "wrots_l10w100_dat"))
# Choose A1
rectangles <- eafdiff(A1, A2, intervals = 5, rectangles = TRUE)
rectangles <- choose_eafdiff(rectangles, left = TRUE)
reference <- c(max(A1[, 1], A2[, 1]), max(A1[, 2], A2[, 2]))
x <- split.data.frame(A1[,1:2], A1[,3])
hv_A1 <- sapply(split.data.frame(A1[, 1:2], A1[, 3]),
                 hypervolume, reference=reference)
hv_A2 <- sapply(split.data.frame(A2[, 1:2], A2[, 3]),
                 hypervolume, reference=reference)
print(fivenum(hv_A1))
print(fivenum(hv_A2))
whv_A1 <- sapply(split.data.frame(A1[, 1:2], A1[, 3]),
                 whv_rect, rectangles=rectangles, reference=reference)
whv_A2 <- sapply(split.data.frame(A2[, 1:2], A2[, 3]),
                 whv_rect, rectangles=rectangles, reference=reference)
print(fivenum(whv_A1))
print(fivenum(whv_A2))
}

}
\concept{eaf}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/vorob.R
\name{vorob_t}
\alias{vorob_t}
\alias{vorob_dev}
\title{Vorob'ev threshold, expectation and deviation}
\usage{
vorob_t(x, sets, reference, maximise = FALSE)

vorob_dev(x, sets, reference, ve = NULL, maximise = FALSE)
}
\arguments{
\item{x}{\code{matrix()}|\code{data.frame()}\cr Matrix or data frame of numerical
values that represents multiple sets of points, where each row represents
a point. If \code{sets} is missing, the last column of \code{x} gives the sets.}

\item{sets}{\code{integer()}\cr Vector that indicates the set of each point in
\code{x}. If missing, the last column of \code{x} is used instead.}

\item{reference}{\code{numeric()}\cr Reference point as a vector of numerical
values.}

\item{maximise}{\code{logical()}\cr Whether the objectives must be maximised
instead of minimised. Either a single logical value that applies to all
objectives or a vector of logical values, with one value per objective.}

\item{ve}{\code{matrix()}\cr Vorob'ev expectation, e.g., as returned by \code{\link[=vorob_t]{vorob_t()}}.}
}
\value{
\code{vorob_t} returns a list with elements \code{threshold},
\code{ve}, and \code{avg_hyp} (average hypervolume)

\code{vorob_dev} returns the Vorob'ev deviation.
}
\description{
Compute Vorob'ev threshold, expectation and deviation. Also, displaying the
symmetric deviation function is possible.  The symmetric deviation
function is the probability for a given target in the objective space to
belong to the symmetric difference between the Vorob'ev expectation and a
realization of the (random) attained set.
}
\details{
Let \eqn{\mathcal{A} = \{A_1, \dots, A_n\}} be a multi-set of \eqn{n} sets
\eqn{A_i \subset \mathbb{R}^d} of mutually nondominated vectors, with finite
(but not necessarily equal) cardinality.  If bounded by a reference point
\eqn{\vec{r}} that is strictly dominated by any point in any set, then these
sets can be seen a samples from a random closed set
\citep{Molchanov2005theory}.

Let the \eqn{\beta}-quantile be the subset of the empirical attainment
function \eqn{\mathcal{Q}_\beta = \{\vec{z}\in \mathbb{R}^d :
\hat{\alpha}_{\mathcal{A}}(\vec{z}) \geq \beta\}}.

The Vorob'ev \emph{expectation} is the \eqn{\beta^{*}}-quantile set
\eqn{\mathcal{Q}_{\beta^{*}}} such that the mean value hypervolume of the
sets is equal (or as close as possible) to the hypervolume of
\eqn{\mathcal{Q}_{\beta^{*}}}, that is, \eqn{\text{hyp}(\mathcal{Q}_\beta)
\leq \mathbb{E}[\text{hyp}(\mathcal{A})] \leq
\text{hyp}(\mathcal{Q}_{\beta^{*}})}, \eqn{\forall \beta > \beta^{*}}. Thus,
the Vorob'ev expectation provides a definition of the notion of \emph{mean}
nondominated set.

The value \eqn{\beta^{*} \in [0,1]} is called the Vorob'ev
\emph{threshold}. Large differences from the median quantile (0.5) indicate a
skewed distribution of \eqn{\mathcal{A}}.

The Vorob'ev \emph{deviation} is the mean hypervolume of the symmetric difference
between the Vorob'ev expectation and any set in \eqn{\mathcal{A}}, that is,
\eqn{\mathbb{E}[\text{hyp}(\mathcal{Q}_{\beta^{*}} \ominus \mathcal{A})]},
where the symmetric difference is defined as \eqn{A \ominus B = (A \setminus
B) \cup (B \setminus A)}.  Low deviation values indicate that the sets are
very similar, in terms of the location of the weakly dominated space, to the
Vorob'ev expectation.

For more background, see
\citet{BinGinRou2015gaupar,Molchanov2005theory,CheGinBecMol2013moda}.
}
\references{
\insertAllCited{}
}
\author{
Mickael Binois
}
\concept{eaf}
\examples{
data(CPFs)
res <- vorob_t(CPFs, reference = c(2, 200))
res$threshold
res$avg_hyp
# Now print Vorob'ev deviation
vd <- vorob_dev(CPFs, ve = res$ve, reference = c(2, 200))
vd
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eaf.R
\name{eaf}
\alias{eaf}
\title{Exact computation of the Empirical Attainment Function (EAF)}
\usage{
eaf(x, sets, percentiles = NULL, maximise = FALSE, groups = NULL)
}
\arguments{
\item{x}{\code{matrix()}|\code{data.frame()}\cr Matrix or data frame of numerical
values that represents multiple sets of points, where each row represents
a point. If \code{sets} is missing, the last column of \code{x} gives the sets.}

\item{sets}{\code{integer()}\cr Vector that indicates the set of each point in
\code{x}. If missing, the last column of \code{x} is used instead.}

\item{percentiles}{\code{numeric()}\cr Vector indicating which percentiles are
computed.  \code{NULL} computes all.}

\item{maximise}{\code{logical()}\cr Whether the objectives must be maximised
instead of minimised. Either a single logical value that applies to all
objectives or a vector of logical values, with one value per objective.}

\item{groups}{\code{factor()}\cr Indicates that the EAF must be computed
separately for data belonging to different groups.}
}
\value{
\code{data.frame()}\cr A data frame containing the exact representation
of EAF. The last column gives the percentile that corresponds to each
point. If groups is not \code{NULL}, then an additional column indicates to
which group the point belongs.
}
\description{
This function computes the EAF given a set of 2D or 3D points and a vector
\code{set} that indicates to which set each point belongs.
}
\details{
The empirical first-order attainment function (EAF) is used to assess the
performance of stochastic multiobjective optimisers such as multiobjective
evolutionary algorithms \citep{Grunert01}.  It is an estimator for the
first-order attainment function, which provides information about the
location and, to some extent, the variability of the random sets of
nondominated objective vectors produced by such optimisers when applied to
given problem instances.

Given a set \eqn{A \subset \mathbb{R}^d}, the attainment function of
\eqn{A}, denoted by \eqn{\alpha_{A}\colon \mathbb{R}^d\to \{0,1\}},
specifies which points in the objective space are weakly dominated by
\eqn{A}, where \eqn{\alpha_A(\vec{z}) = 1} if \eqn{\exists \vec{a} \in A,
\vec{a} \leq \vec{z}}, and \eqn{\alpha_A(\vec{z}) = 0}, otherwise.

Let \eqn{\mathcal{A} = \{A_1, \dots, A_n\}} be a multi-set of \eqn{n} sets
\eqn{A_i \subset \mathbb{R}^d}, the EAF \citep{Grunert01,GruFon2009:emaa} is
the function \eqn{\hat{\alpha}_{\mathcal{A}}\colon \mathbb{R}^d\to [0,1]},
such that:

\deqn{\hat{\alpha}_{\mathcal{A}}(\vec{z}) = \frac{1}{n}\sum_{i=1}^n \alpha_{A_i}(\vec{z})}

The EAF is a coordinate-wise non-decreasing step function, similar to the
empirical cumulative distribution function (ECDF)
\citep{LopVerDreDoe2025}.  Thus, a finite representation of the EAF
can be computed as the set of minima, in terms of Pareto optimality, with a
value of the EAF not smaller than a given \eqn{t/n}, where
\eqn{t=1,\dots,n} \citep{FonGueLopPaq2011emo}. Formally, the EAF can
be represented by the sequence \eqn{(L_1, L_2, \dots, L_n)}, where:

\deqn{L_t = \min \{\vec{z} \in \mathbb{R}^d : \hat{\alpha}_{\mathcal{A}}(\vec{z}) \geq t/n\}}

It is also common to refer to the \eqn{k\% \in [0,100]} percentile. For
example, the \emph{median} (or 50\%) attainment surface corresponds to
\eqn{L_{\lceil n/2 \rceil}} and it is the lower boundary of the vector space
attained by at least 50\% of the input sets \eqn{A_i}. Similarly, \eqn{L_1}
is called the \emph{best} attainment surface (\eqn{\frac{1}{n}}\%) and
represents the lower boundary of the space attained by at least one input
set, whereas \eqn{L_{100}} is called the \emph{worst} attainment surface (100\%)
and represents the lower boundary of the space attained by all input sets.

In the current implementation, the EAF is computed using the algorithms
proposed by \citet{FonGueLopPaq2011emo}, which have complexity \eqn{\Theta(m\log
m + nm)} in 2D and \eqn{O(n^2 m \log m)} in 3D, where \eqn{n} is the number
of input sets and \eqn{m} is the total number of input points.
}
\note{
There are several examples of data sets in
\code{system.file(package="moocore","extdata")}.  The current implementation only
supports two and three dimensional points.
}
\examples{
extdata_path <- system.file(package="moocore", "extdata")

x <- read_datasets(file.path(extdata_path, "example1_dat"))
# Compute full EAF (sets is the last column)
str(eaf(x))

# Compute only best, median and worst
str(eaf(x[,1:2], sets = x[,3], percentiles = c(0, 50, 100)))

x <- read_datasets(file.path(extdata_path, "spherical-250-10-3d.txt"))
y <- read_datasets(file.path(extdata_path, "uniform-250-10-3d.txt"))
x <- rbind(data.frame(x, groups = "spherical"),
           data.frame(y, groups = "uniform"))
# Compute only median separately for each group
z <- eaf(x[,1:3], sets = x[,4], groups = x[,5], percentiles = 50)
str(z)
}
\references{
\insertAllCited{}
}
\seealso{
\code{\link[=read_datasets]{read_datasets()}}
}
\author{
Manuel \enc{López-Ibáñez}{Lopez-Ibanez}
}
\concept{eaf}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/epsilon.R
\name{epsilon}
\alias{epsilon}
\alias{epsilon_additive}
\alias{epsilon_mult}
\title{Epsilon metric}
\usage{
epsilon_additive(x, reference, maximise = FALSE)

epsilon_mult(x, reference, maximise = FALSE)
}
\arguments{
\item{x}{\code{matrix()}|\code{data.frame()}\cr Matrix or data frame of numerical
values, where each row gives the coordinates of a point.}

\item{reference}{\code{matrix}|\code{data.frame}\cr Reference set as a matrix or
data.frame of numerical values.}

\item{maximise}{\code{logical()}\cr Whether the objectives must be maximised
instead of minimised. Either a single logical value that applies to all
objectives or a vector of logical values, with one value per objective.}
}
\value{
\code{numeric(1)} A single numerical value.
}
\description{
Computes the epsilon metric, either additive or multiplicative.
}
\details{
The epsilon metric of a set \eqn{A \subset \mathbb{R}^m} with respect to a
reference set \eqn{R \subset \mathbb{R}^m} is defined as

\deqn{epsilon(A,R) = \max_{r \in R} \min_{a \in A} \max_{1 \leq i \leq m} epsilon(a_i, r_i)}

where \eqn{a} and \eqn{r} are objective vectors of length \eqn{m}.

In the case of minimization of objective \eqn{i}, \eqn{epsilon(a_i,r_i)} is
computed as \eqn{a_i/r_i} for the multiplicative variant (respectively,
\eqn{a_i - r_i} for the additive variant), whereas in the case of
maximization of objective \eqn{i}, \eqn{epsilon(a_i,r_i) = r_i/a_i} for the
multiplicative variant (respectively, \eqn{r_i - a_i} for the additive
variant). This allows computing a single value for problems where some
objectives are to be maximized while others are to be minimized. Moreover, a
lower value corresponds to a better approximation set, independently of the
type of problem (minimization, maximization or mixed). However, the meaning
of the value is different for each objective type. For example, imagine that
objective 1 is to be minimized and objective 2 is to be maximized, and the
multiplicative epsilon computed here for \eqn{epsilon(A,R) = 3}. This means
that \eqn{A} needs to be multiplied by 1/3 for all \eqn{a_1} values and by 3
for all \eqn{a_2} values in order to weakly dominate \eqn{R}.

The multiplicative variant can be computed as \eqn{\exp(epsilon_{+}(\log(A),
\log(R)))}, which makes clear that the computation of the multiplicative
version for zero or negative values doesn't make sense. See the examples
below.

The current implementation uses the naive algorithm that requires
\eqn{O(m \cdot |A| \cdot |R|)}, where \eqn{m} is the number of objectives
(dimension of vectors).
}
\references{
\insertRef{ZitThiLauFon2003:tec}{moocore}
}
\author{
Manuel \enc{López-Ibáñez}{Lopez-Ibanez}
}
\concept{metrics}
\examples{
# Fig 6 from Zitzler et al. (2003).
A1 <- matrix(c(9,2,8,4,7,5,5,6,4,7), ncol=2, byrow=TRUE)
A2 <- matrix(c(8,4,7,5,5,6,4,7), ncol=2, byrow=TRUE)
A3 <- matrix(c(10,4,9,5,8,6,7,7,6,8), ncol=2, byrow=TRUE)
if (requireNamespace("graphics", quietly = TRUE)) {
   plot(A1, xlab=expression(f[1]), ylab=expression(f[2]),
        panel.first=grid(nx=NULL), pch=4, cex=1.5, xlim = c(0,10), ylim=c(0,8))
   points(A2, pch=0, cex=1.5)
   points(A3, pch=1, cex=1.5)
   legend("bottomleft", legend=c("A1", "A2", "A3"), pch=c(4,0,1),
          pt.bg="gray", bg="white", bty = "n", pt.cex=1.5, cex=1.2)
}
epsilon_mult(A1, A3) # A1 epsilon-dominates A3 => e = 9/10 < 1
epsilon_mult(A1, A2) # A1 weakly dominates A2 => e = 1
epsilon_mult(A2, A1) # A2 is epsilon-dominated by A1 => e = 2 > 1
# Equivalence between additive and multiplicative
exp(epsilon_additive(log(A2), log(A1)))

# A more realistic example
extdata_path <- system.file(package="moocore","extdata")
path.A1 <- file.path(extdata_path, "ALG_1_dat.xz")
path.A2 <- file.path(extdata_path, "ALG_2_dat.xz")
A1 <- read_datasets(path.A1)[,1:2]
A2 <- read_datasets(path.A2)[,1:2]
ref <- filter_dominated(rbind(A1, A2))
epsilon_additive(A1, ref)
epsilon_additive(A2, ref)
# Multiplicative version of epsilon metric
ref <- filter_dominated(rbind(A1, A2))
epsilon_mult(A1, ref)
epsilon_mult(A2, ref)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/r2.R
\name{r2_exact}
\alias{r2_exact}
\title{Exact R2 indicator}
\usage{
r2_exact(x, reference, maximise = FALSE)
}
\arguments{
\item{x}{\code{matrix()}|\code{data.frame()}\cr Matrix or data frame of numerical
values, where each row gives the coordinates of a point.}

\item{reference}{\code{numeric()}\cr Reference (ideal) point as a vector of numerical
values.}

\item{maximise}{\code{logical()}\cr Whether the objectives must be maximised
instead of minimised. Either a single logical value that applies to all
objectives or a vector of logical values, with one value per objective.}
}
\value{
\code{numeric(1)} A single numerical value.
}
\description{
Computes the exact R2 indicator with respect to a given ideal/utopian
reference point assuming minimization of all objectives.
}
\details{
The unary R2 indicator is a quality indicator for a set \eqn{A \subset
\mathbb{R}^m}{A in R^m} w.r.t. an ideal or utopian reference point
\eqn{\vec{r} \in \mathbb{R}^m}{r in R^m}.  It was originally proposed by
\citet{HanJas1998} and is defined as the expected Tchebycheff utility under a
uniform distribution of weight vectors (w.l.o.g. assuming minimization):

\deqn{R2(A) := \int_{w \in W} \min_{a \in A} \left\{ \max_{i=1,\dots,m} w_i
(a_i - r_i) \right\} \, dw,}{R2(A) := integral over w in W of min over a in
A of max over i of w_i (a_i - r_i) dw,}

where \eqn{W} denotes the uniform distribution across weights:

\deqn{W = \{w \in \mathbb{R}^m \mid w_i \geq 0, \sum_{i=1}^m w_i = 1\}.}{W
= \{w in R^m | w_i >= 0, sum_i w_i = 1\}.}

The R2 indicator is to be minimized and has an optimal value of 0 when
\eqn{\vec{r} \in A}{r in A}.

The exact R2 indicator is strongly Pareto-compliant, i.e., compatible with
Pareto-optimality:

\deqn{\forall A, B \subset \mathbb{R}^m: A \prec B \Rightarrow R2(A) <
R2(B).}{for all A, B in R^m: A dominates B implies R2(A) < R2(B).}

Given an ideal or utopian reference point, which is available in most
scenarios, all non-dominated solutions always contribute to the value of
the exact R2 indicator.  However, it is scale-dependent and care should be
taken such that all objectives contribute approximately equally to the
indicator, e.g., by normalizing the Pareto front to the unit hypercube.

The current implementation exclusively supports bi-objective solution sets
and runs in \eqn{O(n \log n)} following \citet{SchKer2025r2v2}.
}
\references{
\insertAllCited{}
}
\concept{metrics}
\examples{
dat <- matrix(c(5, 5, 4, 6, 2, 7, 7, 4), ncol = 2, byrow = TRUE)
r2_exact(dat, reference = c(0, 0))

# This function assumes minimisation by default. We can easily specify maximisation:
r2_exact(dat, reference = c(10, 10), maximise = TRUE)

# Merge all the sets of a dataset by removing the set number column:
extdata_path <- system.file(package="moocore","extdata")
dat <- read_datasets(file.path(extdata_path, "example1_dat"))[, 1:2]
nrow(dat)

# Dominated points are ignored, so this:
r2_exact(dat, reference = 0)

# gives the same exact R2 value as this:
dat <- filter_dominated(dat)
nrow(dat)
r2_exact(dat, reference = 0)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eafdiff.R
\name{compute_eafdiff_call}
\alias{compute_eafdiff_call}
\title{Same as \code{\link[=eafdiff]{eafdiff()}} but performs no checks and does not transform the input
or the output. This function should be used by other packages that want to
avoid redundant checks and transformations.}
\usage{
compute_eafdiff_call(x, y, cumsizes_x, cumsizes_y, intervals, ret)
}
\arguments{
\item{x, y}{\code{matrix}|\code{data.frame()}\cr Data frames corresponding to the input data of
left and right sides, respectively. Each data frame has at least three
columns, the last one is the set of each point. See also
\code{\link[=read_datasets]{read_datasets()}}.}

\item{cumsizes_x, cumsizes_y}{Cumulative size of the different sets of points in \code{x} and \code{y}.}

\item{intervals}{\code{integer(1)}\cr The absolute range of the differences
\eqn{[0, 1]} is partitioned into the number of intervals provided.}

\item{ret}{(\code{"points"|"rectangles"|"polygons"})\cr The format of the returned EAF differences.}
}
\value{
With \code{rectangle=FALSE}, a \code{data.frame} containing points where there
is a transition in the value of the EAF differences.  With
\code{rectangle=TRUE}, a \code{matrix} where the first 4 columns give the
coordinates of two corners of each rectangle. In both cases, the last
column gives the difference in terms of sets in \code{x} minus sets in \code{y} that
attain each point (i.e., negative values are differences in favour \code{y}).
}
\description{
Same as \code{\link[=eafdiff]{eafdiff()}} but performs no checks and does not transform the input
or the output. This function should be used by other packages that want to
avoid redundant checks and transformations.
}
\seealso{
\code{\link[=as_double_matrix]{as_double_matrix()}} \code{\link[=transform_maximise]{transform_maximise()}}
}
\concept{eaf}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hv.R
\name{hypervolume}
\alias{hypervolume}
\title{Hypervolume metric}
\usage{
hypervolume(x, reference, maximise = FALSE)
}
\arguments{
\item{x}{\code{matrix()}|\code{data.frame()}\cr Matrix or data frame of numerical
values, where each row gives the coordinates of a point.}

\item{reference}{\code{numeric()}\cr Reference point as a vector of numerical
values.}

\item{maximise}{\code{logical()}\cr Whether the objectives must be maximised
instead of minimised. Either a single logical value that applies to all
objectives or a vector of logical values, with one value per objective.}
}
\value{
\code{numeric(1)} A single numerical value.
}
\description{
Compute the hypervolume metric with respect to a given reference point
assuming minimization of all objectives.
}
\details{
The hypervolume of a set of multidimensional points \eqn{A \subset
\mathbb{R}^m}{A in R^m}, where \eqn{m} is the dimension of the points, with
respect to a reference point \eqn{\vec{r} \in \mathbb{R}^m}{r in R^m} is the
volume of the region dominated by the set and bounded by the reference point
\citep{ZitThi1998ppsn}.  Points in \eqn{A} that do not strictly dominate
\eqn{\vec{r}} do not contribute to the hypervolume value, thus, ideally, the
reference point must be strictly dominated by all points in the true Pareto
front.

More precisely, the hypervolume is the \href{https://en.wikipedia.org/wiki/Lebesgue_measure}{Lebesgue measure} of the union of
axis-aligned hyperrectangles
(\href{https://en.wikipedia.org/wiki/Hyperrectangle}{orthotopes}), where each
hyperrectangle is defined by one point from \eqn{\vec{a} \in A} and the
reference point.  The union of axis-aligned hyperrectangles is also called
an \emph{orthogonal polytope}.

The hypervolume is compatible with Pareto-optimality
\citep{KnoCor2002cec,ZitThiLauFon2003:tec}, that is, \eqn{\nexists A,B
\subset \mathbb{R}^m}{it does not exist A,B subsets of R^m}, such that
\eqn{A} is better than \eqn{B} according to Pareto-optimality and
\eqn{\text{hyp}(A) \leq \text{hyp}(B)}{hyp(A) <= hyp(B)}. In other words, if
the hypervolume of a set is larger than the hypervolume of another, then we
know for sure than the latter set cannot be better than the former in terms
of Pareto-optimality. The hypervolume is also the only unary metric known
that is complete according to Pareto-optimality
\citep{ZitThiLauFon2003:tec}, that is, if \eqn{A} is better than \eqn{B}
according to Pareto-optimality, then \eqn{\text{hyp}(A) > \text{hyp}(B)}.

Like most measures of unions of high-dimensional geometric objects,
computing the hypervolume is #P-hard \citep{BriFri2010approx}, which means
that the best possible algorithm takes an exponential time on the number of
objectives or points, in the worst-case.

For 2D and 3D, the algorithms used
\citep{FonPaqLop06:hypervolume,BeuFonLopPaqVah09:tec} have \eqn{O(n \log n)}
complexity, where \eqn{n} is the number of input points. The 3D case uses
the \eqn{\text{HV3D}^{+}} algorithm \citep{GueFon2017hv4d}, which has the
sample complexity as the HV3D algorithm
\citep{FonPaqLop06:hypervolume,BeuFonLopPaqVah09:tec}, but it is faster in
practice.

For 4D, the algorithm used is \eqn{\text{HV4D}^{+}} \citep{GueFon2017hv4d},
which has \eqn{O(n^2)} complexity.  Compared to the \href{https://github.com/apguerreiro/HVC/}{original implementation}, this implementation
correctly handles weakly dominated points and has been further optimized for
speed.

For 5D or higher and up to 12 points, the implementation uses the
inclusion-exclusion algorithm \citep{WuAza2001metrics}, which has \eqn{O(m
2^{n})} time and \eqn{O(n\cdot m)} space complexity, but it is very fast for
such small sets.  For larger number of points, it uses a recursive algorithm
\citep{FonPaqLop06:hypervolume} that computes 4D contributions
\citep{GueFon2017hv4d} as the base case, resulting in a \eqn{O(n^{m-2})}
time complexity and \eqn{O(n)} space complexity in the worst-case.
Experimental results show that the pruning techniques used may reduce the
time complexity even further.  The original proposal
\citep{FonPaqLop06:hypervolume} had the HV3D algorithm as the base case,
giving a time complexity of \eqn{O(n^{m-2} \log n)}.  Andreia P. Guerreiro
enhanced the numerical stability of the algorithm by avoiding floating-point
comparisons of partial hypervolumes.

The hypervolume of 1D inputs is defined as \code{max(0, ref - min(x))}.
}
\references{
\insertAllCited{}
}
\author{
Manuel \enc{López-Ibáñez}{Lopez-Ibanez}
}
\concept{metrics}
\examples{
dat = matrix(c(5, 5, 4, 6, 2, 7, 7, 4), ncol=2, byrow=TRUE)
hypervolume(dat, ref=c(10, 10))
hypervolume(dat, ref=0, maximise=TRUE)

data(SPEA2minstoptimeRichmond)
# The second objective must be maximized
# We calculate the hypervolume of the union of all sets.
hypervolume(SPEA2minstoptimeRichmond[, 1:2], reference = c(250, 0),
            maximise = c(FALSE, TRUE))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/moocore-package.R
\docType{data}
\name{CPFs}
\alias{CPFs}
\title{Conditional Pareto fronts obtained from Gaussian processes simulations.}
\format{
A data frame with 2967 observations on the following 3 variables.
\describe{
\item{\code{f1}}{first objective values.}
\item{\code{f2}}{second objective values.}
\item{\code{set}}{indices of corresponding conditional Pareto fronts.}
}
}
\source{
\insertRef{BinGinRou2015gaupar}{moocore}
}
\usage{
CPFs
}
\description{
The data has the only goal of providing an example of use of \code{\link[=vorob_t]{vorob_t()}} and
\code{\link[=vorob_dev]{vorob_dev()}}. It has been obtained by fitting two Gaussian processes on 20
observations of a bi-objective problem, before generating conditional
simulation of both GPs at different locations and extracting non-dominated
values of coupled simulations.
}
\examples{
data(CPFs)
vorob_t(CPFs, reference = c(2, 200))
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eaf.R
\name{compute_eaf_call}
\alias{compute_eaf_call}
\title{Same as \code{\link[=eaf]{eaf()}} but performs no checks and does not transform the input or
the output. This function should be used by other packages that want to
avoid redundant checks and transformations.}
\usage{
compute_eaf_call(x, cumsizes, percentiles)
}
\arguments{
\item{x}{\code{matrix()}|\code{data.frame()}\cr Matrix or data frame of numerical
values that represents multiple sets of points, where each row represents
a point. If \code{sets} is missing, the last column of \code{x} gives the sets.}

\item{cumsizes}{\code{integer()}\cr Cumulative size of the different sets of points in \code{x}.}

\item{percentiles}{\code{numeric()}\cr Vector indicating which percentiles are
computed.  \code{NULL} computes all.}
}
\value{
\code{data.frame()}\cr A data frame containing the exact representation
of EAF. The last column gives the percentile that corresponds to each
point. If groups is not \code{NULL}, then an additional column indicates to
which group the point belongs.
}
\description{
Same as \code{\link[=eaf]{eaf()}} but performs no checks and does not transform the input or
the output. This function should be used by other packages that want to
avoid redundant checks and transformations.
}
\seealso{
\code{\link[=as_double_matrix]{as_double_matrix()}} \code{\link[=transform_maximise]{transform_maximise()}}
}
\concept{eaf}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/moocore-package.R
\docType{data}
\name{SPEA2minstoptimeRichmond}
\alias{SPEA2minstoptimeRichmond}
\title{Results of SPEA2 when minimising electrical cost and maximising the
minimum idle time of pumps on Richmond water network.}
\format{
A data frame as produced by \code{\link[=read_datasets]{read_datasets()}}. The second
column measures time in seconds and corresponds to a maximisation problem.
}
\source{
\insertRef{LopezIbanezPhD}{moocore}
}
\usage{
SPEA2minstoptimeRichmond
}
\description{
Results of SPEA2 when minimising electrical cost and maximising the
minimum idle time of pumps on Richmond water network.
}
\examples{
data(SPEA2minstoptimeRichmond)
str(SPEA2minstoptimeRichmond)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/moocore-package.R
\docType{package}
\name{moocore-package}
\alias{moocore}
\alias{moocore-package}
\title{moocore: Core Mathematical Functions for Multi-Objective Optimization}
\description{
Fast implementations of mathematical operations and performance metrics for multi-objective optimization, including filtering and ranking of dominated vectors according to Pareto optimality, hypervolume metric, C.M. Fonseca, L. Paquete, M. López-Ibáñez (2006) \doi{10.1109/CEC.2006.1688440}, epsilon indicator, inverted generational distance, computation of the empirical attainment function, V.G. da Fonseca, C.M. Fonseca, A.O. Hall (2001) \doi{10.1007/3-540-44719-9_15}, and Vorob'ev threshold, expectation and deviation, M. Binois, D. Ginsbourger, O. Roustant (2015) \doi{10.1016/j.ejor.2014.07.032}, among others.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://multi-objective.github.io/moocore/r/}
  \item \url{https://github.com/multi-objective/moocore}
  \item Report bugs at \url{https://github.com/multi-objective/moocore/issues}
}

}
\author{
\strong{Maintainer}: Manuel López-Ibáñez \email{manuel.lopez-ibanez@manchester.ac.uk} (\href{https://orcid.org/0000-0001-9974-1295}{ORCID})

Other contributors:
\itemize{
  \item Carlos Fonseca [contributor]
  \item Luís Paquete [contributor]
  \item Andreia P. Guerreiro [contributor]
  \item Mickaël Binois [contributor]
  \item Michael H. Buselli (AVL-tree library) [copyright holder]
  \item Wessel Dankers (AVL-tree library) [copyright holder]
  \item NumPy Developers (RNG and ziggurat constants) [copyright holder]
  \item Jean-Sebastien Roy (mt19937 library) [copyright holder]
  \item Makoto Matsumoto (mt19937 library) [copyright holder]
  \item Takuji Nishimura (mt19937 library) [copyright holder]
}

}
\keyword{internal}

\newcommand{\citep}{\insertCite{#1}{moocore}}%
\newcommand{\citet}{\insertCite{#1;textual}{moocore}}%
\newcommand{\geq}{\ge}
\newcommand{\leq}{\le}

% Macro to create link to an HTML vignette from a help page.
% arg 1 is the name of the vignette (without .html)
% arg 2 is the text to show on the link.
%% See also:
%% https://github.com/dmurdoch/rgl/blob/ec1e6cf237f91064a3e00b4627c6e39005779f83/man/macros/macros.Rd#L10
\newcommand{\HTMLVignette}{\href{https://multi-objective.github.io/moocore/r/articles/#1.html}{#2}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/moocore-package.R
\docType{data}
\name{tpls50x20_1_MWT}
\alias{tpls50x20_1_MWT}
\title{Various strategies of Two-Phase Local Search applied to the Permutation
Flowshop Problem with Makespan and Weighted Tardiness objectives.}
\format{
A data frame with 1511 observations of 4 variables:
\describe{
\item{\code{algorithm}}{TPLS search strategy}
\item{\code{Makespan}}{first objective values.}
\item{\code{WeightedTardiness}}{second objective values.}
\item{\code{run}}{index of the run.}
}
}
\source{
\insertRef{DubLopStu2011amai}{moocore}
}
\usage{
tpls50x20_1_MWT
}
\description{
Various strategies of Two-Phase Local Search applied to the Permutation
Flowshop Problem with Makespan and Weighted Tardiness objectives.
}
\examples{
data(tpls50x20_1_MWT)
str(tpls50x20_1_MWT)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eafdiff.R
\name{eafdiff}
\alias{eafdiff}
\title{Compute empirical attainment function differences}
\usage{
eafdiff(x, y, intervals = NULL, maximise = FALSE, rectangles = FALSE)
}
\arguments{
\item{x, y}{\code{matrix}|\code{data.frame()}\cr Data frames corresponding to the input data of
left and right sides, respectively. Each data frame has at least three
columns, the last one is the set of each point. See also
\code{\link[=read_datasets]{read_datasets()}}.}

\item{intervals}{\code{integer(1)}\cr The absolute range of the differences
\eqn{[0, 1]} is partitioned into the number of intervals provided.}

\item{maximise}{\code{logical()}\cr Whether the objectives must be maximised
instead of minimised. Either a single logical value that applies to all
objectives or a vector of logical values, with one value per objective.}

\item{rectangles}{\code{logical(1)}\cr If TRUE, the output is in the form of rectangles of the same color.}
}
\value{
With \code{rectangle=FALSE}, a \code{data.frame} containing points where there
is a transition in the value of the EAF differences.  With
\code{rectangle=TRUE}, a \code{matrix} where the first 4 columns give the
coordinates of two corners of each rectangle. In both cases, the last
column gives the difference in terms of sets in \code{x} minus sets in \code{y} that
attain each point (i.e., negative values are differences in favour \code{y}).
}
\description{
Calculate the differences between the empirical attainment functions of two
data sets.
}
\details{
This function calculates the differences between the EAFs of two
data sets.
}
\seealso{
\code{\link[=read_datasets]{read_datasets()}}
}
\concept{eaf}
\examples{
A1 <- read_datasets(text='
 3 2
 2 3

 2.5 1
 1 2

 1 2
')

A2 <- read_datasets(text='
 4 2.5
 3 3
 2.5 3.5

 3 3
 2.5 3.5

 2 1
')
d <- eafdiff(A1, A2)
str(d)
d









d <- eafdiff(A1, A2, rectangles = TRUE)
str(d)
d











}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hv.R
\name{hv_contributions}
\alias{hv_contributions}
\title{Hypervolume contribution of a set of points}
\usage{
hv_contributions(x, reference, maximise = FALSE, ignore_dominated = TRUE)
}
\arguments{
\item{x}{\code{matrix()}|\code{data.frame()}\cr Matrix or data frame of numerical
values, where each row gives the coordinates of a point.}

\item{reference}{\code{numeric()}\cr Reference point as a vector of numerical
values.}

\item{maximise}{\code{logical()}\cr Whether the objectives must be maximised
instead of minimised. Either a single logical value that applies to all
objectives or a vector of logical values, with one value per objective.}

\item{ignore_dominated}{\code{logical(1)}\cr Whether dominated points are ignored
when computing the contribution of nondominated points.  The value of this
parameter has an effect on the return values only if the input contains
dominated points. Setting this to \code{FALSE} slows down the computation
significantly.  See the Notes below for a detailed explanation.}
}
\value{
\code{numeric()}\cr A numerical vector
}
\description{
Computes the hypervolume contribution of each point of a set of points with
respect to a given reference point. Duplicated and dominated points have
zero contribution.  By default, dominated points are ignored, that is, they
do not affect the contribution of other points.  See the Notes below for
more details.  For details about the hypervolume, see \code{\link[=hypervolume]{hypervolume()}}.
}
\details{
The hypervolume contribution of point \eqn{\vec{p} \in X} is defined as
\eqn{\text{hvc}(\vec{p}) = \text{hyp}(X) - \text{hyp}(X \setminus
\{\vec{p}\})}.  This definition implies that duplicated points have zero
contribution even if not dominated, because removing one of the duplicates
does not change the hypervolume of the remaining set.  Moreover, dominated
points also have zero contribution. However, a point that is dominated by a
single (dominating) nondominated point reduces the contribution of the
latter, because removing the dominating point makes the dominated one become
nondominated.

Handling this special case is non-trivial and makes the computation more
expensive, thus the default (\code{ignore_dominated=TRUE}) ignores all dominated
points in the input, that is, their contribution is set to zero and their
presence does not affect the contribution of any other point.  Setting
\code{ignore_dominated=FALSE} will consider dominated points according to the
mathematical definition given above, but the computation will be slower.

When the input only consists of mutually nondominated points, the value of
\code{ignore_dominated} does not change the result, but the default value is
significantly faster.

Like the hypervolume, computing the hypervolume contribution is #P-hard \citep{BriFri2012tcs}.

The current implementation uses a \eqn{O(n\log n)} dimension-sweep
algorithm for 2D.  With \code{ignore_dominated=TRUE}, the 3D case uses the HVC3D
algorithm \citep{GueFon2017hv4d}, which has \eqn{O(n\log n)} complexity.
Otherwise, the implementation uses the naive algorithm that requires
calculating the hypervolume \eqn{|X|+1} times.
}
\references{
\insertAllCited{}
}
\seealso{
\code{\link[=hypervolume]{hypervolume()}}
}
\author{
Manuel \enc{López-Ibáñez}{Lopez-Ibanez}
}
\concept{metrics}
\examples{
x <- matrix(c(5,1, 1,5, 4,2, 4,4, 5,1), ncol=2, byrow=TRUE)
hv_contributions(x, reference=c(6,6))
# hvc[(5,1)] = 0 = duplicated
# hvc[(1,5)] = 3 = (4 - 1) * (6 - 5)
# hvc[(4,2)] = 3 = (5 - 4) * (5 - 2)
# hvc[(4,4)] = 0 = dominated
# hvc[(5,1)] = 0 = duplicated
hv_contributions(x, reference=c(6,6), ignore_dominated = FALSE)
# hvc[(5,1)] = 0 = duplicated
# hvc[(1,5)] = 3 = (4 - 1) * (6 - 5)
# hvc[(4,2)] = 2 = (5 - 4) * (4 - 2)
# hvc[(4,4)] = 0 = dominated
# hvc[(5,1)] = 0 = duplicated
data(SPEA2minstoptimeRichmond)
# The second objective must be maximized
# We calculate the hypervolume contribution of each point of the union of all sets.
hv_contributions(SPEA2minstoptimeRichmond[, 1:2], reference = c(250, 0),
            maximise = c(FALSE, TRUE))

# Duplicated points show zero contribution above, even if not
# dominated. However, filter_dominated removes all duplicates except
# one. Hence, there are more points below with nonzero contribution.
hv_contributions(filter_dominated(SPEA2minstoptimeRichmond[, 1:2], maximise = c(FALSE, TRUE)),
                 reference = c(250, 0), maximise = c(FALSE, TRUE))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{transform_maximise}
\alias{transform_maximise}
\title{Transform matrix according to maximise parameter}
\usage{
transform_maximise(x, maximise)
}
\arguments{
\item{x}{\code{matrix()}|\code{data.frame()}\cr Matrix or data frame of numerical
values, where each row gives the coordinates of a point.}

\item{maximise}{\code{logical()}\cr Whether the objectives must be maximised
instead of minimised. Either a single logical value that applies to all
objectives or a vector of logical values, with one value per objective.}
}
\value{
\code{x} transformed such that every column where \code{maximise} is \code{TRUE} is multiplied by \code{-1}.
}
\description{
Transform matrix according to maximise parameter
}
\examples{
x <- data.frame(f1=1:10, f2=101:110)
rownames(x) <- letters[1:10]
transform_maximise(x, maximise=c(FALSE,TRUE))
transform_maximise(x, maximise=TRUE)
x <- as.matrix(x)
transform_maximise(x, maximise=c(FALSE,TRUE))
transform_maximise(x, maximise=TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/igd.R
\name{igd}
\alias{igd}
\alias{IGDX}
\alias{igd_plus}
\alias{avg_hausdorff_dist}
\title{Inverted Generational Distance (IGD and IGD+) and Averaged Hausdorff Distance}
\usage{
igd(x, reference, maximise = FALSE)

igd_plus(x, reference, maximise = FALSE)

avg_hausdorff_dist(x, reference, maximise = FALSE, p = 1L)
}
\arguments{
\item{x}{\code{matrix()}|\code{data.frame()}\cr Matrix or data frame of numerical
values, where each row gives the coordinates of a point.}

\item{reference}{\code{matrix}|\code{data.frame}\cr Reference set as a matrix or
data.frame of numerical values.}

\item{maximise}{\code{logical()}\cr Whether the objectives must be maximised
instead of minimised. Either a single logical value that applies to all
objectives or a vector of logical values, with one value per objective.}

\item{p}{\code{integer(1)}\cr Hausdorff distance parameter (default: \code{1L}).}
}
\value{
\code{numeric(1)} A single numerical value.
}
\description{
Functions to compute the inverted generational distance (IGD and IGD+) and
the averaged Hausdorff distance between nondominated sets of points.
}
\details{
The generational distance (GD) of a set \eqn{A} is defined as the distance
between each point \eqn{a \in A} and the closest point \eqn{r} in a
reference set \eqn{R}, averaged over the size of \eqn{A}. Formally,

\deqn{GD_p(A,R) = \left(\frac{1}{|A|}\sum_{a\in A}\min_{r\in R} d(a,r)^p\right)^{\frac{1}{p}} }{GD(A,R) = (1/|A|) * ( sum_{a in A} min_{r in R} d(a,r)^p )^(1/p)}
where the distance in our implementation is the Euclidean distance:
\deqn{d(a,r) = \sqrt{\sum_{k=1}^m (a_k - r_k)^2} }{d(a,r) = sqrt( sum_{k=1}^m (a_k - r_k)^2)}

The inverted generational distance (IGD) is calculated as \eqn{IGD_p(A,R) = GD_p(R,A)}.

The modified inverted generational distanced (IGD+) was proposed by
\citet{IshMasTanNoj2015igd} to ensure that IGD+ is weakly Pareto compliant,
similarly to \code{\link[=epsilon_additive]{epsilon_additive()}} or \code{\link[=epsilon_mult]{epsilon_mult()}}. Formally,

\deqn{IGD^+(A,R) = \left(\frac{1}{|R|}\sum_{r\in R}\min_{a\in A} d^+(r,a)\right)}{IGD+(A,R) = (1/|R|) * ( sum_{r in R} min_{a in A} d^+(r,a))}

Assuming minimization of all objectives, IGD+ uses the following modified
distance:

\deqn{d^+(r,a) = \sqrt{\sum_{k=1}^m (\max\{a_k - r_k, 0\})^2}}{d^+(r,a) = sqrt(sum_{k=1}^m (max {a_k - r_k, 0 })^2)}

The average Hausdorff distance (\eqn{\Delta_p}) was proposed by
\citet{SchEsqLarCoe2012tec} and it is calculated as:

\deqn{\Delta_p(A,R) = \max\{ IGD_p(A,R), IGD_p(R,A) \}}

IGDX \citep{ZhoZhaJin2009igdx} is the application of IGD to decision vectors
instead of objective vectors to measure closeness and diversity in decision
space. One can use the functions \code{igd()} or \code{igd_plus()} (recommended)
directly, just passing the decision vectors as \code{data}.

There are different formulations of the GD and IGD metrics in the literature
that differ on the value of \eqn{p}, on the distance metric used and on
whether the term \eqn{|A|^{-1}} is inside (as above) or outside the exponent
\eqn{1/p}.  GD was first proposed by \citet{VelLam1998gp} with \eqn{p=2} and
the term \eqn{|A|^{-1}} outside the exponent. IGD seems to have been
mentioned first by \citet{CoeSie2004igd}, however, some people also used the
name D-metric for the same concept with \eqn{p=1} and later papers have
often used IGD/GD with \eqn{p=1}. \citet{SchEsqLarCoe2012tec} proposed to
place the term \eqn{|A|^{-1}} inside the exponent, as in the formulation
shown above.  This has a significant effect for GD and less so for IGD given
a constant reference set. IGD+ also follows this formulation.  We refer to
\citet{IshMasTanNoj2015igd} and \citet{BezLopStu2017emo} for a more detailed
historical perspective and a comparison of the various variants.

Following \citet{IshMasTanNoj2015igd}, we always use \eqn{p=1} in our
implementation of IGD and IGD+ because (1) it is the setting most used in
recent works; (2) it makes irrelevant whether the term \eqn{|A|^{-1}} is
inside or outside the exponent \eqn{1/p}; and (3) the meaning of IGD becomes
the average Euclidean distance from each reference point to its nearest
objective vector. It is also slightly faster to compute.

GD should never be used directly to compare the quality of approximations to
a Pareto front, because it is not weakly Pareto-compliant and often
contradicts Pareto optimality.

IGD is still popular due to historical reasons, but we strongly recommend
IGD+ instead of IGD, because IGD contradicts Pareto optimality in some cases
(see examples below) whereas IGD+ is weakly Pareto-compliant.

The average Hausdorff distance \eqn{\Delta_p(A,R)} is also not weakly
Pareto-compliant, as shown in the examples below.
}
\references{
\insertAllCited{}
}
\author{
Manuel \enc{López-Ibáñez}{Lopez-Ibanez}
}
\concept{metrics}
\examples{
# Example 4 from Ishibuchi et al. (2015)
ref <- matrix(c(10,0,6,1,2,2,1,6,0,10), ncol=2, byrow=TRUE)
A <- matrix(c(4,2,3,3,2,4), ncol=2, byrow=TRUE)
B <- matrix(c(8,2,4,4,2,8), ncol=2, byrow=TRUE)
if (requireNamespace("graphics", quietly = TRUE)) {
   plot(ref, xlab=expression(f[1]), ylab=expression(f[2]),
        panel.first=grid(nx=NULL), pch=23, bg="gray", cex=1.5)
   points(A, pch=1, cex=1.5)
   points(B, pch=19, cex=1.5)
   legend("topright", legend=c("Reference", "A", "B"), pch=c(23,1,19),
          pt.bg="gray", bg="white", bty = "n", pt.cex=1.5, cex=1.2)
}
cat("A is better than B in terms of Pareto optimality,\n however, IGD(A)=",
    igd(A, ref), "> IGD(B)=", igd(B, ref),
    "and AvgHausdorff(A)=", avg_hausdorff_dist(A, ref),
    "> AvgHausdorff(B)=", avg_hausdorff_dist(B, ref),
    ", which both contradict Pareto optimality.\nBy contrast, IGD+(A)=",
    igd_plus(A, ref), "< IGD+(B)=", igd_plus(B, ref), ", which is correct.\n")
# A less trivial example.
extdata_path <- system.file(package="moocore","extdata")
path.A1 <- file.path(extdata_path, "ALG_1_dat.xz")
path.A2 <- file.path(extdata_path, "ALG_2_dat.xz")
A1 <- read_datasets(path.A1)[,1:2]
A2 <- read_datasets(path.A2)[,1:2]
ref <- filter_dominated(rbind(A1, A2))
igd(A1, ref)
igd(A2, ref)

# IGD+ (Pareto compliant)
igd_plus(A1, ref)
igd_plus(A2, ref)

# Average Haussdorff distance
avg_hausdorff_dist(A1, ref)
avg_hausdorff_dist(A2, ref)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/read_datasets.R
\name{write_datasets}
\alias{write_datasets}
\title{Write data sets}
\usage{
write_datasets(x, file = "")
}
\arguments{
\item{x}{\code{matrix}|\code{data.frame()}\cr Dataset with at least three
columns, the last one is the set of each point. See also
\code{\link[=read_datasets]{read_datasets()}}.}

\item{file}{Either a character string naming a file or a connection open for
writing. \code{""} indicates output to the console.}
}
\value{
No return value, called for side effects
}
\description{
Write data sets to a file in the same format as \code{\link[=read_datasets]{read_datasets()}}.
}
\examples{
x <- read_datasets(text="1 2\n3 4\n\n5 6\n7 8\n", col_names=c("obj1", "obj2"))
write_datasets(x)

}
\seealso{
\code{\link[utils:write.table]{utils::write.table()}}, \code{\link[=read_datasets]{read_datasets()}}
}
\keyword{file}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{as_double_matrix}
\alias{as_double_matrix}
\title{Convert input to a matrix with \code{"double"} storage mode (\code{\link[base:mode]{base::storage.mode()}}).}
\usage{
as_double_matrix(x)
}
\arguments{
\item{x}{\code{data.frame()}|\code{matrix()}\cr A numerical data frame or matrix with at least 1 row and 2 columns.}
}
\value{
\code{x} is coerced to a numerical \code{matrix()}.
}
\description{
Convert input to a matrix with \code{"double"} storage mode (\code{\link[base:mode]{base::storage.mode()}}).
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nondominated.R
\name{pareto_rank}
\alias{pareto_rank}
\title{Rank points according to Pareto-optimality (nondominated sorting).}
\usage{
pareto_rank(x, maximise = FALSE)
}
\arguments{
\item{x}{\code{matrix()}|\code{data.frame()}\cr Matrix or data frame of numerical
values, where each row gives the coordinates of a point.}

\item{maximise}{\code{logical()}\cr Whether the objectives must be maximised
instead of minimised. Either a single logical value that applies to all
objectives or a vector of logical values, with one value per objective.}
}
\value{
An integer vector of the same length as the number of rows of the
input \code{x}, where each value gives the rank of each point (lower is
better).
}
\description{
\code{\link[=pareto_rank]{pareto_rank()}} is meant to be used like \code{rank()}, but it assigns ranks
according to Pareto dominance, where rank 1 indicates those solutions not
dominated by any other solution in the input set.  Duplicated points are
assigned the same rank.  The resulting ranking can be used to partition
points into a list of matrices, each matrix representing a nondominated
front \citep{Deb02nsga2} (see examples below).
}
\details{
Given a finite set of points \eqn{X \subset \mathbb{R}^m},
the rank of a point \eqn{x \in X} is defined as:

\deqn{\operatorname{rank}(x) = r \iff x \in F^c_{r} \land \nexists y \in F^c_{r}, y \prec x}

where \eqn{y \prec x} means that \eqn{y} dominates \eqn{x} according to
Pareto optimality, \eqn{F^c_r = X \setminus \bigcup_{i=1}^{r-1} F_i} and
\eqn{F_r = \{x \in X \land \operatorname{rank}(x) = r\}}.  The sets
\eqn{F_c}, with \eqn{c=1,\dots,k}, partition \eqn{X} into \eqn{k}
\emph{fronts}, that is, mutually nondominated subsets of \eqn{X}.

With \eqn{m=2}, i.e., \code{ncol(data)=2}, the code uses the best-known
\eqn{O(n \log n)} algorithm by \citet{Jen03}.  When \eqn{m \geq 3}, it
uses the naive algorithm that identifies one \emph{front} at a time, which
requires \eqn{O(n^2\log n)} for \eqn{m=3}, and \eqn{O(n^2 \log^{m-2} n)}
for \eqn{m \geq 4}.
}
\references{
\insertAllCited{}
}
\seealso{
\code{\link[=is_nondominated]{is_nondominated()}}
}
\concept{dominance}
\examples{
three_fronts = matrix(c(1, 2, 3,
                        3, 1, 2,
                        2, 3, 1,
                        10, 20, 30,
                        30, 10, 20,
                        20, 30, 10,
                        100, 200, 300,
                        300, 100, 200,
                        200, 300, 100), ncol=3, byrow=TRUE)
pareto_rank(three_fronts)

split.data.frame(three_fronts, pareto_rank(three_fronts))
path_A1 <- file.path(system.file(package="moocore"),"extdata","ALG_1_dat.xz")
set <- read_datasets(path_A1)[,1:2]
ranks <- pareto_rank(set)
str(ranks)
if (requireNamespace("graphics", quietly = TRUE)) {
   colors <- colorRampPalette(c("red","yellow","springgreen","royalblue"))(max(ranks))
   plot(set, col = colors[ranks], type = "p", pch = 20)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/whv.R
\name{whv_rect}
\alias{whv_rect}
\alias{total_whv_rect}
\title{Compute (total) weighted hypervolume given a set of rectangles}
\usage{
whv_rect(x, rectangles, reference, maximise = FALSE)

total_whv_rect(
  x,
  rectangles,
  reference,
  maximise = FALSE,
  ideal = NULL,
  scalefactor = 0.1
)
}
\arguments{
\item{x}{\code{matrix()}|\code{data.frame()}\cr Matrix or data frame of numerical
values, where each row gives the coordinates of a point.}

\item{rectangles}{\code{matrix()}\cr Weighted rectangles that will bias the
computation of the hypervolume. Maybe generated by \code{\link[=eafdiff]{eafdiff()}} with
\code{rectangles=TRUE} or by \code{\link[=choose_eafdiff]{choose_eafdiff()}}.}

\item{reference}{\code{numeric()}\cr Reference point as a vector of numerical
values.}

\item{maximise}{\code{logical()}\cr Whether the objectives must be maximised
instead of minimised. Either a single logical value that applies to all
objectives or a vector of logical values, with one value per objective.}

\item{ideal}{\code{numeric()}\cr Ideal point as a vector of numerical values.  If
\code{NULL}, it is calculated as minimum (or maximum if maximising that
objective) of each objective in the input data.}

\item{scalefactor}{\code{numeric(1)}\cr Real value within \eqn{(0,1]} that scales
the overall weight of the differences. This is parameter psi (\eqn{\psi})
in \citet{DiaLop2020ejor}.}
}
\value{
\code{numeric(1)} A single numerical value.
}
\description{
Calculates the hypervolume weighted by a set of rectangles (with zero weight
outside the rectangles). The function \code{\link[=total_whv_rect]{total_whv_rect()}} calculates the
total weighted hypervolume as \code{\link[=hypervolume]{hypervolume()}}\code{ + scalefactor * abs(prod(reference - ideal)) * whv_rect()}. The details of the computation
are given by \citet{DiaLop2020ejor}.
}
\details{
TODO
}
\references{
\insertAllCited{}
}
\seealso{
\code{\link[=read_datasets]{read_datasets()}}, \code{\link[=eafdiff]{eafdiff()}}, \code{\link[=choose_eafdiff]{choose_eafdiff()}}, \code{\link[=whv_hype]{whv_hype()}}
}
\concept{metrics}
\examples{
rectangles <- as.matrix(read.table(header=FALSE, text='
 1.0  3.0  2.0  Inf    1
 2.0  3.5  2.5  Inf    2
 2.0  3.0  3.0  3.5    3
'))
whv_rect(matrix(2, ncol=2), rectangles, reference = 6)
whv_rect(matrix(c(2, 1), ncol=2), rectangles, reference = 6)
whv_rect(matrix(c(1, 2), ncol=2), rectangles, reference = 6)
total_whv_rect(matrix(2, ncol=2), rectangles, reference = 6, ideal = c(1,1))
total_whv_rect(matrix(c(2, 1), ncol=2), rectangles, reference = 6, ideal = c(1,1))
total_whv_rect(matrix(c(1, 2), ncol=2), rectangles, reference = 6, ideal = c(1,1))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/normalise.R
\name{normalise}
\alias{normalise}
\title{Normalise points}
\usage{
normalise(x, to_range = c(1, 2), lower = NA, upper = NA, maximise = FALSE)
}
\arguments{
\item{x}{\code{matrix()}|\code{data.frame()}\cr Matrix or data frame of numerical
values, where each row gives the coordinates of a point.}

\item{to_range}{\code{numerical(2)}\cr Normalise values to this range. If the objective is
maximised, it is normalised to \code{c(to_range[1], to_range[0])}
instead.}

\item{lower, upper}{\code{numerical()}\cr Bounds on the values. If \code{NA}, the maximum and minimum
values of each coordinate are used.}

\item{maximise}{\code{logical()}\cr Whether the objectives must be maximised
instead of minimised. Either a single logical value that applies to all
objectives or a vector of logical values, with one value per objective.}
}
\value{
\code{matrix()}\cr A numerical matrix
}
\description{
Normalise points per coordinate to a range, e.g., \code{c(1,2)}, where the
minimum value will correspond to 1 and the maximum to 2. If bounds are
given, they are used for the normalisation.
}
\examples{

data(SPEA2minstoptimeRichmond)
# The second objective must be maximized
head(SPEA2minstoptimeRichmond[, 1:2])

head(normalise(SPEA2minstoptimeRichmond[, 1:2], maximise = c(FALSE, TRUE)))

head(normalise(SPEA2minstoptimeRichmond[, 1:2], to_range = c(0,1), maximise = c(FALSE, TRUE)))

}
\author{
Manuel \enc{López-Ibáñez}{Lopez-Ibanez}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/eaf.R
\name{attsurf2df}
\alias{attsurf2df}
\title{Convert a list of attainment surfaces to a single EAF \code{data.frame}.}
\usage{
attsurf2df(x)
}
\arguments{
\item{x}{\code{list()}\cr List of \code{data.frames} or matrices. The names of the list
give the percentiles of the attainment surfaces.  This is the format
returned by \code{\link[=eaf_as_list]{eaf_as_list()}}.}
}
\value{
\code{data.frame()}\cr Data frame with as many columns as objectives and an additional column \code{percentiles}.
}
\description{
Convert a list of attainment surfaces to a single EAF \code{data.frame}.
}
\examples{

data(SPEA2relativeRichmond)
attsurfs <- eaf_as_list(eaf(SPEA2relativeRichmond, percentiles = c(0,50,100)))
str(attsurfs)
eaf_df <- attsurf2df(attsurfs)
str(eaf_df)
}
\seealso{
\code{\link[=eaf_as_list]{eaf_as_list()}}
}
\concept{eaf}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/whv.R
\name{whv_hype}
\alias{whv_hype}
\title{Approximation of the (weighted) hypervolume by Monte-Carlo sampling (2D only)}
\usage{
whv_hype(
  x,
  reference,
  ideal,
  maximise = FALSE,
  nsamples = 100000L,
  seed = NULL,
  dist = "uniform",
  mu = NULL
)
}
\arguments{
\item{x}{\code{matrix()}|\code{data.frame()}\cr Matrix or data frame of numerical
values, where each row gives the coordinates of a point.}

\item{reference}{\code{numeric()}\cr Reference point as a vector of numerical
values.}

\item{ideal}{\code{numeric()}\cr Ideal point as a vector of numerical values.}

\item{maximise}{\code{logical()}\cr Whether the objectives must be maximised
instead of minimised. Either a single logical value that applies to all
objectives or a vector of logical values, with one value per objective.}

\item{nsamples}{\code{integer(1)}\cr Number of samples for Monte-Carlo
sampling. Higher values typically produce more accurate approximations of
the true hypervolume, but require more time.}

\item{seed}{\code{integer(1)}\cr Random seed.}

\item{dist}{\code{character(1)}\cr Weight distribution type. See Details.}

\item{mu}{\code{numeric()}\cr Parameter of the weight distribution. See Details.}
}
\value{
A single numerical value.
}
\description{
Return an estimation of the hypervolume of the space dominated by the input
data following the procedure described by \citet{AugBadBroZit2009gecco}. A
weight distribution describing user preferences may be specified.
}
\details{
The current implementation only supports 2 objectives.

A weight distribution  \citep{AugBadBroZit2009gecco} can be provided via the \code{dist} argument. The ones currently supported are:
\itemize{
\item \code{"uniform"} corresponds to the default hypervolume (unweighted).
\item \code{"point"} describes a goal in the objective space, where the parameter \code{mu} gives the coordinates of the goal. The resulting weight distribution is a multivariate normal distribution centred at the goal.
\item \code{"exponential"} describes an exponential distribution with rate parameter \code{1/mu}, i.e., \eqn{\lambda = \frac{1}{\mu}}.
}
}
\references{
\insertAllCited{}
}
\seealso{
\code{\link[=read_datasets]{read_datasets()}}, \code{\link[=eafdiff]{eafdiff()}}, \code{\link[=whv_rect]{whv_rect()}}
}
\concept{metrics}
\examples{
whv_hype(matrix(2, ncol=2), reference = 4, ideal = 1, seed = 42)
whv_hype(matrix(c(3,1), ncol=2), reference = 4, ideal = 1, seed = 42)
whv_hype(matrix(2, ncol=2), reference = 4, ideal = 1, seed = 42,
         dist = "exponential", mu=0.2)
whv_hype(matrix(c(3,1), ncol=2), reference = 4, ideal = 1, seed = 42,
         dist = "exponential", mu=0.2)
whv_hype(matrix(2, ncol=2), reference = 4, ideal = 1, seed = 42,
         dist = "point", mu=c(2.9,0.9))
whv_hype(matrix(c(3,1), ncol=2), reference = 4, ideal = 1, seed = 42,
         dist = "point", mu=c(2.9,0.9))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/generate.R
\name{generate_ndset}
\alias{generate_ndset}
\title{Generate a random set of mutually nondominated points}
\usage{
generate_ndset(n, d, method, seed = NULL, integer = FALSE)
}
\arguments{
\item{n}{\code{integer(1)}\cr Number of rows in the output.}

\item{d}{\code{integer(1)}\cr Number of columns in the output.}

\item{method}{\code{character(1)}\cr Method used to generate the random nondominated set.
See \strong{Details} below for more information.}

\item{seed}{\code{integer(1)}\cr Integer seed for random number generation. If \code{NULL},
a random seed is generated.}

\item{integer}{\code{logical(1)}\cr If \code{TRUE}, return integer-valued points.}
}
\value{
A numeric matrix of size \verb{n × d} containing nondominated points.
}
\description{
Generate a random set of \code{n} mutually nondominated points of dimension \code{d}
with the shape defined by \code{method}.

When \code{integer = FALSE} (the default), the points are generated within the
hypercube \eqn{(0,1)^d} and can be scaled to another range using
\code{\link[=normalise]{normalise()}}.  Otherwise, points are scaled to a non-negative integer range
that keeps the points mutually nondominated.

See the visualisations in the vignette \HTMLVignette{generate}{Sampling Random Nondominated Sets}.
}
\details{
The available methods are:

\describe{
\item{\code{'simplex'}, \code{'linear'}, or \code{'L'}}{

Uniformly samples points on the standard simplex.  This shape of
nondominated set is also called \code{'linear'} in the literature
\citep{LacKlaFon2017box}.

The standard \eqn{(d-1)}-simplex is defined by \eqn{\{x \in \mathbb{R}_+^d :
   \sum_i x_i = 1\}}.  Each point \eqn{\vec{z} \in (0,1)^d \subset
   \mathbb{R}^d} is generated by sampling \eqn{d} independent and
identically distributed values \eqn{(x_1,x_2, \dots, x_d)} from the
exponential distribution, then dividing each value by the L1-norm of the
vector, \eqn{z_i = x_i / \sum_{i=1}^d x_i} \citep{RubMel1998simulation}.
Values sampled from the exponential distribution are guaranteed to be
positive.  Sampling from either the standard normal distribution
\citep{GueFonPaq2021hv} or the uniform distribution
\citep{LacKlaFon2017box} does not produce a uniform distribution when
projected onto the simplex.

}
\item{\code{'concave-sphere'}, \code{'sphere'}, or \code{'C'}}{

Uniformly samples points on the positive orthant of the hyper-sphere,
which is concave when all objectives are minimised.

Each point \eqn{\vec{z} \in (0,1)^d \subset \mathbb{R}^d} is generated by
sampling \eqn{d} independent and identically distributed values
\eqn{\vec{x}=(x_1,x_2, \dots, x_d)} from the standard normal
distribution, then dividing each value by the l2-norm of the vector,
\eqn{z_i = \frac{|x_i|}{\|\vec{x}\|_2}} \citep{Muller1959sphere}. The
absolute value in the numerator ensures that points are sampled on the
positive orthant of the hyper-sphere.  Sampling from the uniform
distribution \citep{LacKlaFon2017box} would not result in a uniform
sampling when projected onto the surface of the hyper-sphere.

}
\item{\code{'convex-sphere'} or \code{'X'}}{

Equivalent to \code{1 - generate_ndset(..., method='concave-sphere')}, which
is convex for minimisation.  This shape has also been called \emph{inverted
convex} \citep{IshHeSha2019regular}. This sampling is uniform.

It corresponds to translating points from the negative orthant of the
hyper-sphere to the positive orthant. Thus, the sampling remains uniform.

}
\item{\code{'convex-simplex'}}{

Equivalent to \code{generate_ndset(..., method='simplex')^2}, which is convex
for minimisation. Such a set cannot be obtained by any affine
transformation of a subset of the hyper-sphere.  This sampling is \emph{not}
uniform.

The corresponding surface is equivalent to a simplex curved towards the
origin. The generated points \eqn{\vec{z} \in (0,1)^d \subset
   \mathbb{R}^d} satisfy \eqn{\sum_{i=1}^d \sqrt{z_i} = 1}.  Although the
sampling on the simplex is uniform, the transformed points are not.

}
\item{\code{'concave-simplex'}}{

Equivalent to \code{1 - generate_ndset(..., method='convex-simplex')}, which
is concave for minimisation.  This shape has also been called \emph{inverted
concave} \citep{IshHeSha2019regular}.  This sampling is \emph{not} uniform
because \code{method='convex-simplex'} is also not uniform.

}
\item{\code{'inverted-simplex'} or \code{'inverted-linear'}}{

Equivalent to \code{1 - generate_ndset(..., method='simplex')}. This sampling
is uniform.

}
}

Methods \code{'inverted-simplex'}, \code{'concave-simplex'} and \code{'convex-sphere'} are
translations of \code{'simplex'}, \code{'convex-simplex'} and \code{'concave'}, respectively.
These translations have been called \emph{inverted} shapes in the literature and
have different properties than their \emph{regular} counterparts
\citep{IshHeSha2019regular}.
}
\examples{
generate_ndset(5, 3, "simplex", seed = 42)
generate_ndset(5, 3, "simplex", seed = 42, integer = TRUE)
generate_ndset(4, 2, "sphere", seed = 123)
generate_ndset(3, 5, "convex-sphere", seed = 123)
generate_ndset(4, 4, "convex-simplex", seed = 123)

}
\references{
\insertAllCited{}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/moocore-package.R
\docType{data}
\name{HybridGA}
\alias{HybridGA}
\title{Results of Hybrid GA on Vanzyl and Richmond water networks}
\format{
A list with two data frames, each of them with three columns, as
produced by \code{\link[=read_datasets]{read_datasets()}}.
\describe{
\item{\verb{$vanzyl}}{data frame of results on Vanzyl network}
\item{\verb{$richmond}}{data frame of results on Richmond
network. The second column is filled with \code{NA}}
}
}
\source{
\insertRef{LopezIbanezPhD}{moocore}.
}
\usage{
HybridGA
}
\description{
Results of Hybrid GA on Vanzyl and Richmond water networks
}
\examples{
data(HybridGA)
print(HybridGA$vanzyl)
print(HybridGA$richmond)
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/hv_approx.R
\name{hv_approx}
\alias{hv_approx}
\title{Approximate the hypervolume indicator.}
\usage{
hv_approx(
  x,
  reference,
  maximise = FALSE,
  nsamples = 262144L,
  seed = NULL,
  method = c("Rphi-FWE+", "DZ2019-HW", "DZ2019-MC")
)
}
\arguments{
\item{x}{\code{matrix()}|\code{data.frame()}\cr Matrix or data frame of numerical
values, where each row gives the coordinates of a point.}

\item{reference}{\code{numeric()}\cr Reference point as a vector of numerical
values.}

\item{maximise}{\code{logical()}\cr Whether the objectives must be maximised
instead of minimised. Either a single logical value that applies to all
objectives or a vector of logical values, with one value per objective.}

\item{nsamples}{\code{integer(1)}\cr Number of samples for Monte-Carlo
sampling. Higher values typically produce more accurate approximations of
the true hypervolume, but require more time.}

\item{seed}{\code{integer(1)}\cr Random seed.}

\item{method}{\code{character(1)}\cr Method to generate the sampling weights. See `Details'.}
}
\value{
A single numerical value.
}
\description{
Approximate the value of the hypervolume metric with respect to a given
reference point assuming minimization of all objectives. Methods
\code{"Rphi-FWE+"} and \code{"DZ2019-HW"} are deterministic and ignore the parameter
\code{seed}, while \code{method="DZ2019-MC"} relies on Monte-Carlo sampling
\citep{DenZha2019approxhv}.  All methods tend to get more accurate with
higher values of \code{nsamples}, but the increase in accuracy is not monotonic,
as shown in the vignette \HTMLVignette{hv_approx}{Approximating the
hypervolume}.
}
\details{
All available methods approximate the hypervolume as a
\eqn{(m-1)}-dimensional integral over the surface of hypersphere
\citep{DenZha2019approxhv}:

\deqn{\widehat{HV}_r(A) = \frac{2\pi^\frac{m}{2}}{\Gamma(\frac{m}{2})}\frac{1}{m 2^m}\frac{1}{n}\sum_{i=1}^n \max_{y \in A} s(w^{(i)}, y)^m}

where \eqn{m} is the number of objectives, \eqn{w^{(i)}} are weights
uniformly distributed on \eqn{S_{+}}, i.e., the positive orthant of the
\eqn{(m-1)}-D unit hypersphere, \eqn{n} is the number of weights sampled,
\eqn{\Gamma()} is the gamma function \code{\link[=gamma]{gamma()}}, i.e., the analytical
continuation of the factorial function, and \eqn{s(w, y) = \min_{k=1}^m (r_k
- y_k)/w_k}.

In the default \code{method="Rphi-FWE+"} \citep{Lop2026hvapprox}, the weights
\eqn{w^{(i)}, i=1\ldots n} are defined using the deterministic
low-discrepancy sequence \eqn{R_\phi} \citep{Rob2018unreasonable} mapped to
the positive orthant of the hypersphere using a modified version of Fang and
Wang efficient mapping \citep{FanWan1994numtheory}.

In \code{method="DZ2019-HW"} \citep{DenZha2019approxhv}, the weights
\eqn{w^{(i)}, i=1\ldots n} are defined using a deterministic low-discrepancy
sequence. The weight values depend on their number (\code{nsamples}), thus
increasing the number of weights may not necessarily increase accuracy
because the set of weights would be different.

In \code{method="DZ2019-MC"} \citep{DenZha2019approxhv}, the weights
\eqn{w^{(i)}, i=1\ldots n} are sampled from the unit normal vector such that
each weight \eqn{w = \frac{|x|}{\|x\|_2}} where each component of \eqn{x} is
independently sampled from the standard normal distribution
\citep{Muller1959sphere}.

The original source code in C++/MATLAB for both \code{"DZ2019-HW"} and
\code{"DZ2019-MC"} methods can be found at
\url{https://github.com/Ksrma/Hypervolume-Approximation-using-polar-coordinate}.

\citet{Lop2026hvapprox} empirically shows that \code{"Rphi-FWE+"} typically
produces an approximation error as low as the other methods, with a
computational cost similar to \code{"DZ2019-MC"} and significantly faster than
\code{"DZ2019-HW"}.
}
\references{
\insertAllCited{}
}
\author{
Manuel \enc{López-Ibáñez}{Lopez-Ibanez}
}
\concept{metrics}
\examples{
x <- matrix(c(5, 5, 4, 6, 2, 7, 7, 4), ncol=2, byrow=TRUE)
hypervolume(x, ref=10)
hv_approx(x, ref=10, method="Rphi-FWE+")
hv_approx(x, ref=10, method="DZ2019-HW")
hv_approx(x, ref=10, seed=42, method="DZ2019-MC")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/nondominated.R
\name{is_nondominated}
\alias{is_nondominated}
\alias{filter_dominated}
\alias{any_dominated}
\title{Identify and remove dominated points according to Pareto optimality}
\usage{
is_nondominated(x, maximise = FALSE, keep_weakly = FALSE)

filter_dominated(x, maximise = FALSE, keep_weakly = FALSE)

any_dominated(x, maximise = FALSE, keep_weakly = FALSE)
}
\arguments{
\item{x}{\code{matrix()}|\code{data.frame()}\cr Matrix or data frame of numerical
values, where each row gives the coordinates of a point.}

\item{maximise}{\code{logical()}\cr Whether the objectives must be maximised
instead of minimised. Either a single logical value that applies to all
objectives or a vector of logical values, with one value per objective.}

\item{keep_weakly}{\code{logical(1)}\cr If \code{FALSE}, return \code{FALSE} for any
duplicates of nondominated points, except the last one.}
}
\value{
\code{\link[=is_nondominated]{is_nondominated()}} returns a logical vector of the same length
as the number of rows of \code{data}, where \code{TRUE} means that the
point is not dominated by any other point.

\code{\link[=filter_dominated]{filter_dominated()}} returns a matrix or data.frame with only mutually nondominated points.

\code{\link[=any_dominated]{any_dominated()}} returns \code{TRUE} if \code{x} contains any (weakly-)dominated points, \code{FALSE} otherwise.
}
\description{
Identify nondominated points with \code{\link[=is_nondominated]{is_nondominated()}} and remove dominated
ones with \code{\link[=filter_dominated]{filter_dominated()}}.

\code{\link[=any_dominated]{any_dominated()}} quickly detects if a set contains any dominated point.
}
\details{
Given \eqn{n} points of dimension \eqn{m}, the current implementation always
uses the best-known \eqn{O(n \log n)} dimension-sweep algorithm
\citep{KunLucPre1975jacm} for \eqn{m \leq 3}. For \eqn{m \geq 4}, functions
\code{is_nondominated()} and \code{filter_dominated()} use the best-known \eqn{O(n
\log^{m-2} n)} algorithm \citep{KunLucPre1975jacm} when \eqn{n > 16}, and
the naive \eqn{O(m n^2)} algorithm otherwise.  Function \code{any_dominated()}
always uses the naive algorithm for \eqn{m \geq 4}.
}
\references{
\insertAllCited{}
}
\seealso{
\code{\link[=pareto_rank]{pareto_rank()}}
}
\author{
Manuel \enc{López-Ibáñez}{Lopez-Ibanez}
}
\concept{dominance}
\examples{
S = matrix(c(1,1,0,1,1,0,1,0), ncol = 2, byrow = TRUE)
is_nondominated(S)
is_nondominated(S, maximise = TRUE)
filter_dominated(S)
filter_dominated(S, keep_weakly = TRUE)
any_dominated(S)
any_dominated(S, keep_weakly = TRUE)
any_dominated(filter_dominated(S))
path_A1 <- file.path(system.file(package="moocore"),"extdata","ALG_1_dat.xz")
set <- read_datasets(path_A1)[,1:2]
is_nondom <- is_nondominated(set)
cat("There are ", sum(is_nondom), " nondominated points\n")

if (requireNamespace("graphics", quietly = TRUE)) {
   plot(set, col = "blue", type = "p", pch = 20)
   ndset <- filter_dominated(set)
   points(ndset[order(ndset[,1]),], col = "red", pch = 21)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/largest_eafdiff.R
\name{largest_eafdiff}
\alias{largest_eafdiff}
\title{Identify largest EAF differences}
\usage{
largest_eafdiff(x, maximise = FALSE, intervals = 5L, reference, ideal = NULL)
}
\arguments{
\item{x}{\code{list()}\cr A list of matrices or data frames with at least 3 columns (last column indicates the set).}

\item{maximise}{\code{logical()}\cr Whether the objectives must be maximised
instead of minimised. Either a single logical value that applies to all
objectives or a vector of logical values, with one value per objective.}

\item{intervals}{\code{integer(1)}\cr The absolute range of the differences
\eqn{[0, 1]} is partitioned into the number of intervals provided.}

\item{reference}{\code{numeric()}\cr Reference point as a vector of numerical
values.}

\item{ideal}{\code{numeric()}\cr Ideal point as a vector of numerical values.  If
\code{NULL}, it is calculated as minimum (or maximum if maximising that
objective) of each objective in the input data.}
}
\value{
\code{list()}\cr A list with two components \code{pair} and \code{value}.
}
\description{
Given a list of datasets, return the indexes of the pair with the largest
EAF differences according to the method proposed by \citet{DiaLop2020ejor}.
}
\examples{
# FIXME: This example is too large, we need a smaller one.
data(tpls50x20_1_MWT)
nadir <- apply(tpls50x20_1_MWT[,2:3], 2L, max)
x <- largest_eafdiff(split.data.frame(tpls50x20_1_MWT[,2:4], tpls50x20_1_MWT[, 1L]),
                     reference = nadir)
str(x)

}
\references{
\insertAllCited{}
}
\concept{eaf}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/moocore-package.R
\docType{data}
\name{SPEA2relativeVanzyl}
\alias{SPEA2relativeVanzyl}
\title{Results of SPEA2 with relative time-controlled triggers on Vanzyl's
water network.}
\format{
An object of class \code{data.frame} with 107 rows and 3 columns.
}
\source{
\insertRef{LopezIbanezPhD}{moocore}
}
\usage{
SPEA2relativeVanzyl
}
\description{
Results of SPEA2 with relative time-controlled triggers on Vanzyl's
water network.
}
\examples{
data(SPEA2relativeVanzyl)
str(SPEA2relativeVanzyl)
}
\keyword{datasets}
